/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/

/*
 *  binarizelow.c
 *                     
 *      Floyd-Steinberg dithering
 *          void       ditherFSLow()
 *          void       ditherFSLineLow()
 *
 *      Simple (pixelwise) binarization with fixed threshold
 *          void       thresholdToBinaryLow()
 *          void       thresholdToBinaryLineLow()
 *
 */

#include <stdio.h>
#include <string.h>

#include "allheaders.h"

    /* for ditherFSLow(): if within this distance from white or black,
     * set to pure white and black, and do not propagate error */
#define   CLIP_UPPER_TO_WHITE     10
#define   CLIP_LOWER_TO_BLACK     10



/*------------------------------------------------------------------*
 *             Binarization by Floyd-Steinberg Dithering            *
 *------------------------------------------------------------------*/
/*
 *  ditherFSLow()
 *
 */
void
ditherFSLow(UINT32  *datad,
            INT32    w,
            INT32    h,
            INT32    wpld,
	    UINT32  *datas,
	    INT32    wpls,
	    UINT32  *bufs1,
	    UINT32  *bufs2)
{
INT32      i;
UINT32    *lined;

        /* do all lines except last line */
    memcpy(bufs2, datas, 4 * wpls);  /* prime the buffer */
    for (i = 0; i < h - 1; i++) {
	memcpy(bufs1, bufs2, 4 * wpls);
	memcpy(bufs2, datas + (i + 1) * wpls, 4 * wpls);
	lined = datad + i * wpld;
	ditherFSLineLow(lined, w, bufs1, bufs2, 0);
    }

        /* do last line */
    memcpy(bufs1, bufs2, 4 * wpls);
    lined = datad + (h - 1) * wpld;
    ditherFSLineLow(lined, w, bufs1, bufs2, 1);
    return;
}


/*
 *  ditherFSLineLow()
 *   
 *       Input:  lined  (ptr to beginning of dest line
 *               w   (width of image in pixels)
 *               bufs1 (buffer of current source line)
 *               bufs2 (buffer of next source line)
 *               lastlineflag  (0 if not last dest line, 1 if last dest line)
 *       Return: void
 *
 *       Dispatches FS error diffusion dithering for
 *       a single line of the image.  If lastlineflag == 0,
 *       both source buffers are used; otherwise, only bufs1
 *       is used.  We use source buffers because the error
 *       is propagated into them, and we don't want to change
 *       the input src image. 
 *
 *       We break dithering out line by line to make it
 *       easier to combine functions like interpolative
 *       scaling and error diffusion dithering, as such a
 *       combination of operations obviates the need to
 *       generate a 2x grayscale image as an intermediary.
 */
void
ditherFSLineLow(UINT32  *lined,
                INT32    w,
		UINT32  *bufs1,
		UINT32  *bufs2,
		INT32    lastlineflag)
{
INT32      j;
INT32      oval, eval;
UINT8      fval1, fval2, rval, bval, dval;

    if (lastlineflag == 0) {
	for (j = 0; j < w - 1; j++) {
	    oval = GET_DATA_BYTE(bufs1, j);
	    if (oval > 127) {   /* binarize to OFF */
		if ((eval = 255 - oval) > CLIP_UPPER_TO_WHITE) {
			/* subtract from neighbors */
		    fval1 = (3 * eval) / 8;
		    fval2 = eval / 4;
		    rval = GET_DATA_BYTE(bufs1, j + 1);
		    rval = MAX(0, rval - fval1);
		    SET_DATA_BYTE(bufs1, j + 1, rval);
		    bval = GET_DATA_BYTE(bufs2, j);
		    bval = MAX(0, bval - fval1);
		    SET_DATA_BYTE(bufs2, j, bval);
		    dval = GET_DATA_BYTE(bufs2, j + 1);
		    dval = MAX(0, dval - fval2);
		    SET_DATA_BYTE(bufs2, j + 1, dval);
		}
	    }
	    else {   /* oval <= 127; binarize to ON  */
		SET_DATA_BIT(lined, j);   /* ON pixel */
		if (oval > CLIP_LOWER_TO_BLACK) {
		        /* add to neighbors */
		    fval1 = (3 * oval) / 8;
		    fval2 = oval / 4;
		    rval = GET_DATA_BYTE(bufs1, j + 1);
		    rval = MIN(255, rval + fval1);
		    SET_DATA_BYTE(bufs1, j + 1, rval);
		    bval = GET_DATA_BYTE(bufs2, j);
		    bval = MIN(255, bval + fval1);
		    SET_DATA_BYTE(bufs2, j, bval);
		    dval = GET_DATA_BYTE(bufs2, j + 1);
		    dval = MIN(255, dval + fval2);
		    SET_DATA_BYTE(bufs2, j + 1, dval);
		}
	    }
	}

	    /* do last column: j = w - 1 */
	oval = GET_DATA_BYTE(bufs1, j);
	if (oval > 127) {  /* binarize to OFF */
	    if ((eval = 255 - oval) > CLIP_UPPER_TO_WHITE) {
		    /* subtract from neighbors */
		fval1 = (3 * eval) / 8;
		bval = GET_DATA_BYTE(bufs2, j);
		bval = MAX(0, bval - fval1);
		SET_DATA_BYTE(bufs2, j, bval);
	    }
	}
	else {  /*oval <= 127; binarize to ON */
	    SET_DATA_BIT(lined, j);   /* ON pixel */
	    if (oval > CLIP_LOWER_TO_BLACK) { 
		    /* add to neighbors */
		fval1 = (3 * oval) / 8;
		bval = GET_DATA_BYTE(bufs2, j);
		bval = MIN(255, bval + fval1);
		SET_DATA_BYTE(bufs2, j, bval);
	    }
	}
    }
    else {   /* lastlineflag == 1 */
	for (j = 0; j < w - 1; j++) {
	    oval = GET_DATA_BYTE(bufs1, j);
	    if (oval > 127) {   /* binarize to OFF */
		if ((eval = 255 - oval) > CLIP_UPPER_TO_WHITE) {
		        /* subtract from neighbors */
		    fval1 = (3 * eval) / 8;
		    rval = GET_DATA_BYTE(bufs1, j + 1);
		    rval = MAX(0, rval - fval1);
		    SET_DATA_BYTE(bufs1, j + 1, rval);
		}
	    }
	    else {   /* oval <= 127; binarize to ON  */
		SET_DATA_BIT(lined, j);   /* ON pixel */
		if (oval > CLIP_LOWER_TO_BLACK) { 
		        /* add to neighbors */
		    fval1 = (3 * oval) / 8;
		    rval = GET_DATA_BYTE(bufs1, j + 1);
		    rval = MIN(255, rval + fval1);
		    SET_DATA_BYTE(bufs1, j + 1, rval);
		}
	    }
	}

	    /* do last pixel: (i, j) = (h - 1, w - 1) */
	oval = GET_DATA_BYTE(bufs1, j);
	if (oval < 128)
	    SET_DATA_BIT(lined, j);   /* ON pixel */
    }

    return;
}



/*------------------------------------------------------------------*
 *             Simple binarization with fixed threshold             *
 *------------------------------------------------------------------*/
/*
 *  thresholdToBinaryLow()
 *
 */
void
thresholdToBinaryLow(UINT32  *datad,
                     INT32    w,
                     INT32    h,
                     INT32    wpld,
	             UINT32  *datas,
                     INT32    d,
	             INT32    wpls,
	             INT32    thresh)
{
INT32      i;
UINT32    *lines, *lined;

    for (i = 0; i < h; i++) {
	lines = datas + i * wpls;
	lined = datad + i * wpld;
	thresholdToBinaryLineLow(lined, w, lines, d, thresh);
    }
    return;
}


/*
 *  thresholdToBinaryLineLow()
 *
 */
void
thresholdToBinaryLineLow(UINT32  *lined,
                         INT32    w,
			 UINT32  *lines,
			 INT32    d,
			 INT32    thresh)
{
INT32    j, gval;

    switch (d)
    {
    case 4:
	for (j = 0; j < w; j++) {
	    gval = GET_DATA_QBIT(lines, j);
	    if (gval < thresh) 
		SET_DATA_BIT(lined, j);
	}
	break;
    case 8:
	for (j = 0; j < w; j++) {
	    gval = GET_DATA_BYTE(lines, j);
	    if (gval < thresh) 
		SET_DATA_BIT(lined, j);
	}
	break;
    default:
        fprintf(stderr, "in thresholdToBinaryLineLow(): invalid depth\n");
    }
    return;
}


#if 0
/*
 *  thresholdToBinaryLow()
 *
 */
void
thresholdToBinaryLow(UINT32  *datad,
                     INT32    w,
                     INT32    h,
                     INT32    wpld,
	             UINT32  *datas,
                     INT32    d,
	             INT32    wpls,
	             INT32    thresh)
{
INT32      i, j;
INT32      gval;
UINT32    *lines, *lined;

    for (i = 0; i < h; i++) {
	lines = datas + i * wpls;
	lined = datad + i * wpld;
	if (d == 4) {
	    for (j = 0; j < w; j++) {
		gval = GET_DATA_QBIT(lines, j);
		if (gval < thresh) 
		    SET_DATA_BIT(lined, j);
	    }
	}
	else {  /* d == 8 */
	    for (j = 0; j < w; j++) {
		gval = GET_DATA_BYTE(lines, j);
		if (gval < thresh) 
		    SET_DATA_BIT(lined, j);
	    }
	}
    }
    return;
}
#endif

