/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *  convolve.c
 *
 *      Top level grayscale or color block convolution
 *          PIX      *pixBlockconv()
 *
 *      Color block convolution
 *          PIX      *pixBlockconvColor()
 *          
 *      Grayscale block convolution
 *          PIX      *pixBlockconvGray()
 *          PIX      *pixBlockconvAccum()
 *
 *      Binary block sum and rank filter
 *          PIX      *pixBlockrank()
 *          PIX      *pixBlocksum()
 */

#include <stdio.h>
#include <stdlib.h>

#include "allheaders.h"




/*----------------------------------------------------------------------*
 *             Top-level grayscale or color block convolution           *
 *----------------------------------------------------------------------*/
/*!
 *  pixBlockconv()
 *
 *          Input:  pixs (8 or 32 bpp)
 *                  wc, hc   (half width/height of convolution kernel)
 *          Return: pixd, or null on error
 *
 *          Note:  The full width and height of the convolution kernel
 *                  are (2 * wc + 1) and (2 * hc + 1).
 */
PIX  *
pixBlockconv(PIX    *pixs,
             INT32   wc,
	     INT32   hc)
{
INT32     d;

    PROCNAME("pixBlockconv");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (wc <= 0 || hc <= 0) {
	WARNING("convolution kernel too small", procName);
	return pixClone(pixs);
    }

    d = pixGetDepth(pixs);
    if (d == 8)
	return pixBlockconvGray(pixs, NULL, wc, hc);
    else if (d == 32)
	return pixBlockconvColor(pixs, wc, hc);
    else
	return (PIX *)ERROR_PTR("d not 8 or 32 bpp", procName, NULL);
}


/*----------------------------------------------------------------------*
 *                        Color block convolution                       *
 *----------------------------------------------------------------------*/
/*!
 *  pixBlockconvColor()
 *
 *          Input:  pixs (32 bpp; 24 bpp RGB color)
 *                  wc, hc   (half width/height of convolution kernel)
 *          Return: pixd, or null on error
 *
 *          Note:  The full width and height of the convolution kernel
 *                  are (2 * wc + 1) and (2 * hc + 1).
 */
PIX  *
pixBlockconvColor(PIX    *pixs,
                  INT32   wc,
		  INT32   hc)
{
PIX      *pixRed, *pixGreen, *pixBlue;
PIX      *pixRedConv, *pixGreenConv, *pixBlueConv;
PIX      *pixd;

    PROCNAME("pixBlockconvColor");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (wc <= 0 || hc <= 0) {
	WARNING("convolution kernel too small", procName);
	return pixClone(pixs);
    }

    if (pixGetDepth(pixs) != 32)
	return (PIX *)ERROR_PTR("pix not 32 bpp", procName, NULL);

    pixRed = pixGetRGBComponent(pixs, COLOR_RED);
    pixRedConv = pixBlockconvGray(pixRed, NULL, wc, hc);
    pixDestroy(&pixRed);
    pixGreen = pixGetRGBComponent(pixs, COLOR_GREEN);
    pixGreenConv = pixBlockconvGray(pixGreen, NULL, wc, hc);
    pixDestroy(&pixGreen);
    pixBlue = pixGetRGBComponent(pixs, COLOR_BLUE);
    pixBlueConv = pixBlockconvGray(pixBlue, NULL, wc, hc);
    pixDestroy(&pixBlue);

    if ((pixd = pixCreateRGBImage(pixRedConv, pixGreenConv, pixBlueConv))
	    == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    pixDestroy(&pixRedConv);
    pixDestroy(&pixGreenConv);
    pixDestroy(&pixBlueConv);

    return pixd;
}



/*----------------------------------------------------------------------*
 *                     Grayscale block convolution                      *
 *----------------------------------------------------------------------*/
/*!
 *  pixBlockconvGray()
 *          Input:  pix (8 bpp)
 *                  accum pix (32 bpp; can be null)
 *                  wc, hc   (half width/height of convolution kernel)
 *          Return: pix (8 bpp)
 *
 *          Notes:
 *             (1)  if accum pix is null, make one and destroy it
 *                  before returning.  Otherwise, just use the
 *                  input accum pix.
 *             (2)  The full width and height of the convolution kernel
 *                  are (2 * wc + 1) and (2 * hc + 1).
 */
PIX *
pixBlockconvGray(PIX    *pixs,
	         PIX    *pixa,
	         INT32   wc,
	         INT32   hc)
{
INT32      w, h, wpl, wpla;
UINT32    *datad, *dataa;
PIX       *pixd, *pixt;

    PROCNAME("pixBlockconvGray");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 8)
	return (PIX *)ERROR_PTR("pixs not 8 bpp", procName, NULL);

    if (pixa) {
	if (pixGetDepth(pixa) != 32)
	    return (PIX *)ERROR_PTR("pra not 32 bpp", procName, NULL);
	pixt = pixClone(pixa);
    }
    else {
	if ((pixt = pixBlockconvAccum(pixs)) == NULL)
	    return (PIX *)ERROR_PTR("pixt not made", procName, NULL);
    }
	
    if ((pixd = pixCreateTemplate(pixs)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    
    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    wpl = pixGetWpl(pixs);
    wpla = pixGetWpl(pixt);
    datad = pixGetData(pixd);
    dataa = pixGetData(pixt);

    blockconvLow(datad, w, h, wpl, dataa, wpla, wc, hc);

    pixDestroy(&pixt);
    return pixd;
}


/*!
 *   pixBlockconvAccum()
 *
 *       Input:  pixs (1 or 8 bpp)
 *       Return: accum pix (32 bpp), or null on error.
 *
 *       Action:  The general recursion relation is
 *                   a(i,j) = v(i,j) + a(i-1, j) + a(i, j-1) - a(i-1, j-1)
 *
 *                For the first line, this reduces to the special case
 *                   a(i,j) = v(i,j) + a(i, j-1)
 *                For the first column, the special case is
 *                   a(i,j) = v(i,j) + a(i-1, j)
 */
PIX *
pixBlockconvAccum(PIX    *pixs)
{
INT32      w, h, d, wpls, wpld;
UINT32    *datas, *datad;
PIX       *pixd;

    PROCNAME("pixBlockconvAccum");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);

    d = pixGetDepth(pixs);
    if (d != 1 && d != 8)
	return (PIX *)ERROR_PTR("pixs not 1 or 8 bpp", procName, NULL);

    w = pixGetWidth(pixs); 
    h = pixGetHeight(pixs); 
    if ((pixd = pixCreate(w, h, 32)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    datas = pixGetData(pixs);
    datad = pixGetData(pixd);
    wpls = pixGetWpl(pixs);
    wpld = pixGetWpl(pixd);

    blockconvAccumLow(datad, w, h, wpld, datas, d, wpls);

    return pixd;
}



/*----------------------------------------------------------------------*
 *                        Binary block sum/rank                         *
 *----------------------------------------------------------------------*/
/*!
 *  pixBlockrank()
 *          Input:  pix (1 bpp)
 *                  accum pix (32 bpp)  -- can be null
 *                  wc, hc   (half width/height of block sum/rank kernel)
 *                  rank   (between 0.0 and 1.0; 0.5 is median filter)
 *          Return: pix (1 bpp)
 *
 *          Notes:
 *             (1)  if accum pix is NULL, make one and destroy it
 *                  before returning.  Otherwise, just use the input pix.
 *             (2)  The full width and height of the convolution kernel
 *                  are (2 * wc + 1) and (2 * hc + 1).
 */
PIX *
pixBlockrank(PIX     *pixs,
	     PIX     *pixa,
	     INT32    wc,
	     INT32    hc,
	     FLOAT32  rank)
{
INT32     w, h, wplb, wpla, wpld;
INT32     i, j;
UINT8     thresh, val;
UINT32   *datad, *dataa, *datab, *lineb, *lined;
PIX      *pixd, *pixt1, *pixt2;

    PROCNAME("pixBlockrank");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 1)
	return (PIX *)ERROR_PTR("pixs not 1 bpp", procName, NULL);
    if (rank < 0.0 || rank > 1.0)
	return (PIX *)ERROR_PTR("rank must be in [0.0, 1.0]", procName, NULL);

    if (pixa) {
	if (pixGetDepth(pixa) != 32)
	    return (PIX *)ERROR_PTR("pixa not 32 bpp", procName, NULL);
	pixt1 = pixClone(pixa);
    }
    else {
	if ((pixt1 = pixBlockconvAccum(pixs)) == NULL)
	    return (PIX *)ERROR_PTR("pixt1 not made", procName, NULL);
    }
	
	/* 8 bpp block sum output */
    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    if ((pixt2 = pixCreate(w, h, 8)) == NULL)
	return (PIX *)ERROR_PTR("pixt2 not made", procName, NULL);

    wplb = pixGetWpl(pixt2);
    wpla = pixGetWpl(pixt1);
    datab = pixGetData(pixt2);
    dataa = pixGetData(pixt1);
    blocksumLow(datab, w, h, wplb, dataa, wpla, wc, hc);

	/* 1 bpp block rank filter output */
    if ((pixd = pixCreateTemplate(pixs)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    datad = pixGetData(pixd);
    lined = datad;
    lineb = datab;
    wpld = pixGetWpl(pixd);
    
    thresh = (INT32)(rank * (FLOAT32)255);
    for (i = 0; i < h; i++) {
	for (j = 0; j < w; j++) {
	    val = GET_DATA_BYTE(lineb, j);
	    if (val >= thresh)
		SET_DATA_BIT(lined, j);
	}
	lineb += wplb;
	lined += wpld;
    }

    pixDestroy(&pixt1);
    pixDestroy(&pixt2);
    return pixd;
}


/*!
 *  pixBlocksum()
 *          Input:  pix (1 bpp)
 *                  accum pix (32 bpp)  -- can be null
 *                  wc, hc   (half width/height of block sum/rank kernel)
 *          Return: pix (8 bpp)
 *
 *          Notes:
 *             (1)  if accum pix is NULL, make one and destroy it
 *                  before returning.  Otherwise, just use the input pix.
 *             (2)  The full width and height of the convolution kernel
 *                  are (2 * wc + 1) and (2 * hc + 1).
 */
PIX *
pixBlocksum(PIX     *pixs,
	    PIX     *pixa,
	    INT32    wc,
	    INT32    hc)
{
INT32     w, h, wplb, wpla;
UINT32   *dataa, *datab;
PIX      *pixt, *pixb;

    PROCNAME("pixBlocksum");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 1)
	return (PIX *)ERROR_PTR("pixs not 1 bpp", procName, NULL);

    if (pixa) {
	if (pixGetDepth(pixa) != 32)
	    return (PIX *)ERROR_PTR("pixa not 32 bpp", procName, NULL);
	pixt = pixClone(pixa);
    }
    else {
	if ((pixt = pixBlockconvAccum(pixs)) == NULL)
	    return (PIX *)ERROR_PTR("pixt not made", procName, NULL);
    }
	
	/* 8 bpp block sum output */
    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    if ((pixb = pixCreate(w, h, 8)) == NULL)
	return (PIX *)ERROR_PTR("pixb not made", procName, NULL);

    wplb = pixGetWpl(pixb);
    wpla = pixGetWpl(pixt);
    datab = pixGetData(pixb);
    dataa = pixGetData(pixt);
    blocksumLow(datab, w, h, wplb, dataa, wpla, wc, hc);

    pixDestroy(&pixt);
    return pixb;
}


