/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 * jbig2.c
 *
 *     These are the classification and utility functions for simple
 *     implementations of jbig2:
 *
 *     JBIG2 application-level functions
 *
 *         INT32     jbig2rankhaus()      [rank hausdorff encoder]
 *         INT32     jbig2correl()        [correlation encoder]
 *         PIX      *jbig2errors()        [compares encoding with ground truth]
 *         INT32     jbig2substitutions() [merges classes to improve encoding]
 *
 *     Strict hausdorff classifier (rank = 1.0)
 *
 *         INT32     jbig2ClassifyHaus()
 *         INT32     pixHaustest()
 *
 *     Rank hausdorff classifier
 *
 *         INT32     jbig2ClassifyRankHaus()
 *         INT32     pixRankHaustest()
 *
 *     Correlation classifier
 *
 *         INT32     jbig2ClassifyCorrel()
 *         FLOAT32   pixCorrelationScore()
 *
 *     JBIG2 utility functions
 *
 *         PTA      *jbig2GetULCorners()
 *         JB2DATA  *jbig2CreateData()
 *         void      jbig2DestroyData()
 *         INT32     jbig2WriteData()
 *         PIX      *jbig2DecodeData()
 *         JB2DATA  *jbig2ReadData()
 *
 *
 *     Note: this is NOT an implementation of the JPEG jbig2
 *     proposed standard encoder, the specifications for which
 *     can be found at http://www.jpeg.org/jbigpt2.html.
 *     It is an implementation of the lower-level part of an encoder that:
 *
 *        (1) identifies connected components (either 4 or 8 connected),
 *        (2) puts them in similarity classes (this is an unsupervised
 *            classifier), and
 *        (3) stores the result in a simple file format (2 files,
 *            one for templates and one for coordinate/template-index
 *            triplets).
 *
 *     An implementation of the official jbig2 encoder would start with
 *     parts (1) and (2), and would then compress the triplets
 *     according to the standards requirements (e.g., Huffman coding
 *     coordinate differences) and compress the image templates by
 *     some method (I believe they suggest but do not require
 *     arithmetic coding). 
 *
 *     The low-level part of the encoder provided here has the
 *     following useful features:
 *
 *         - it is accurate in the identification of templates
 *           and classes because it uses a windowed hausdorff
 *           distance metric
 *         - it is accurate in the placement of the connected
 *           components, because it aligns the centroids of the
 *           template with those of each instance
 *         - it is fast because it uses a morphologically based
 *           matching algorithm to implement the hausdorff criterion
 *
 *     We actually provide two different classifiers, one using
 *     Hausdorff distance and one using a simple image correlation.
 *     The Hausdorff method works better because it gives a
 *     small effective weight to foreground boundary pixels and
 *     a large effective weight to foreground pixels that are not
 *     on the boundary, whereas the correlation method weights all
 *     pixels equally.  The Hausdorff weighting corresponds better
 *     to the expected probabilities of the pixel values in a
 *     scanned image.
 *
 *     The Hausdorff scoring is not a rank weighting, because a
 *     single pixel beyond the given distance will cause a match
 *     failure.  A rank Hausdorff is more robust to non-boundary noise,
 *     but it is also more susceptible to confusing components that
 *     should be in different classes.  For implementing a jbig2
 *     application for visually lossless binary image compression,
 *     it is recommended that you use a 3x3 structuring element
 *     (size = 3) and a strict hausdorff comparison (rank = 1.0
 *     in the rank hausdorff function).  This will result in some
 *     excess splitting of character classes, and associated
 *     storage for the extra templates, but that is a small price
 *     to pay for avoiding the serious mistake of putting different
 *     characters into the same class!
 *
 *     The core of several application level functions are also given here.
 *     jbig2rankhaus() does rank hausdorff classification and generates
 *         data and template files
 *     jbig2correl() does correlation classification and generates data and
 *         template files
 *     jbig2errors() finds the errors between two segmentations:
 *         a "ground truth" and a test version
 *     jbig2substitutions() substitutes classes specified in an input
 *         file.  This is useful for quickly making a "ground
 *         truth" jbig2 segmentation from an approximate one.
 */

#include <stdio.h>
#include <stdlib.h>

#include "allheaders.h"

#ifndef  NO_CONSOLE_IO
#define   DEBUG_JBIG2_HAUS   0
#define   DEBUG_JBIG2_CORREL 0
#define   DEBUG_ERRORS       0
#define   DEBUG_SUBS         0
#endif  /* ~NO_CONSOLE_IO */


#define   BUF_SIZE         512

    /* for jbig2ClassifyHaus(): size of border added around
     * pix of each c.c., to allow further processing  */
#define   ADDED_PIXELS       6   /* should be at least the sum of
                                  * the MAX_DIFF_HEIGHT (or MAX_DIFF_WIDTH)
				  * and one-half the size of the SE */

    /* for pixHaustest(); choose these to be 2 or greater */
#define   MAX_DIFF_HEIGHT    2   /* use at least 2 */
#define   MAX_DIFF_WIDTH     2   /* use at least 2 */

    /* for jbig2haus() and jbig2correl():  the lattice cell is
     * used to store the template images, and it must be large
     * enough to contain thelargest components (e.g., characters)
     * that you want to save.  Components larger in either
     * dimension are discarded. */
#define   LATTICE_CELL_W   120   /* output grid width spacing */
#define   LATTICE_CELL_H   120   /* output grid height spacing */

    /* for jbig2haus() and jbig2correl():  component connectivity;
     * must be either 4 or 8.  You get fewer touching characters with
     * 4 than with 8, but you get a smaller number of classes and
     * better results for small blobs with 8 rather than 4.  */
#define   CONNECTIVITY       8

    /* for jbig2errors() */
#define   MAX_SLOP_SIZE      3    /* 3 seems to cover them all */



/*----------------------------------------------------------------------*
 *                 jbig2 application-level functions                    *
 *----------------------------------------------------------------------*/
/*!
 *    jbig2rankhaus()
 *
 *	 Input:  filein
 *               size  (of square structuring element; 3, representing
 *                      3x3 sel, is best for accuracy without
 *                      undue class expansion)
 *               rank (rank val of match, each way)
 *               rootout  (root name of two output files, for templates
 *                         and location and labeling of c.c. data)
 *       Return: 0 if OK; 1 on error
 *
 *       Note: all c.c. larger than LATTICE_CELL_W (or H) are not saved
 *
 */
INT32
jbig2rankhaus(char     *filein,
              INT32     size,
              FLOAT32   rank,
	      char     *rootout)
{
INT32           w, h, i, ncc, nclass;
BOXA           *boxa;
JB2DATA        *jb2data;
NUMA           *naclass;
PIX            *pixs, *pix, *pixd;
PIXA           *pixas, *pixa, *pixac;
PIXAA          *pixaa; /* 2-level: all instances in each class */
PTA            *pta;   /* centroids of bordered images, rel. to UL corner */
PTA            *ptat;  /* centroids of bordered templates */
PTA            *ptaout;  /* UL corners (adjusted) of unbordered images */

    PROCNAME("jbig2rankhaus");

    if (!filein)
	return ERROR_INT("filein not defined", procName, 1);
    if (!rootout)
	return ERROR_INT("rootout not defined", procName, 1);
    if (size < 1 || size > 10)
	return ERROR_INT("size not reasonable", procName, 1);
    if (rank < 0.5 || rank > 1.0)
	return ERROR_INT("rank not reasonable", procName, 1);

    if ((pixs = pixRead(filein)) == NULL)
	return ERROR_INT("pixs not made", procName, 1);
    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);

	/* get the cc's and their bounding boxes */
    if ((boxa = pixConnComp(pixs, &pixas, CONNECTIVITY)) == NULL)
	return ERROR_INT("boxa not made", procName, 1);

	/* get classes using hausdorff with given size structuring
	 * element and specified rank  */
    if (rank == 1.0) {
	if (jbig2ClassifyHaus(boxa, pixas, size, &pixaa, &pta, &ptat, &naclass))
	    return ERROR_INT("error in haus classifier", procName, 1);
    }
    else {
	if (jbig2ClassifyRankHaus(boxa, pixas, size, rank,
				  &pixaa, &pta, &ptat, &naclass))
	    return ERROR_INT("error in rank haus classifier", procName, 1);
    }

	/* make a ptaout, which gives the global UL corners adjusted for
	 * each component so that it can be replaced by the template
	 * for its class and still have the centroid in the
	 * same place */
    if ((ptaout = jbig2GetULCorners(boxa, naclass, pta, ptat)) == NULL)
	return ERROR_INT("ptatout not made", procName, 1);

	/* make the class template image file, using the first image
	 * of each pixa in the pixaa.  We use a cutoff
	 * on the size of displayed templates; those larger than 
	 * the lattice cell size (e.g., 120x120 pixels) are left blank
	 * in the template image.  Thus, they cannot be reconstructed
	 * from the resulting image of templates. */
    nclass = pixaaGetCount(pixaa);
    fprintf(stderr, "%d classes with dilation %d\n", nclass, size);
    ncc = ptaGetCount(pta);
    if ((pixac = pixaCreate(nclass)) == NULL)
	return ERROR_INT("pixac not made", procName, 1);
    for (i = 0; i < nclass; i++) {
	pixa = pixaaGetPixa(pixaa, i, NEWREF);
	pix = pixaGetPix(pixa, 0, CLONE);
	pixaDestroy(&pixa);
	pixaAddPix(pixac, pix, INSERT);
#if  DEBUG_JBIG2_HAUS
	fprintf(stderr, "number in class[%d] = %d\n",
	        i, pixaGetCount(pixaa->pixa[i]));
#endif  /* DEBUG_JBIG2_HAUS */
    }
    if ((pixd = pixaDisplayOnLattice(pixac, LATTICE_CELL_W, LATTICE_CELL_H))
		    == NULL)
	return ERROR_INT("pixd not made", procName, 1);
    pixaDestroy(&pixac);

	/* make a memory data structure and write out data files */
    if ((jb2data = jbig2CreateData(pixd, w, h, nclass, LATTICE_CELL_W,
                              LATTICE_CELL_H, naclass, ptaout)) == NULL)
	return ERROR_INT("jb2data not made", procName, 1);
    jbig2WriteData(rootout, jb2data);

#if  DEBUG_JBIG2_HAUS
{
char    buf[BUF_SIZE];
char   *templatefile, *datafile;
PIX    *pixt;

	/* read data file and reconstruct image */
    sprintf(buf, "%s%s", rootout, JBIG2_TEMPLATE_EXT); 
    templatefile = stringNew(buf);
    sprintf(buf, "%s%s", rootout, JBIG2_DATA_EXT); 
    datafile = stringNew(buf);
    pixt = jbig2DecodeData(templatefile, datafile, NULL);
    pixWrite("debug_output_recon1", pixt, IFF_PNG);
    FREE((void *)templatefile);
    FREE((void *)datafile);
    pixDestroy(&pixt);
}
#endif  /* DEBUG_JBIG2_HAUS */

    jbig2DestroyData(&jb2data);
    pixaDestroy(&pixas);
    ptaDestroy(&pta);
    ptaDestroy(&ptat);
    ptaDestroy(&ptaout);
    boxaDestroy(&boxa);
    pixDestroy(&pixs);
    pixDestroy(&pixd);
    pixaaDestroy(&pixaa);
    numaDestroy(&naclass);

    return 0;
}


/*!
 *    jbig2correl()
 *
 *	 Input:  filein
 *               thresh  (correlation threshold; 0.80 is reasonable for
 *                        fair accuracy and not too many classes)
 *               rootout  (root name of two output files, for templates
 *                         and location and labeling of c.c. data)
 *       Return: 0 if OK; 1 on error
 *
 *       Note: all c.c. larger than LATTICE_CELL_W (or H) are not saved
 *
 */
INT32
jbig2correl(char    *filein,
            FLOAT32  thresh,
	    char    *rootout)
{
INT32           w, h, i, ncc, nclass;
BOXA           *boxa;
JB2DATA        *jb2data;
NUMA           *naclass;
PIX            *pixs, *pix, *pixd;
PIXA           *pixas, *pixa, *pixac;
PIXAA          *pixaa; /* all instances in each class */
PTA            *pta;   /* centroids of bordered images */
PTA            *ptat;  /* centroids of bordered templates */
PTA            *ptaout;  /* UL corners (adjusted) of unbordered images */

    PROCNAME("jbig2correl");

    if (!filein)
	return ERROR_INT("filein not defined", procName, 1);
    if (!rootout)
	return ERROR_INT("rootout not defined", procName, 1);

    if ((pixs = pixRead(filein)) == NULL)
	return ERROR_INT("pixs not made", procName, 1);
    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);

	/* get the cc's and their bounding boxes */
    if ((boxa = pixConnComp(pixs, &pixas, CONNECTIVITY)) == NULL)
	return ERROR_INT("boxa not defined", procName, 1);

	/* get classes using correlation with threshold */
    if (jbig2ClassifyCorrel(boxa, pixas, thresh, &pixaa, &pta, &ptat, &naclass))
	return ERROR_INT("error in correl classifier", procName, 1);

	/* we make a ptaout, which gives the UL corners adjusted for
	 * each component so that it can be replaced by the template
	 * for its class and still have the centroid in the
	 * same place */
    if ((ptaout = jbig2GetULCorners(boxa, naclass, pta, ptat)) == NULL)
	return ERROR_INT("ptaout not made", procName, 1);
    ncc = ptaGetCount(pta);
	
	/* make the class template image file, using the first image
	 * of each pixa in the pixaa.  We use a cutoff
	 * on the size of displayed templates; those larger than
	 * the lattice cell size (e.g., 120x120 pixels) are left blank
	 * in the template image.  Thus, they cannot be reconstructed
	 * from the resulting image of templates. */
    nclass = pixaaGetCount(pixaa);
    fprintf(stderr, "%d classes with threshold %4.2f\n", nclass, thresh);
    if ((pixac = pixaCreate(nclass)) == NULL)
	return ERROR_INT("pixac not made", procName, 1);
    for (i = 0; i < nclass; i++) {
	pixa = pixaaGetPixa(pixaa, i, NEWREF);
	pix = pixaGetPix(pixa, 0, CLONE);
	pixaDestroy(&pixa);
	pixaAddPix(pixac, pix, INSERT);
#if  DEBUG_JBIG2_CORREL
	fprintf(stderr, "number in class[%d] = %d\n",
	        i, pixaGetCount(pixaa->pixa[i]));
#endif  /* DEBUG_JBIG2_CORREL */
    }
    if ((pixd = pixaDisplayOnLattice(pixac, LATTICE_CELL_W, LATTICE_CELL_H))
		    == NULL)
	return ERROR_INT("pixd not made", procName, 1);
    pixaDestroy(&pixac);

        /* make a memory data structure and write out data files */
    if ((jb2data = jbig2CreateData(pixd, w, h, nclass, LATTICE_CELL_W,
			           LATTICE_CELL_H, naclass, ptaout)) == NULL)
	return ERROR_INT("jb2data not made", procName, 1);
    jbig2WriteData(rootout, jb2data);

#if  DEBUG_JBIG2_CORREL
{
char    buf[BUF_SIZE];
char   *templatefile, *datafile;
PIX    *pixt;

	/* read data file and reconstruct image */
    sprintf(buf, "%s%s", rootout, JBIG2_TEMPLATE_EXT);
    templatefile = stringNew(buf);
    sprintf(buf, "%s%s", rootout, JBIG2_DATA_EXT);
    datafile = stringNew(buf);
    pixt = jbig2DecodeData(templatefile, datafile, NULL);
    pixWrite("output_recon1", pixt, IFF_PNG);
    FREE((void *)templatefile);
    FREE((void *)datafile);
    pixDestroy(&pixt);
}
#endif  /* DEBUG_JBIG2_CORREL */

    jbig2DestroyData(&jb2data);
    pixaDestroy(&pixas);
    ptaDestroy(&pta);
    ptaDestroy(&ptat);
    ptaDestroy(&ptaout);
    boxaDestroy(&boxa);
    pixDestroy(&pixs);
    pixDestroy(&pixd);
    pixaaDestroy(&pixaa);
    numaDestroy(&naclass);

    return 0;
}


/*!
 *    jbig2errors()
 *
 *	 Input:   gtroot    (root name of "ground truth" jbig2 files)
 *	          testroot  (root name of test jbig2 files)
 *                &nerrors  (<return> number of errors in test)
 *       Return:  pixd (for output substitution image), or null on error
 *
 *       Here's the basic idea.  The ground truth (gt) assigns
 *       all instances to some class.  All instances in a class
 *       are visually identical, in that they are observed to represent
 *       the same character in the same type style and size. 
 *       If another classifier splits one of these classes, it is less
 *       efficient but is not considered to be an error, because
 *       the reconstructed page will not be noticeably different.
 *       However, as each gt class is assumed to be visually distinct,
 *       if another classifier combines instances from two different
 *       gt classes into one class, that is an error.  
 *
 *       So we consider each class in the test set.  The instances
 *       in that class correspond to a set of classes in the gt set.
 *       If that set consists of just one class, there are no errors.
 *       But if the instances in the test set consist of more than
 *       one class in the gt set, errors are made in the test set.
 *       The test set uses the first instance in the class as the
 *       representative (template).  This corresponds to some class
 *       in the gt set.  All other classes in the gt set that corresponds
 *       to the test set are errors.  Sum over all test set classes
 *       for the total number of errors.
 */
PIX *
jbig2errors(char    *gtroot,
            char    *testroot,
	    INT32   *pnerrors)
{
char            buf1[BUF_SIZE], buf2[BUF_SIZE];
INT32           foundit, nmissed, nerrors, slop;
INT32           i, j, k, jclass, kclass, gtclass, x1, y1, x2, y2;
INT32           nclassgt, nclasstest, nccgt, ncctest;
INT32           nfoundtest, w, h;
JB2DATA        *gtdata, *testdata;
NUMA           *naclassgt, *naclasstest;
PIX            *pix, *pixd;
PIXA           *pixa;
PTA            *ptagt, *ptatest;

    PROCNAME("jbig2errors");

    if (!gtroot)
	return (PIX *)ERROR_PTR("gtroot not defined", procName, NULL);
    if (!testroot)
	return (PIX *)ERROR_PTR("testroot not defined", procName, NULL);

	/* read gt data */
    sprintf(buf1, "%s%s", gtroot, JBIG2_TEMPLATE_EXT);
    sprintf(buf2, "%s%s", gtroot, JBIG2_DATA_EXT);
    if ((gtdata = jbig2ReadData(buf1, buf2)) == NULL)
	return (PIX *)ERROR_PTR("gtdata not made", procName, NULL);

        /* read test data; errorful reconstruction; decode pix from test */
    sprintf(buf1, "%s%s", testroot, JBIG2_TEMPLATE_EXT);
    sprintf(buf2, "%s%s", testroot, JBIG2_DATA_EXT);
    if ((pixd = jbig2DecodeData(buf1, buf2, &testdata)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    ptagt = gtdata->pta;   /* locations of gt components */
    naclassgt = gtdata->naclass;  /* class ids of gt components */
    nclassgt = gtdata->nclass;   /* number of gt classes */
    nccgt = numaGetCount(naclassgt);  /* number of gt components */

    ptatest = testdata->pta;   /* locations of test components */
    naclasstest = testdata->naclass;  /* class ids of test components */
    nclasstest = testdata->nclass;   /* number of test classes */
    ncctest = numaGetCount(naclasstest);  /* number of test components */

	/* make a pixa of test templates */
    if ((pixa = pixaCreateFromPix(testdata->pix, nclasstest,
                             testdata->latticew, testdata->latticeh)) == NULL)
	return (PIX *)ERROR_PTR("pixa not made", procName, NULL);

    fprintf(stderr,
        "nclassgt = %d, nccgt = %d, nclasstest = %d, ncctest = %d\n",
        nclassgt, nccgt, nclasstest, ncctest);

    nerrors = 0;
    nmissed = 0;   /* want to set MAX_SLOP_SIZE big enough to keep this 0 */
    for (i = 0; i < nclasstest; i++) {
	pix = pixaGetPix(pixa, i, CLONE);
	w = pixGetWidth(pix);
	h = pixGetHeight(pix);
	slop = MIN(MAX_SLOP_SIZE, 1 + (w + h) / 6);
	nfoundtest = 0;
        for (j = 0; j < ncctest; j++) {
	    numaGetIValue(naclasstest, j, &jclass);
	    if (jclass != i)  /* not in i-th test class */
	        continue;
	    nfoundtest++;
	    ptaGetIPt(ptatest, j, &x2, &y2);

		/* search gt for this instance */
	    foundit = 0;
	    for (k = 0; k < nccgt; k++) {
		ptaGetIPt(ptagt, k, &x1, &y1);
		if (x1 >= x2 - slop && x1 <= x2 + slop &&
		    y1 >= y2 - slop && y1 <= y2 + slop)
		{  /* found the c.c. */
		    numaGetIValue(naclassgt, k, &kclass);
		    foundit = 1;
		    break;
		}
	    }
	    if (foundit == 0) {
		fprintf(stderr, "missed one!\n");
/*		fprintf(stderr,
		   "    w = %d, h = %d, i = %d, slop = %d, (x2,y2) = (%d,%d)\n",
		     w, h, i, slop, x2, y2);  */
		nmissed++;
		continue;
	    }

	    if (nfoundtest == 1) {   /* the first one; the template */
		gtclass = kclass;   /* the corresp. gtclass to testclass i */
#if  DEBUG_ERRORS
		fprintf(stderr, "testclass %d --> gtclass %d\n", i, gtclass);
#endif  /* DEBUG_ERRORS */
	    }
	    else {    /* check and record inverted B/W if substitution */
		if (kclass != gtclass) {
		    nerrors++;
#if  DEBUG_ERRORS
		    fprintf(stderr,
			"cc %d at (%d,%d) in testclass %d --> gtclass %d\n",
			nfoundtest, x2, y2, i, kclass);
#endif  /* DEBUG_ERRORS */
		    pixRasterop(pixd, x2, y2, w, h, PIX_NOT(PIX_SRC),
				pix, 0, 0);
		}
	    }
	}
	pixDestroy(&pix);
    }
		
    *pnerrors = nerrors;
    if (nmissed > 0)
	WARNING_INT("Number of components missed = %d\n", procName, nmissed);

    jbig2DestroyData(&gtdata);
    jbig2DestroyData(&testdata);
    pixaDestroy(&pixa);

    return pixd;
}


/*!
 *  jbig2substitutions()
 *
 *       Input: inroot   (root name of input jbig2 files)
 *              subsfile (file of class substitutions)
 *              outroot  (root name of output jbig2 files)
 *       Return:  0 if OK, 1 on error
 *           
 *
 *       The substitutions file contains pairs:
 *             old-class     new-class
 *             ...
 * 
 *       The purpose is to allow one to generate a "ground truth"
 *       from a first approximation.  There is no attempt to get
 *       the alignment perfectly correct in the output data file.
 *
 *       For each pair in the subsfile, the new-class is to be
 *       substituted for the old-class.  We then remove the
 *       discarded old-class templates, and renumber the
 *       classes accordingly.  This also requires using
 *       the renumbered classes in the data file.
 *       To avoid interactions, the renumbering is done in two
 *       steps:  first renumber directly according to the
 *       substitution file; second, eliminate the skipped class
 *       numbers by renumbering them so that they are consecutive.
 */
INT32
jbig2substitutions(char    *inroot,
                   char    *subsfile,
		   char    *outroot)
{
char            buf1[BUF_SIZE], buf2[BUF_SIZE];
char           *str, *linestr;
INT32           latticew, latticeh;
INT32           i, j, nbytes, oldval, newval, classval, foundit, renum;
INT32           index, len, nsaved;
INT32           nclassin, nclassout, ncc, nsubfile, nsubs;
JB2DATA        *indata, *outdata;
NUMA           *naclassin, *naclassout, *naold, *nanew, *napost, *narenumber;
SARRAY         *sa;
PIX            *pixout, *pix, *pixd;
PIXA           *pixa, *pixar;

    PROCNAME("jbig2substitutions");

    if (!inroot)
	return ERROR_INT("inroot not defined", procName, 1);
    if (!subsfile)
	return ERROR_INT("subsfile not defined", procName, 1);
    if (!outroot)
	return ERROR_INT("outroot not defined", procName, 1);

	/* read input jbig2 data */
    sprintf(buf1, "%s%s", inroot, JBIG2_TEMPLATE_EXT);
    sprintf(buf2, "%s%s", inroot, JBIG2_DATA_EXT);
    if ((indata = jbig2ReadData(buf1, buf2)) == NULL)
	return ERROR_INT("indata not made", procName, 1);

    naclassin = indata->naclass;  /* class ids of indata components */
    nclassin = indata->nclass;   /* number of indata classes */
    ncc = numaGetCount(naclassin);  /* number of indata components */
    latticew = indata->latticew;
    latticeh = indata->latticeh;

        /* start with this and change it */
    if ((naclassout = numaCopy(naclassin)) == NULL)
	return ERROR_INT("naclassout not made", procName, 1);
    
	/* make a pixa of indata templates */
    if ((pixa = pixaCreateFromPix(indata->pix, nclassin, latticew, latticeh))
		== NULL)
	return ERROR_INT("pixa not made", procName, 1);

    fprintf(stderr, "nclassin = %d, ncc = %d\n", nclassin, ncc);

	/* read in subs file data */
    if ((str = (char *)fileToArray(subsfile, &nbytes)) == NULL)
	return ERROR_INT("str not read", procName, 1);
    if ((sa = sarrayCreateLinesFromString(str, 0)) == NULL)
	return ERROR_INT("sa not made", procName, 1);
    nsubfile = sarrayGetCount(sa);

	/* parse the subs file */
    if ((naold = numaCreate(nsubfile)) == NULL)
	return ERROR_INT("naold not made", procName, 1);
    if ((nanew = numaCreate(nsubfile)) == NULL)
	return ERROR_INT("nanew not made", procName, 1);
    for (i = 0; i < nsubfile; i++) {
	linestr = sarrayGetString(sa, i, 1);
	if (linestr[0] == '%') {  /* ignore */
	    FREE((void *)linestr);
	    continue;
	}
	len = strlen(linestr);
	for (j = 0; j < len; j++) {   /* remove extraneous characters */
	    if (linestr[j] != '0' && linestr[j] != '1' &&
		linestr[j] != '2' && linestr[j] != '3' &&
		linestr[j] != '4' && linestr[j] != '5' &&
		linestr[j] != '6' && linestr[j] != '7' &&
		linestr[j] != '8' && linestr[j] != '9')
		    linestr[j] = ' ';
	}
	if (sscanf(linestr, "%d %d", &oldval, &newval) == 2) {
	    numaAddNumber(naold, (FLOAT32)oldval);
	    numaAddNumber(nanew, (FLOAT32)newval);
	}
	else
	    fprintf(stderr, " data line %d, data: '%s' not read properly\n",
		    i, linestr);
	FREE((void *)linestr);
    }
    nsubs = numaGetCount(naold);
    nclassout = nclassin - nsubs;

	/* do the class value substitutions, without renumbering */
    for (i = 0; i < nsubs; i++) {
	numaGetIValue(naold, i, &oldval);
	numaGetIValue(nanew, i, &newval);
	for (j = 0; j < ncc; j++) {
	    numaGetIValue(naclassout, j, &classval);
	    if (classval == oldval)
		numaSetValue(naclassout, j, newval);
	}
    }

	/* get an numa of the remaining classes,
	 * and a renumbering array relating the new number to the old */
    napost = numaCreate(nclassin);   /* hold remaining classes */
    narenumber = numaCreate(nclassin);  /* has na[i] = new-num[old-num] */
    nsaved = 0;
    for (i = 0; i < nclassin; i++) {
	foundit = 0;
	for (j = 0; j < nsubs; j++) {
	    numaGetIValue(naold, j, &oldval);
	    if (i == oldval) {
		foundit = 1;
		break;
	    }
	}
	if (foundit == 0) {
	    nsaved++;
	    numaAddNumber(napost, i);
	}
	numaAddNumber(narenumber, nsaved - 1);
    }
    if (nclassout != numaGetCount(napost))
	fprintf(stderr, "oops: nclassout = %d, size of napost = %d\n", 
	        nclassout, numaGetCount(napost));

	/* Use the numa of remaining classes to compress the
	 * template array so that it only has the remaining classes.
	 * This causes a renumbering of the classes.  */
    if ((pixar = pixaCreate(nclassin)) == NULL)
	return ERROR_INT("pixar not made", procName, 1);
    for (i = 0; i < nclassout; i++) {
	numaGetIValue(napost, i, &index);
	pix = pixaGetPix(pixa, index, COPY);
	pixaAddPix(pixar, pix, INSERT);
    }

	/* Renumber the class values in the new instance list,
	 * to agree with the numbers of remaining templates. */
    for (i = 0; i < ncc; i++) {
	numaGetIValue(naclassout, i, &index);
	numaGetIValue(narenumber, index, &renum);
	numaSetValue(naclassout, i, renum);
    }

	/* Generate the output data and write it to file */
    if ((pixout = pixaDisplayOnLattice(pixar, latticew, latticeh)) == NULL)
	return ERROR_INT("pixout not made", procName, 1);
    if ((outdata = jbig2CreateData(pixout, indata->w, indata->h,
                                   nclassout, latticew, latticeh,
                                   naclassout, indata->pta)) == NULL)
	return ERROR_INT("outdata not made", procName, 1);
    jbig2WriteData(outroot, outdata);

    fprintf(stderr, "nclassout = %d, ncc = %d\n", nclassout, ncc);

#if  DEBUG_SUBS
	/* Test the output results visually by decoding */
    sprintf(buf1, "%s%s", outroot, JBIG2_TEMPLATE_EXT);
    sprintf(buf2, "%s%s", outroot, JBIG2_DATA_EXT);
    pixd = jbig2DecodeData(buf1, buf2, NULL);
    pixWrite("junksubs", pixd, IFF_PNG);
    pixDestroy(&pixd);
#endif  /* DEBUG_SUBS */

    FREE((void *)str);
    numaDestroy(&naclassout);
    numaDestroy(&naold);
    numaDestroy(&nanew);
    numaDestroy(&napost);
    numaDestroy(&narenumber);
    sarrayDestroy(&sa);
    pixDestroy(&pixout);
    pixaDestroy(&pixa);
    pixaDestroy(&pixar);
    jbig2DestroyData(&indata);
    jbig2DestroyData(&outdata);

    exit(0);
}



/*----------------------------------------------------------------------*
 *       Classification using windowed rank = 1.0 hausdorff metric      *
 *----------------------------------------------------------------------*/
/*!
 *  jbig2ClassifyHaus()
 *
 *      Input:  boxa (of c.c.)
 *              pixas (of c.c.)
 *              size (size of square structuring element)
 *              &pixaa (<return> 2-level array of instances for each class)
 *              &pta (<return> centroids of all bordered c.c.)
 *              &ptat (<return> centroids of all bordered template c.c.)
 *              &naclass (<return> class label of all c.c.)
 *      Return: 0 if OK; 1 on error
 */
INT32
jbig2ClassifyHaus(BOXA     *boxa,
		  PIXA     *pixas,
		  INT32     size,
		  PIXAA   **ppixaa,
		  PTA     **ppta,
		  PTA     **pptat,
		  NUMA    **pnaclass)
{
INT32          n, n3, i, j, index, found, testval;
FLOAT32        x1, y1, x2, y2;
BOX           *box;
NUMA          *naclass;
PIX           *pix, *pix1, *pix2, *pix3, *pix4;
PIXA          *pixa, *pixa1, *pixa2, *pixa3, *pixa4;
PIXAA         *pixaa;
PTA           *pta, *ptat;
SEL           *sel;
SELA          *sela;

    PROCNAME("jbig2ClassifyHaus");

    if (!boxa)
	return ERROR_INT("boxa not found", procName, 1);
    if (!pixas)
	return ERROR_INT("pixas not found", procName, 1);
    if (size < 2 || size > 9)
	return ERROR_INT("size not in {2, ..., 9}", procName, 1);
    if (!ppta)
	return ERROR_INT("&pta not defined", procName, 1);
    if (!pptat)
	return ERROR_INT("&ptat not defined", procName, 1);
    if (!pnaclass)
	return ERROR_INT("&pnaclass not defined", procName, 1);

    sela = selaAddBasic(NULL);
    if (size == 2)
	selaFindSelByName(sela, "sel_2", &index, &sel);
    else if (size == 3)
	selaFindSelByName(sela, "sel_3", &index, &sel);
    else if (size == 4)
	selaFindSelByName(sela, "sel_4", &index, &sel);
    else if (size == 5)
	selaFindSelByName(sela, "sel_5", &index, &sel);
    else if (size == 6)
	selaFindSelByName(sela, "sel_6", &index, &sel);
    else if (size == 7)
	selaFindSelByName(sela, "sel_7", &index, &sel);
    else if (size == 8)
	selaFindSelByName(sela, "sel_8", &index, &sel);
    else  /* size == 9 */
	selaFindSelByName(sela, "sel_9", &index, &sel);
/*    selWriteStream(stderr, sel); */

	/* generate the bordered pixa, with and without dilation.
	 * pixa1 and pixa2 contain all the input c.c. */
    n = pixaGetCount(pixas);
    pixa1 = pixaCreate(n);
    pixa2 = pixaCreate(n);
    for (i = 0; i < n; i++) {
	pix = pixaGetPix(pixas, i, CLONE);
	pix1 = pixAddBorderGeneral(pix, ADDED_PIXELS, ADDED_PIXELS,
	            ADDED_PIXELS, ADDED_PIXELS, 0);
	pix2 = pixDilate(NULL, pix1, sel);
	pixaAddPix(pixa1, pix1, INSERT);   /* un-dilated */
	pixaAddPix(pixa2, pix2, INSERT);   /* dilated */
	pixDestroy(&pix);
    }

	/* Get the centroids of all the bordered images.
	 * These are relative to the UL corner of each (bordered) pix.  */
    pta = pixaCentroids(pixa1);
    *ppta = pta;
    ptat = ptaCreate(0);  /* this will hold centroids of templates */
    *pptat = ptat;
	
	/* save the class of each c.c. */
    naclass = numaCreate(0);
    *pnaclass = naclass;

	/* Make a pixaa to store un-bordered pix, in a hierarchical
	 * set of arrays.  There is one pixa for each class,
	 * and the pix in each pixa are all the instances found
	 * of that class.  This is actually more than we need
	 * for the jbig2 encoder, but it is pedagogically useful to
	 * have these images around, because we can see when the
	 * classifier fails.  Also, the set of instances
	 * for each class can be used to make an improved (or even
	 * grayscale) template, rather than simply using the first
	 * one in the set. */
    pixaa = pixaaCreate(0);
    *ppixaa = pixaa;

	/* temporary arrays to store class exemplars (templates) */
    pixa3 = pixaCreate(0);   /* un-dilated */
    pixa4 = pixaCreate(0);   /* dilated */

	/* Fill up the pixaa tree with the template exemplars as
	 * the first pix in each pixa.  As we add each pix,
	 * we also add the associated box to the pixa.
	 * We also keep track of the centroid of each pix,
	 * and use the difference between centroids (of the
	 * pix with the exemplar we are checking it with)
	 * to align the two when checking that the Hausdorff
	 * distance does not exceed a threshold.
	 * The threshold is set by the SE used for dilating.
	 * For example, a 3x3 brick, sel_3, corresponds to a
	 * Hausdorff distance of 1.  In general, for an NxN brick,
	 * with N odd, corresponds to a Hausdorff distance of (N - 1)/2.
	 * We can't reasonably expect to use a smaller size than 3x3.
	 * Use of a SE larger than 3x3 should result in fewer classes,
	 * but also in the likelihood of more errors.
	 * Substitute sel_5 and sel_7 to see what happens
	 * to the number of classes and the apparent errors. */
    for (i = 0; i < n; i++) {
	pix1 = pixaGetPix(pixa1, i, CLONE);
	pix2 = pixaGetPix(pixa2, i, CLONE);
	ptaGetPt(pta, i, &x1, &y1); 
	n3 = pixaGetCount(pixa3);
	found = FALSE;
	for (j = 0; j < n3; j++) {
	    pix3 = pixaGetPix(pixa3, j, CLONE);
	    pix4 = pixaGetPix(pixa4, j, CLONE);
	    ptaGetPt(ptat, j, &x2, &y2);
	    testval = pixHaustest(pix1, pix2, pix3, pix4, x1 - x2, y1 - y2);
	    pixDestroy(&pix3);
	    pixDestroy(&pix4);
	    if (testval == 1) {
		found = TRUE;
		numaAddNumber(naclass, j);
		pixa = pixaaGetPixa(pixaa, j, NEWREF);
		pix = pixaGetPix(pixas, i, CLONE);
		pixaAddPix(pixa, pix, INSERT);
		box = boxaGetBox(boxa, i, CLONE);
		pixaAddBox(pixa, box, INSERT);
		pixaDestroy(&pixa);
		break;
	    }
	}
	if (found == FALSE) {  /* new class */
	    numaAddNumber(naclass, n3);
	    pixa = pixaCreate(0);
	    pix = pixaGetPix(pixas, i, CLONE);
	    pixaAddPix(pixa, pix, INSERT);
	    box = boxaGetBox(boxa, i, CLONE);
	    pixaAddBox(pixa, box, INSERT);
	    pixaaAddPixa(pixaa, pixa, INSERT);
	    ptaAddPt(ptat, x1, y1);
	    pixaAddPix(pixa3, pix1, INSERT);
	    pixaAddPix(pixa4, pix2, INSERT);
	}
	else {   /* don't save them */
	    pixDestroy(&pix1);
	    pixDestroy(&pix2);
	}
    }

    pixaDestroy(&pixa1);
    pixaDestroy(&pixa2);
    pixaDestroy(&pixa3);
    pixaDestroy(&pixa4);
    selaDestroy(&sela);

    return 0;
}


/*!
 *  pixHaustest()
 *
 *        Input:   pix1   (new pix, not dilated)
 *                 pix2   (new pix, dilated)
 *                 pix3   (exemplar pix, not dilated)
 *                 pix4   (exemplar pix, dilated)
 *                 delx   (x comp of centroid difference)
 *                 dely   (y comp of centroid difference)
 *        Return:  0 (FALSE) if no match, 1 (TRUE) if the new
 *                 pix is in the same class as the exemplar.
 *
 *        Note: we check first that the two pix are roughly
 *        the same size.  Only if they meet that criterion do
 *        we compare the bitmaps.  The Hausdorff is a 2-way
 *        check.  The centroid difference is used to align the two
 *        images to the nearest integer for each of the checks.
 *        These check that the dilated image of one contains
 *        ALL the pixels of the undilated image of the other.
 *        Checks are done in both direction.  A single pixel not
 *        contained in either direction results in failure of the test.
 *
 */
INT32
pixHaustest(PIX      *pix1,
            PIX      *pix2,
	    PIX      *pix3,
	    PIX      *pix4,
	    FLOAT32   delx,   /* x(1) - x(3) */
	    FLOAT32   dely)   /* y(1) - y(3) */
{
INT32   wi, hi, wt, ht, delw, delh, idelx, idely, boolval;
PIX    *pixt;

	/* eliminate possible matches based on size difference */
    wi = pixGetWidth(pix1);
    hi = pixGetHeight(pix1);
    wt = pixGetWidth(pix3);
    ht = pixGetHeight(pix3);
    delw = ABS(wi - wt);
    if (delw > MAX_DIFF_WIDTH)
	return FALSE;
    delh = ABS(hi - ht);
    if (delh > MAX_DIFF_HEIGHT)
	return FALSE;

	/* round difference in centroid location to nearest integer;
	 * use this as a shift when doing the matching. */
    if (delx >= 0)
	idelx = (INT32)(delx + 0.5);
    else
	idelx = (INT32)(delx - 0.5);
    if (dely >= 0)
	idely = (INT32)(dely + 0.5);
    else
	idely = (INT32)(dely - 0.5);

	/*  Do 1-direction hausdorff, checking that every pixel in pix1
	 *  is within a dilation distance of some pixel in pix3.  Namely,
	 *  that pix4 entirely covers pix1:
	 *       pixt = pixSubtract(NULL, pix1, pix4), including shift
	 *  where pixt has no ON pixels.  */
    pixt = pixCreateTemplate(pix1);
    pixRasterop(pixt, 0, 0, wi, hi, PIX_SRC, pix1, 0, 0);
    pixRasterop(pixt, idelx, idely, wi, hi, PIX_DST & PIX_NOT(PIX_SRC),
                pix4, 0, 0);
    pixZero(pixt, &boolval);
    if (boolval == 0) {
	pixDestroy(&pixt);
	return FALSE;
    }

	/*  Do 1-direction hausdorff, checking that every pixel in pix3
	 *  is within a dilation distance of some pixel in pix1.  Namely,
	 *  that pix2 entirely covers pix3:
	 *      pixSubtract(pixt, pix3, pix2), including shift
	 *  where pixt has no ON pixels. */
    pixRasterop(pixt, idelx, idely, wt, ht, PIX_SRC, pix3, 0, 0);
    pixRasterop(pixt, 0, 0, wt, ht, PIX_DST & PIX_NOT(PIX_SRC), pix2, 0, 0);
    pixZero(pixt, &boolval);
    pixDestroy(&pixt);
    return boolval;
}



/*----------------------------------------------------------------------*
 *         Classification using windowed rank hausdorff metric          *
 *----------------------------------------------------------------------*/
/*!
 *  jbig2ClassifyRankHaus()
 *
 *      Input:  boxa (of c.c.)
 *              pixas (of c.c.)
 *              size (size of square structuring element)
 *              rank (rank val of match, each way)
 *              &pixaa (<return> 2-level array of instances for each class)
 *              &pta (<return> centroids of all bordered c.c.)
 *              &ptat (<return> centroids of all bordered template c.c.)
 *              &naclass (<return> class label of all c.c.)
 *      Return: 0 if OK; 1 on error
 */
INT32
jbig2ClassifyRankHaus(BOXA     *boxa,
		      PIXA     *pixas,
		      INT32     size,
		      FLOAT32   rank,
		      PIXAA   **ppixaa,
		      PTA     **ppta,
		      PTA     **pptat,
		      NUMA    **pnaclass)
{
INT32          n, n3, i, j, index, found, testval, area1, area3;
INT32         *tab8;
FLOAT32        x1, y1, x2, y2;
BOX           *box;
NUMA          *naclass;
NUMA          *nafg;   /* fg area of all instances */
NUMA          *nafgt;  /* fg area of all templates */
PIX           *pix, *pix1, *pix2, *pix3, *pix4;
PIXA          *pixa, *pixa1, *pixa2, *pixa3, *pixa4;
PIXAA         *pixaa;
PTA           *pta, *ptat;
SEL           *sel;
SELA          *sela;

    PROCNAME("jbig2ClassifyRankHaus");

    if (!boxa)
	return ERROR_INT("boxa not found", procName, 1);
    if (!pixas)
	return ERROR_INT("pixas not found", procName, 1);
    if (size < 1 || size > 9)
	return ERROR_INT("size not in {2, ..., 9}", procName, 1);
    if (rank < 0.3 || rank > 1.0)
	return ERROR_INT("rank not in range 0.3 - 1.0", procName, 1);
    if (!ppta)
	return ERROR_INT("&pta not defined", procName, 1);
    if (!pptat)
	return ERROR_INT("&ptat not defined", procName, 1);
    if (!pnaclass)
	return ERROR_INT("&pnaclass not defined", procName, 1);

    if ((tab8 = makePixelSumTab8()) == NULL)
	return ERROR_INT("tab8 not made", procName, 1);

    sela = selaAddBasic(NULL);
    if (size == 1)
	selaFindSelByName(sela, "sel_1", &index, &sel);
    else if (size == 2)
	selaFindSelByName(sela, "sel_2", &index, &sel);
    else if (size == 3)
	selaFindSelByName(sela, "sel_3", &index, &sel);
    else if (size == 4)
	selaFindSelByName(sela, "sel_4", &index, &sel);
    else if (size == 5)
	selaFindSelByName(sela, "sel_5", &index, &sel);
    else if (size == 6)
	selaFindSelByName(sela, "sel_6", &index, &sel);
    else if (size == 7)
	selaFindSelByName(sela, "sel_7", &index, &sel);
    else if (size == 8)
	selaFindSelByName(sela, "sel_8", &index, &sel);
    else  /* size == 9 */
	selaFindSelByName(sela, "sel_9", &index, &sel);
/*    selWriteStream(stderr, sel); */

	/* generate the bordered pixa, with and without dilation.
	 * pixa1 and pixa2 contain all the input c.c. */
    n = pixaGetCount(pixas);
    if ((nafg = pixaCountPixels(pixas)) == NULL)
	return ERROR_INT("nafg not made", procName, 1);
    pixa1 = pixaCreate(n);
    pixa2 = pixaCreate(n);
    for (i = 0; i < n; i++) {
	pix = pixaGetPix(pixas, i, CLONE);
	pix1 = pixAddBorderGeneral(pix, ADDED_PIXELS, ADDED_PIXELS,
	            ADDED_PIXELS, ADDED_PIXELS, 0);
	pix2 = pixDilate(NULL, pix1, sel);
	pixaAddPix(pixa1, pix1, INSERT);   /* un-dilated */
	pixaAddPix(pixa2, pix2, INSERT);   /* dilated */
	pixDestroy(&pix);
    }

	/* Get the centroids of all the bordered images.
	 * These are relative to the UL corner of each (bordered) pix.  */
    pta = pixaCentroids(pixa1);
    *ppta = pta;
    ptat = ptaCreate(0);  /* this will hold centroids of templates */
    *pptat = ptat;
	
	/* save the class of each c.c. */
    naclass = numaCreate(0);
    *pnaclass = naclass;

	/* Make a pixaa to store un-bordered pix, in a hierarchical
	 * set of arrays.  There is one pixa for each class,
	 * and the pix in each pixa are all the instances found
	 * of that class.  This is actually more than we need
	 * for the jbig2 encoder, but it is pedagogically useful to
	 * have these images around, because we can see when the
	 * classifier fails.  Also, the set of instances
	 * for each class can be used to make an improved (or even
	 * grayscale) template, rather than simply using the first
	 * one in the set. */
    pixaa = pixaaCreate(0);
    *ppixaa = pixaa;

	/* temporary arrays to store class exemplars (templates) */
    pixa3 = pixaCreate(0);   /* un-dilated */
    nafgt = numaCreate(0); /* fg areas of undilated templates */
    pixa4 = pixaCreate(0);   /* dilated */

	/* Fill up the pixaa tree with the template exemplars as
	 * the first pix in each pixa.  As we add each pix,
	 * we also add the associated box to the pixa.
	 * We also keep track of the centroid of each pix,
	 * and use the difference between centroids (of the
	 * pix with the exemplar we are checking it with)
	 * to align the two when checking that the Hausdorff
	 * distance does not exceed a threshold.
	 * The threshold is set by the SE used for dilating.
	 * For example, a 3x3 brick, sel_3, corresponds to a
	 * Hausdorff distance of 1.  In general, for an NxN brick,
	 * with N odd, corresponds to a Hausdorff distance of (N - 1)/2.
	 * We can't reasonably expect to use a smaller size than 3x3.
	 * Use of a SE larger than 3x3 should result in fewer classes,
	 * but also in the likelihood of more errors.
	 * Substitute sel_5 and sel_7 to see what happens
	 * to the number of classes and the apparent errors. */
    for (i = 0; i < n; i++) {   /* all instances */
	pix1 = pixaGetPix(pixa1, i, CLONE);
	numaGetIValue(nafg, i, &area1);
	pix2 = pixaGetPix(pixa2, i, CLONE);
	ptaGetPt(pta, i, &x1, &y1); 
	n3 = pixaGetCount(pixa3);
	found = FALSE;
	for (j = 0; j < n3; j++) {   /* all templates */
	    pix3 = pixaGetPix(pixa3, j, CLONE);
	    numaGetIValue(nafgt, j, &area3);
	    pix4 = pixaGetPix(pixa4, j, CLONE);
	    ptaGetPt(ptat, j, &x2, &y2);
	    testval = pixRankHaustest(pix1, pix2, pix3, pix4,
	                              x1 - x2, y1 - y2,
				      area1, area3, rank, tab8);
	    pixDestroy(&pix3);
	    pixDestroy(&pix4);
	    if (testval == 1) {  /* greedy match; take the first */
		found = TRUE;
		numaAddNumber(naclass, j);
		pixa = pixaaGetPixa(pixaa, j, NEWREF);
		pix = pixaGetPix(pixas, i, CLONE);
		pixaAddPix(pixa, pix, INSERT);
		box = boxaGetBox(boxa, i, CLONE);
		pixaAddBox(pixa, box, INSERT);
		pixaDestroy(&pixa);
		break;
	    }
	}
	if (found == FALSE) {  /* new class */
	    numaAddNumber(naclass, n3);
	    pixa = pixaCreate(0);
	    pix = pixaGetPix(pixas, i, CLONE);
	    pixaAddPix(pixa, pix, INSERT);
	    box = boxaGetBox(boxa, i, CLONE);
	    pixaAddBox(pixa, box, INSERT);
	    pixaaAddPixa(pixaa, pixa, INSERT);
	    ptaAddPt(ptat, x1, y1);
	    pixaAddPix(pixa3, pix1, INSERT);
	    pixaAddPix(pixa4, pix2, INSERT);
	    numaAddNumber(nafgt, area1);
	}
	else {   /* don't save them */
	    pixDestroy(&pix1);
	    pixDestroy(&pix2);
	}
    }

    FREE((void *)tab8);
    numaDestroy(&nafg);
    numaDestroy(&nafgt);
    pixaDestroy(&pixa1);
    pixaDestroy(&pixa2);
    pixaDestroy(&pixa3);
    pixaDestroy(&pixa4);
    selaDestroy(&sela);

    return 0;
}


/*!
 *  pixRankHaustest()
 *
 *        Input:   pix1   (new pix, not dilated)
 *                 pix2   (new pix, dilated)
 *                 pix3   (exemplar pix, not dilated)
 *                 pix4   (exemplar pix, dilated)
 *                 delx   (x comp of centroid difference)
 *                 dely   (y comp of centroid difference)
 *                 area1  (fg pixels in pix1)
 *                 area3  (fg pixels in pix3)
 *                 rank   (rank value of test, each way)
 *                 tab8   (table of pixel sums for byte)
 *
 *        Return:  0 (FALSE) if no match, 1 (TRUE) if the new
 *                 pix is in the same class as the exemplar.
 *
 *        Note: we check first that the two pix are roughly
 *        the same size.  Only if they meet that criterion do
 *        we compare the bitmaps.  We convert the rank value to
 *        a number of pixels by multiplying the rank fraction by the number
 *        of pixels in the undilated image.  The Hausdorff is a 2-way
 *        check.  The centroid difference is used to align the two
 *        images to the nearest integer for each of the checks.
 *        The rank hausdorff checks that the dilated image of one
 *        contains the rank fraction of the pixels of the undilated
 *        image of the other.   Checks are done in both direction. 
 *        Failure of the test in either direction results in failure
 *        of the test.
 *
 */
INT32
pixRankHaustest(PIX      *pix1,
                PIX      *pix2,
	        PIX      *pix3,
	        PIX      *pix4,
	        FLOAT32   delx,   /* x(1) - x(3) */
	        FLOAT32   dely,   /* y(1) - y(3) */
		INT32     area1,
		INT32     area3,
	        FLOAT32   rank,
		INT32    *tab8)
{
INT32   wi, hi, wt, ht, delw, delh, idelx, idely, boolval;
INT32   thresh1, thresh3;
PIX    *pixt;

	/* eliminate possible matches based on size difference */
    wi = pixGetWidth(pix1);
    hi = pixGetHeight(pix1);
    wt = pixGetWidth(pix3);
    ht = pixGetHeight(pix3);
    delw = ABS(wi - wt);
    if (delw > MAX_DIFF_WIDTH)
	return FALSE;
    delh = ABS(hi - ht);
    if (delh > MAX_DIFF_HEIGHT)
	return FALSE;

	/* upper bounds in remaining pixels for allowable match */
    thresh1 = (INT32)(area1 * (1. - rank) + 0.5);
    thresh3 = (INT32)(area3 * (1. - rank) + 0.5);

	/* round difference in centroid location to nearest integer;
	 * use this as a shift when doing the matching. */
    if (delx >= 0)
	idelx = (INT32)(delx + 0.5);
    else
	idelx = (INT32)(delx - 0.5);
    if (dely >= 0)
	idely = (INT32)(dely + 0.5);
    else
	idely = (INT32)(dely - 0.5);

	/*  Do 1-direction hausdorff, checking that every pixel in pix1
	 *  is within a dilation distance of some pixel in pix3.  Namely,
	 *  that pix4 entirely covers pix1:
	 *       pixt = pixSubtract(NULL, pix1, pix4), including shift
	 *  where pixt has no ON pixels.  */
    pixt = pixCreateTemplate(pix1);
    pixRasterop(pixt, 0, 0, wi, hi, PIX_SRC, pix1, 0, 0);
    pixRasterop(pixt, idelx, idely, wi, hi, PIX_DST & PIX_NOT(PIX_SRC),
                pix4, 0, 0);
    pixThresholdPixels(pixt, thresh1, &boolval, tab8);
    if (boolval == 1) { /* above thresh1 */
	pixDestroy(&pixt);
	return FALSE;
    }

	/*  Do 1-direction hausdorff, checking that every pixel in pix3
	 *  is within a dilation distance of some pixel in pix1.  Namely,
	 *  that pix2 entirely covers pix3:
	 *      pixSubtract(pixt, pix3, pix2), including shift
	 *  where pixt has no ON pixels. */
    pixRasterop(pixt, idelx, idely, wt, ht, PIX_SRC, pix3, 0, 0);
    pixRasterop(pixt, 0, 0, wt, ht, PIX_DST & PIX_NOT(PIX_SRC), pix2, 0, 0);
    pixThresholdPixels(pixt, thresh3, &boolval, tab8);
    pixDestroy(&pixt);
    if (boolval == 1)  /* above thresh3 */
	return FALSE;
    else
	return TRUE;
}



/*----------------------------------------------------------------------*
 *            Classification using windowed correlation score           *
 *----------------------------------------------------------------------*/
/*!
 *  jbig2ClassifyCorrel()
 *
 *      Input:  boxa (of c.c.)
 *              pixas (of c.c.)
 *              thresh (correlation threshold score; e.g., 0.75)
 *              &pixaa (<return> 2-level array of instances for each class)
 *              &pta (<return> centroids of all bordered c.c.)
 *              &ptat (<return> centroids of all bordered template c.c.)
 *              &naclass (<return> class label of all c.c.)
 *      Return: 0 if OK; 1 on error
 */
INT32
jbig2ClassifyCorrel(BOXA     *boxa,
                    PIXA     *pixas,
		    FLOAT32   thresh,
		    PIXAA   **ppixaa,
		    PTA     **ppta,
		    PTA     **pptat,
		    NUMA    **pnaclass)
{
INT32          ncc, n2, i, j, found, area1, area2;
INT32         *tab8;
FLOAT32        x1, y1, x2, y2;
FLOAT32        score;
BOX           *box;
NUMA          *naclass;
NUMA          *na;    /* counts of foreground pixels in each c.c. */
NUMA          *nat;   /* counts of foreground pixels in each template */
PIX           *pix, *pix1, *pix2;
PIXA          *pixa, *pixa1, *pixa2;
PIXAA         *pixaa;
PTA           *pta, *ptat;

    PROCNAME("jbig2ClassifyCorrel");

    if (!boxa)
	return ERROR_INT("boxa not found", procName, 1);
    if (!pixas)
	return ERROR_INT("pixas not found", procName, 1);
    if (thresh < 0.4 || thresh > 0.9)
	return ERROR_INT("thresh not in range (0.4 - 0.9)", procName, 1);
    if (!ppta)
	return ERROR_INT("&pta not defined", procName, 1);
    if (!pptat)
	return ERROR_INT("&ptat not defined", procName, 1);
    if (!pnaclass)
	return ERROR_INT("&pnaclass not defined", procName, 1);

	/* generate the bordered pixa1, which contains all the input c.c. */
    ncc = pixaGetCount(pixas);
    pixa1 = pixaCreate(ncc);
    for (i = 0; i < ncc; i++) {
	pix = pixaGetPix(pixas, i, CLONE);
	pix1 = pixAddBorderGeneral(pix, ADDED_PIXELS, ADDED_PIXELS,
	            ADDED_PIXELS, ADDED_PIXELS, 0);
	pixaAddPix(pixa1, pix1, INSERT);
	pixDestroy(&pix);
    }

	/* get the centroids of all the bordered images */
    pta = pixaCentroids(pixa1);
    *ppta = pta;
    ptat = ptaCreate(0);  /* this will hold centroids of templates */
    *pptat = ptat;
	
        /* save the class of each c.c. */
    naclass = numaCreate(0);
    *pnaclass = naclass;

	/* get the number of ON pixels in each c.c. */
    na = pixaCountPixels(pixa1);
    nat = numaCreate(0);  /* this will hold areas of templates */
    tab8 = makePixelSumTab8();

	/* Make a pixaa to store un-bordered pix, in a hierarchical
	 * set of arrays.  There is one pixa for each class,
	 * and the pix in each pixa are all the instances found
	 * of that class.  This is actually more than we need
	 * for the jbig2 encoder, but it is pedagogically useful to
	 * have these images around, because we can see when the
	 * classifier fails.  Also, the set of instances
	 * for each class can be used to make an improved (or even
	 * grayscale) template, rather than simply using the first
	 * one in the set. */
    pixaa = pixaaCreate(0);
    *ppixaa = pixaa;

	/* temporary array to store class exemplars */
    pixa2 = pixaCreate(0);

	/* Fill up the pixaa tree with the template exemplars as
	 * the first pix in each pixa.  As we add each pix,
	 * we also add the associated box to the pixa.
	 * We also keep track of the centroid of each pix,
	 * and use the difference between centroids (of the
	 * pix with the exemplar we are checking it with)
	 * to align the two when checking that the correlation
	 * score exceeds a threshold.  The correlation score
	 * is given by the square of the area of the AND
	 * between aligned instance and template, divided by
	 * the product of areas of each image.  For identical
	 * template and instance, the score is 1.0.
	 * If the threshold is too small, non-equivalent instances
	 * will be placed in the same class; if too large, there will
	 * be an unnecessary division of classes representing the
	 * same character. */
    for (i = 0; i < ncc; i++) {
	pix1 = pixaGetPix(pixa1, i, CLONE);
	numaGetIValue(na, i, &area1);
	ptaGetPt(pta, i, &x1, &y1); 
	n2 = pixaGetCount(pixa2);
	found = FALSE;
	for (j = 0; j < n2; j++) {
	    pix2 = pixaGetPix(pixa2, j, CLONE);
	    numaGetIValue(nat, j, &area2);
	    ptaGetPt(ptat, j, &x2, &y2);
	    score = pixCorrelationScore(pix1, pix2, area1, area2,
	                x1 - x2, y1 - y2, tab8);
	    pixDestroy(&pix2);
	    if (score > thresh) {
		found = TRUE;
		numaAddNumber(naclass, j);
		pixa = pixaaGetPixa(pixaa, j, NEWREF);
		pix = pixaGetPix(pixas, i, CLONE);
		pixaAddPix(pixa, pix, INSERT);
		box = boxaGetBox(boxa, i, CLONE);
		pixaAddBox(pixa, box, INSERT);
		pixaDestroy(&pixa);
		break;
	    }
	}
	if (found == FALSE) {  /* new class */
	    numaAddNumber(naclass, n2);
	    pixa = pixaCreate(0);
	    pix = pixaGetPix(pixas, i, CLONE);
	    pixaAddPix(pixa, pix, INSERT);
	    box = boxaGetBox(boxa, i, CLONE);
	    pixaAddBox(pixa, box, INSERT);
	    pixaaAddPixa(pixaa, pixa, INSERT);
	    ptaAddPt(ptat, x1, y1);
	    numaAddNumber(nat, area1);
	    pixaAddPix(pixa2, pix1, INSERT);
	}
	else {   /* don't save it */
	    pixDestroy(&pix1);
	}
    }

    FREE((void *)tab8);
    pixaDestroy(&pixa1);
    pixaDestroy(&pixa2);
    numaDestroy(&na);
    numaDestroy(&nat);
    return 0;
}


/*!
 *  pixCorrelationScore()
 *
 *        Input:   pix1   (test pix)
 *                 pix2   (exemplar pix)
 *                 area1  (number of on pixels in pix1)
 *                 area2  (number of on pixels in pix2)
 *                 delx   (x comp of centroid difference)
 *                 dely   (y comp of centroid difference)
 *                 tab    (sum tab for byte)
 *        Return:  correlation score 
 *
 *        Note: we check first that the two pix are roughly
 *        the same size.  Only if they meet that criterion do
 *        we compare the bitmaps. 
 *        The centroid difference is used to align the two
 *        images to the nearest integer for the correlation.
 *
 *        The correlation score is the ratio of the square of
 *        the number of pixels in the AND of the two bitmaps 
 *        to the product of the number of ON pixels in each.
 *
 */
FLOAT32
pixCorrelationScore(PIX      *pix1,
                    PIX      *pix2,
		    INT32     area1,
		    INT32     area2,
	            FLOAT32   delx,   /* x(1) - x(3) */
	            FLOAT32   dely,   /* y(1) - y(3) */
		    INT32    *tab)
{
INT32     wi, hi, wt, ht, delw, delh, idelx, idely, count;
FLOAT32   score;
PIX      *pixt;

	/* eliminate based on size difference */
    wi = pixGetWidth(pix1);
    hi = pixGetHeight(pix1);
    wt = pixGetWidth(pix2);
    ht = pixGetHeight(pix2);
    delw = ABS(wi - wt);
    if (delw > MAX_DIFF_WIDTH)
	return FALSE;
    delh = ABS(hi - ht);
    if (delh > MAX_DIFF_HEIGHT)
	return FALSE;

	/* round difference to nearest integer */
    if (delx >= 0)
	idelx = (INT32)(delx + 0.5);
    else
	idelx = (INT32)(delx - 0.5);
    if (dely >= 0)
	idely = (INT32)(dely + 0.5);
    else
	idely = (INT32)(dely - 0.5);

	/*  pixt = pixAnd(NULL, pix1, pix2), including shift */
    pixt = pixCreateTemplate(pix1);
    pixRasterop(pixt, 0, 0, wi, hi, PIX_SRC, pix1, 0, 0);
    pixRasterop(pixt, idelx, idely, wt, ht, PIX_SRC & PIX_DST, pix2, 0, 0);
    pixCountPixels(pixt, &count, tab);
    pixDestroy(&pixt);
    score = (FLOAT32)(count * count) / (FLOAT32)(area1 * area2);
/*    fprintf(stderr, "score = %7.3f, count = %d, area1 = %d, area2 = %d\n",
             score, count, area1, area2); */
    return score;
}


/*----------------------------------------------------------------------*
 *                       jbig2 utility routines                         *
 *----------------------------------------------------------------------*/
/*!
 *  jbig2GetULCorners()
 *
 *     Input:  boxa  (of c.c. bounding rectangles)
 *             naclass (class assignment of all c.c.)
 *             pta   (of centroids of all c.c.)
 *             ptat  (of centroids of template c.c.
 *     Return: ptaout  (of UL corners to give placement of
 *                      template)
 *
 *     Note: ptaout gives the global UL corners, adjusted for each specific
 *           component, so that each component can be replaced by the template
 *           for its class and have the centroid in the template in
 *           the same position as the centroid of the original c.c.
 *           The input arrays pta and ptat give the centroids of
 *           those components relative to the UL corner of each component.
 *           We compute the difference in each component, round to
 *           nearest integer, and correct the box->x and box->y by
 *           the appropriate integral difference.
 */
PTA *
jbig2GetULCorners(BOXA   *boxa,
		  NUMA   *naclass,
                  PTA    *pta,
		  PTA    *ptat)
{
INT32    i, n, iclass, idelx, idely, x, y;
FLOAT32  x1, x2, y1, y2, delx, dely;
BOX     *box;
PTA     *ptaout;

    PROCNAME("jbig2GetULCorners");

    n = boxaGetCount(boxa);
    ptaout = ptaCreate(n);
    for (i = 0; i < n; i++) {
	ptaGetPt(pta, i, &x1, &y1);
	numaGetIValue(naclass, i, &iclass);
	ptaGetPt(ptat, iclass, &x2, &y2);
	delx = x2 - x1;
	dely = y2 - y1;
	if (delx >= 0)
	    idelx = (INT32)(delx + 0.5);
	else
	    idelx = (INT32)(delx - 0.5);
	if (dely >= 0)
	    idely = (INT32)(dely + 0.5);
	else
	    idely = (INT32)(dely - 0.5);
	box = boxaGetBox(boxa, i, CLONE);
	if (!box)
	    return (PTA *)ERROR_PTR("box not found", procName, NULL);
	x = box->x;
	y = box->y;
	boxDestroy(&box);
	ptaAddPt(ptaout, x - idelx, y - idely);
    }

    return ptaout;
}


/*!
 *  jbig2CreateData()
 *
 *      Input:  pix (composite of template images)
 *              w, h  (of original image)
 *              nclass (number of templates)
 *              latticew, latticeh (cell size used to store each c.c.
 *                  in the composite)
 *              naclass (array giving the class for each c.c.)
 *              pta (array of UL corners, relative to the page, at
 *                 which each template is properly aligned to replace
 *                 the actual instance in the original)
 *      Return: jb2data, or null on error.
 *
 *      Note: this routine stores the jbig2 data required for
 *            generating the approximate jbig2 version of the image.
 *            It can be losslessly written to (and read from) two files.
 *
 *      Action: makes copies/clones of the 3 data structures and puts
 *              them into this container. 
 */
JB2DATA *
jbig2CreateData(PIX    *pix,
                INT32   w,
		INT32   h,
		INT32   nclass,
		INT32   latticew,
		INT32   latticeh,
		NUMA   *naclass,
		PTA    *pta)
{
JB2DATA    *data;

    PROCNAME("jbig2CreateData");

    if ((data = (JB2DATA *)CALLOC(1, sizeof(JB2DATA))) == NULL)
	return (JB2DATA *)ERROR_PTR("data not made", procName, NULL);

    data->pix = pixClone(pix);
    data->w = w;
    data->h = h;
    data->nclass = nclass;
    data->latticew = latticew;
    data->latticeh = latticeh;
    data->naclass = numaCopy(naclass);
    data->pta = ptaCopy(pta);

    return data;
}


/*
 *  jbig2DestroyData()
 *
 *     Input:  &data (<to be nulled>)
 *     Return: void
 */
void
jbig2DestroyData(JB2DATA  **pdata)
{
JB2DATA   *data;

    if (!pdata)
        return;
    if ((data = *pdata) == NULL)
        return;

    pixDestroy(&data->pix);
    numaDestroy(&data->naclass);
    ptaDestroy(&data->pta);
    FREE((void *)data);
    *pdata = NULL;
    return;
}
    

/*!
 *  jbig2WriteData()
 *
 *      Input:  rootname (for output files)
 *              jb2data
 *      Return: 0 if OK, 1 on error
 *      Action: writes data in jb2data to file
 */
INT32
jbig2WriteData(char      *rootout,
               JB2DATA   *jb2data)
{
char      buf[BUF_SIZE];
INT32     w, h, nclass, cellw, cellh, ncc, i, x, y, iclass;
NUMA     *naclass;
PTA      *pta;
PIX      *pixt;
FILE     *fp;
    
    PROCNAME("jbig2WriteData");

    if (!rootout)
        return ERROR_INT("no rootout", procName, 1);
    if (!jb2data)
        return ERROR_INT("no jb2data", procName, 1);

    w = jb2data->w;
    h = jb2data->h;
    pixt = jb2data->pix;
    nclass = jb2data->nclass;
    cellw = jb2data->latticew;
    cellh = jb2data->latticeh;
    naclass = jb2data->naclass;
    pta = jb2data->pta;

    sprintf(buf, "%s%s", rootout, JBIG2_TEMPLATE_EXT); 
    pixWrite(buf, pixt, IFF_PNG);

    sprintf(buf, "%s%s", rootout, JBIG2_DATA_EXT); 
    if ((fp = fopen(buf, "w")) == NULL)
	return ERROR_INT("stream not opened", procName, 1);
    ncc = ptaGetCount(pta);
    fprintf(fp, "jbig2 data file\n");
    fprintf(fp, "image size: w = %d, h = %d\n", w, h);
    fprintf(fp, "num c.c. = %d\n", ncc);
    fprintf(fp, "num classes = %d\n", nclass);
    fprintf(fp, "lattice sizes: w = %d, h = %d\n", cellw, cellh);
    for (i = 0; i < ncc; i++) {
	ptaGetIPt(pta, i, &x, &y);
	numaGetIValue(naclass, i, &iclass);
	fprintf(fp, "%d %d %d\n", iclass, x, y);
    }
    fclose(fp);

    return 0;
}



/*!
 *  jbig2DecodeData()
 *
 *          Input:  imagefile (of templates)
 *                  datafile (of class numbers and UL corners for each c.c)
 *                  &jb2data (<returned> of all the data in the two files;
 *                      can be null, in which case nothing is returned)
 *          Return: pix (reconstruction of full image, using templates) or
 *                  null on error
 */
PIX *
jbig2DecodeData(char      *imagefile,
                char      *datafile,
		JB2DATA  **pjb2data)
{
INT32    i, w, h, cellw, cellh, x, y, iclass;
INT32    nclass, ncc, wp, hp;
JB2DATA *jb2data;
NUMA    *naclass;
PIX     *pixt, *pix, *pixd;
PIXA    *pixa;
PTA     *pta;
    
    PROCNAME("jbig2DecodeData");

    if (pjb2data)
	*pjb2data = NULL;

        /* read data into memory from the two files */
    if ((jb2data = jbig2ReadData(imagefile, datafile)) == NULL)
	return (PIX *)ERROR_PTR("jb2data not made", procName, NULL);

    w = jb2data->w;
    h = jb2data->h;
    pixt = jb2data->pix;
    nclass = jb2data->nclass;
    cellw = jb2data->latticew;
    cellh = jb2data->latticeh;
    naclass = jb2data->naclass;
    pta = jb2data->pta;
    ncc = numaGetCount(naclass);
    
	/* reconstruct the original image from the templates and
	 * the data associated with each connected component. */
    if ((pixd = pixCreate(w, h, 1)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    if ((pixa = pixaCreateFromPix(pixt, nclass, cellw, cellh)) == NULL)
	return (PIX *)ERROR_PTR("pixa not made", procName, NULL);
    for (i = 0; i < ncc; i++) {
	numaGetIValue(naclass, i, &iclass);
	pix = pixaGetPix(pixa, iclass, CLONE);
	wp = pixGetWidth(pix);
	hp = pixGetHeight(pix);
	ptaGetIPt(pta, i, &x, &y);
	pixRasterop(pixd, x, y, wp, hp, PIX_SRC | PIX_DST, pix, 0, 0);
	pixDestroy(&pix);
    }

	/* save jb2data if requested */
    if (pjb2data != NULL)
	*pjb2data = jb2data;
    else
        jbig2DestroyData(&jb2data);

    pixaDestroy(&pixa);
    return pixd;
}


/*!
 *  jbig2ReadData()
 *
 *      Input:  imagefile (name of template file)
 *              datafile (name of file containing class number and
 *              UL global coords for each c.c.)
 *      Return: jb2data, or NULL on error
 *      Action: reads data from file and stores it in a jb2data
 */
JB2DATA *
jbig2ReadData(char   *imagefile,
              char   *datafile)
{
char      *linestr;
UINT8     *data;
INT32      nsa, i, w, h, cellw, cellh, x, y, iclass;
INT32      nbytes, nclass, ncc;
JB2DATA   *jb2data;
NUMA      *naclass;
PIX       *pixs;
PTA       *pta;
SARRAY    *sa;
    
    PROCNAME("jbig2ReadData");

    if ((pixs = pixRead(imagefile)) == NULL)
	return (JB2DATA *)ERROR_PTR("pix not read", procName, NULL);

    if ((data = fileToArray(datafile, &nbytes)) == NULL)
	return (JB2DATA *)ERROR_PTR("data not read", procName, NULL);

    if ((sa = sarrayCreateLinesFromString((char *)data, 0)) == NULL)
	return (JB2DATA *)ERROR_PTR("sa not made", procName, NULL);
    nsa = sarrayGetCount(sa);   /* num of c.c. + 5 */
    linestr = sarrayGetString(sa, 1, 0);
    sscanf(linestr, "image size: w = %d, h = %d", &w, &h);
    linestr = sarrayGetString(sa, 2, 0);
    sscanf(linestr, "num c.c. = %d", &ncc);
    linestr = sarrayGetString(sa, 3, 0);
    sscanf(linestr, "num classes = %d\n", &nclass);
    linestr = sarrayGetString(sa, 4, 0);
    sscanf(linestr, "lattice sizes: w = %d, h = %d\n", &cellw, &cellh);

#if 0
    fprintf(stderr, "image size: w = %d, h = %d\n", w, h);
    fprintf(stderr, "num c.c. = %d\n", ncc);
    fprintf(stderr, "num classes = %d\n", nclass);
    fprintf(stderr, "lattice sizes: w = %d, h = %d\n", cellw, cellh);
#endif

    if ((naclass = numaCreate(ncc)) == NULL)
	return (JB2DATA *)ERROR_PTR("naclass not made", procName, NULL);
    if ((pta = ptaCreate(ncc)) == NULL)
	return (JB2DATA *)ERROR_PTR("pta not made", procName, NULL);
    for (i = 5; i < nsa; i++) {
	linestr = sarrayGetString(sa, i, 0);
	sscanf(linestr, "%d %d %d\n", &iclass, &x, &y);
	numaAddNumber(naclass, iclass);
	ptaAddPt(pta, x, y);
    }

    jb2data = jbig2CreateData(pixs, w, h, nclass, cellw, cellh, naclass, pta);

    pixDestroy(&pixs);
    FREE((void *)data);
    sarrayDestroy(&sa);
    numaDestroy(&naclass);
    ptaDestroy(&pta);

    return jb2data;
}

