/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *  pix1.c
 *
 *      Pix creation
 *          PIX          *pixCreate()
 *          PIX          *pixCreateTemplate()
 *          PIX          *pixClone()
 *
 *      Pix destruction
 *          void          pixDestroy()
 *
 *      Pix copy
 *          PIX          *pixCopy()
 *	    INT32         pixCopyColormap()
 *          INT32         pixSizesEqual()
 *
 *      Pix accessors
 *          INT32         pixGetWidth()
 *          INT32         pixSetWidth()
 *          INT32         pixGetHeight()
 *          INT32         pixSetHeight()
 *          INT32         pixGetDepth()
 *          INT32         pixSetDepth()
 *          INT32         pixGetWpl()
 *          INT32         pixSetWpl()
 *          INT32         pixGetRefcount()
 *          INT32         pixChangeRefcount()
 *          UINT32        pixGetXRes()
 *          UINT32        pixGetYRes()
 *          INT32         pixSetXRes()
 *          INT32         pixSetYRes()
 *          char         *pixGetText()
 *          INT32         pixSetText()
 *          INT32         pixFreeColormap()
 *          PIXCMAP      *pixGetColormap()
 *          INT32         pixSetColormap()
 *          UINT8        *pixGetData()
 *          INT32         pixSetData()
 *
 *      Pix debug
 *          INT32         pixPrintStreamInfo()
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "allheaders.h"

#ifndef  NO_CONSOLE_IO
#define   DEBUG    0
#endif  /* ~NO_CONSOLE_IO */

    /* Can use this to pad each rasterline.  Not presently necessary */
#define  EXTRA_WORDS_ON_RIGHT      0



/*--------------------------------------------------------------------*
 *                              Pix Creation                          *
 *--------------------------------------------------------------------*/
/*!
 *  pixCreate() 
 *
 *     Input:  width, height, depth
 *     Return: pixd, or null on error
 *     Action: creates a Pix with data array allocated
 */
PIX *
pixCreate(INT32     width,
	  INT32     height,
	  INT32     depth)
{
INT32      wpl;
PIX       *pixd;
UINT32    *data;

    PROCNAME("pixCreate");

    if ((depth != 1) && (depth != 2) && (depth != 4) && (depth != 8)
	     && (depth != 16) && (depth != 24) && (depth != 32))
	return (PIX *)ERROR_PTR("depth must be {1, 2, 4, 8, 16, 24, 32}",
	     procName, NULL);
    if (width <= 0)
	return (PIX *)ERROR_PTR("width must be > 0", procName, NULL);
    if (height <= 0)
	return (PIX *)ERROR_PTR("height must be > 0", procName, NULL);

    if ((pixd = (PIX *)CALLOC(1, sizeof(PIX))) == NULL)
	return (PIX *)ERROR_PTR("CALLOC fail for pixd", procName, NULL);

    pixSetWidth(pixd, width);
    pixSetHeight(pixd, height);
    pixSetDepth(pixd, depth);

	/* So far, there is no need to add an extra word to each
	 * raster line.  But the mechanism is here if needed, and
	 * should have no side effects if enabled, as all ops that
	 * use the word pad bits use the actual image width, not wpl.
	 * When I got valgrind, I found that rasteropGeneralLow() was
	 * making occasional read errors one word off the end
	 * of the data, and adding an extra word to each raster
	 * line was a quick fix, but is currently not necessary. */  
    wpl = (width * depth + 31) / 32 + EXTRA_WORDS_ON_RIGHT;
    pixSetWpl(pixd, wpl);
    if ((data = (UINT32 *)CALLOC(4 * wpl, height)) == NULL)
	return (PIX *)ERROR_PTR("CALLOC fail for data", procName, NULL);
    pixSetData(pixd, data);
    pixd->refcount = 1;

    return pixd;
}


/*!
 *  pixCreateTemplate()
 *
 *     Input:  pixs
 *     Return: pixd, or null on error
 *     Action: makes a Pix of the same size as the input Pix,
 *             allocating the data array, but does not copy the
 *             image data from pixs
 */
PIX *
pixCreateTemplate(PIX  *pixs)
{
INT32     w, h, d;
PIX      *pixd;

    PROCNAME("pixCreateTemplate");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    d = pixGetDepth(pixs);

    if ((pixd = pixCreate(w, h, d)) == NULL)
        return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    pixSetXRes(pixd, pixGetXRes(pixs));
    pixSetYRes(pixd, pixGetYRes(pixs));
    if (pixCopyColormap(pixd, pixs))
        return (PIX *)ERROR_PTR("colormap copy fail", procName, NULL);
    pixSetText(pixd, pixGetText(pixs));
    
    return pixd;
}


/*!
 *  pixClone()
 *
 *         Input:  pix
 *         Return: same pix (ptr), or null on error
 *
 *         N.B.  Why is this here?  We make a "clone", which
 *         is just another handle to an existing pix, because
 *         (1) images can be large and hence expensive to copy,
 *         (2) extra handles to a data structure that are made
 *         without some control are dangerous because if you
 *         have two (say), how do you know which one should be
 *         used to destroy the pix, and, conversely, if you
 *         destroy the pix, how do you remember that the other
 *         handle is no longer valid?  This is solved by a
 *         simple protocol: (1) whenever you want a new handle
 *         to an existing image, call pixClone(), which just
 *         bumps a ref count, and (2) always call pixDestroy()
 *         on ALL handles, which decrements the ref count and
 *         only destroys the pix when pixDestroy() has
 *         been called on all handles.
 */
PIX *
pixClone(PIX  *pixs)
{

    PROCNAME("pixClone");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);

    pixChangeRefcount(pixs, 1);

    return pixs;
}



/*--------------------------------------------------------------------*
 *                           Pix Destruction                          *
 *--------------------------------------------------------------------*/
/*!
 *  pixDestroy()
 *
 *     Input:  &pix <can be nulled>
 *     Return: void
 *     Action: decrements the ref count and, if 0, destroys the Pix
 *             and nulls the input ptr.
 */
void
pixDestroy(PIX   **ppix)
{
UINT32       *data;
char         *text;
PIXCMAP      *cmap;
PIX          *pix;

    PROCNAME("pixDestroy");

    if (!ppix) {
	WARNING("ptr address is null!", procName);
        return;
    }

    if ((pix = *ppix) == NULL)
	return;

	/* decrement the ref count.  If it is still > 0, just return */
    pixChangeRefcount(pix, -1);
    if (pixGetRefcount(pix) > 0)
	return;

	/* ok, the refcount is 0, so destroy it */
    if (data = pixGetData(pix))
	FREE((void *)data);
    if (text = pixGetText(pix))
	FREE((void *)text);
    if ((cmap = pixGetColormap(pix)) != NULL) { 
	if (cmap->array != NULL )
	    FREE((void *)cmap->array);
	FREE((void *)cmap);
    }
    FREE((void *)pix);

    *ppix = NULL;
    return;
}


/*-------------------------------------------------------------------------*
 *                                 Pix Copy                                *
 *-------------------------------------------------------------------------*/
/*!
 *  pixCopy()
 *
 *    Input:  pixd (<optional>)
 *            pixs
 *    Return: pixd, or null on error
 *
 *    N.B. If pixd = NULL, this makes a new copy, with refcount of 1.
 *         If pixd != NULL, this makes sure pixs and pixd are
 *            the same size, and then copies the image data,
 *            leaving the refcounts of pixs and pixd unchanged.
 *
 *    Note: this operation, like all others that may involve a
 *          pre-existing pixd, will side-effect any existing
 *          clones of pixd.
 */
PIX *
pixCopy(PIX  *pixd,   /* can be null */
	PIX  *pixs)
{
INT32      bytes;
UINT32    *datas, *datad;

    PROCNAME("pixCopy");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, pixd);

	/* total bytes in image data */
    bytes = 4 * pixGetWpl(pixs) * pixGetHeight(pixs);

	/* if we're making a new pix ... */
    if (!pixd) {
	if ((pixd = pixCreateTemplate(pixs)) == NULL)
	    return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
	datas = pixGetData(pixs);
	datad = pixGetData(pixd);
	memcpy((char *)datad, (char *)datas, bytes);
	return pixd;
    }

	/* programmer error! */
    if (pixs == pixd) {
        fprintf(stderr, "Warning: image data in %s copied to itself\n",
	        procName);
	return pixd;
    }

        /* check sizes */
    if (!pixSizesEqual(pixs, pixd))
	    return (PIX *)ERROR_PTR("pix sizes not equal", procName, pixd);

	/* copy the data */
    datas = pixGetData(pixs);
    datad = pixGetData(pixd);
    memcpy((char*)datad, (char*)datas, bytes);
    return pixd;
}


/*!
 *  pixCopyColormap()
 *
 *    Input: src and dest Pix
 *    Return: 0 if OK, 1 on error.
 */
INT32
pixCopyColormap(PIX  *pixd,
                PIX  *pixs)
{
PIXCMAP     *scmap, *dcmap;

    PROCNAME("pixCopyColormap");

    if (!pixs)
	return ERROR_INT("pixs not defined", procName, 1);
    if (!pixd)
	return ERROR_INT("pixd not defined", procName, 1);

    scmap = pixGetColormap(pixs);
    if (scmap) {  /* copy colormap and insert into pixd */
	if ((dcmap = (PIXCMAP *)CALLOC(1, sizeof(PIXCMAP))) == NULL)
            return ERROR_INT("calloc fail for dcmap", procName, 1);
	if ((dcmap->array = (void *)CALLOC(scmap->size, 1)) == NULL)
	    return ERROR_INT("calloc fail for colormap array", procName, 2);
	dcmap->size   = scmap->size;
	dcmap->type  = scmap->type;
	memcpy(dcmap->array, scmap->array, dcmap->size);

	pixSetColormap(pixd, dcmap);
    }

    return 0;
}


/*!
 *  pixSizesEqual()
 *
 *      Input: two Pix
 *      Return: 1 if the two Pix have same {h, w, d}; 0 otherwise.
 */
INT32
pixSizesEqual(PIX  *pix1,
	      PIX  *pix2)
{

    PROCNAME("pixSizesEqual");

    if (!pix1)
        return ERROR_INT("pix1 not defined", procName, 0);
    if (!pix2)
        return ERROR_INT("pix2 not defined", procName, 0);

    if ((pixGetWidth(pix1) != pixGetWidth(pix2)) ||
        (pixGetHeight(pix1) != pixGetHeight(pix2)) ||
        (pixGetDepth(pix1) != pixGetDepth(pix2))) {
#if DEBUG
	fprintf(stderr, "w1 = %d, w2 = %d, h1 = %d, h2 = %d, d1 = %d, d2 = %d\n",
	                 pixGetWidth(pix1), pixGetWidth(pix2),
	                 pixGetHeight(pix1), pixGetHeight(pix2),
	                 pixGetDepth(pix1), pixGetDepth(pix2));
#endif /* DEBUG */
	return 0;
    }
    else
	return 1;
}



/*--------------------------------------------------------------------*
 *                                Accessors                           *
 *--------------------------------------------------------------------*/
INT32
pixGetWidth(PIX  *pix)
{
    PROCNAME("pixGetWidhth");

    if (!pix)
	return ERROR_INT("pix not defined", procName, UNDEF);

    return pix->w;
}


INT32
pixSetWidth(PIX   *pix,
	    INT32  width)
{
    PROCNAME("pixSetWidth");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);
    if (width < 0) {
        pix->w = 0;
	return ERROR_INT("width must be >= 0", procName, 1);
    }

    pix->w = width;
    return 0;
}


INT32
pixGetHeight(PIX  *pix)
{
    PROCNAME("pixGetHeight");

    if (!pix)
	return ERROR_INT("pix not defined", procName, UNDEF);

    return pix->h;
}


INT32
pixSetHeight(PIX  *pix,
	     INT32  height)
{

    PROCNAME("pixSetHeight");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);
    if (height < 0) {
        pix->h = 0;
	return ERROR_INT("h must be >= 0", procName, 1);
    }

    pix->h = height;
    return 0;
}


INT32 
pixGetDepth(PIX  *pix)
{

    PROCNAME("pixGetDepth");

    if (!pix)
	return ERROR_INT("pix not defined", procName, UNDEF);

    return pix->d;
}


INT32  
pixSetDepth(PIX   *pix,
	    INT32  depth)
{

    PROCNAME("pixSetDepth");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);
    if (depth < 1) 
	return ERROR_INT("d must be >= 1", procName, 1);

    pix->d = depth;
    return 0;
}


INT32  
pixGetWpl(PIX  *pix)
{

    PROCNAME("pixGetWpl");

    if (!pix)
	return ERROR_INT("pix not defined", procName, UNDEF);

    return pix->wpl;
}


INT32  
pixSetWpl(PIX   *pix,
	  INT32  wpl)
{

    PROCNAME("pixSetWpl");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);

    pix->wpl = wpl;
    return 0;
}


INT32
pixGetRefcount(PIX  *pix)
{

    PROCNAME("pixGetRefcount");

    if (!pix)
	return ERROR_INT("pix not defined", procName, UNDEF);

    return pix->refcount;
}


INT32  
pixChangeRefcount(PIX   *pix,
	          INT32  delta)
{

    PROCNAME("pixChangeRefcount");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);

    pix->refcount += delta;
    return 0;
}


UINT32  
pixGetXRes(PIX  *pix)
{

    PROCNAME("pixGetXRes");

    if (!pix)
	return ERROR_INT("pix not defined", procName, UNDEF);

    return pix->xres;
}


UINT32  
pixGetYRes(PIX  *pix)
{

    PROCNAME("pixGetYRes");

    if (!pix)
	return ERROR_INT("pix not defined", procName, UNDEF);

    return pix->yres;
}


INT32  
pixSetXRes(PIX    *pix,
	   UINT32  res)
{

    PROCNAME("pixSetXRes");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);

    pix->xres = res;
    return 0;
}


INT32  
pixSetYRes(PIX    *pix,
	   UINT32  res)
{

    PROCNAME("pixSetYRes");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);

    pix->yres = res;
    return 0;
}


char *  
pixGetText(PIX  *pix)
{

    PROCNAME("pixGetText");

    if (!pix)
	return (char *)ERROR_PTR("pix not defined", procName, NULL);

    return pix->text;
}


/*!
 *   pixSetText()
 *
 *      Input: pix
 *             textstring
 *      Return: 0 if OK, 1 on error
 *
 */
INT32  
pixSetText(PIX   *pix,
	   char  *textstring)
{

    PROCNAME("pixSetText");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);

    stringReplace(&pix->text, textstring);
    return 0;
}



/*!
 *   pixFreeColormap()
 *
 *     Input:  pix
 *     Return: 0 if OK, 1 on error
 */
INT32  
pixFreeColormap(PIX   *pix )
{
PIXCMAP   *cmap;
void      *cmaparray;

    PROCNAME("pixFreeColormap");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);

    if ((cmap = pix->colormap) != NULL) {
	if ((cmaparray = cmap->array) != NULL)
	    FREE(cmaparray);
	FREE((void*)cmap);
	pix->colormap = NULL;
    }
    return 0;
}


PIXCMAP *  
pixGetColormap(PIX  *pix)
{

    PROCNAME("pixGetColormap");

    if (!pix)
	return (PIXCMAP *)ERROR_PTR("pix not defined", procName, NULL);

    return pix->colormap;
}


INT32  
pixSetColormap(PIX      *pix,
	       PIXCMAP  *colormap)
{

    PROCNAME("pixSetColormap");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);

    pix->colormap = colormap;
    return 0;
}


UINT32 *
pixGetData(PIX  *pix)
{

    PROCNAME("pixGetData");

    if (!pix)
	return (UINT32 *)ERROR_PTR("pix not defined", procName, NULL);

    return pix->data;
}


INT32  
pixSetData(PIX    *pix,
	   UINT32 *data)
{

    PROCNAME("pixSetData");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);

    pix->data = data;
    return 0;
}


/*--------------------------------------------------------------------*
 *                    Print output for debugging                      *
 *--------------------------------------------------------------------*/
/*!
 *    pixPrintStreamInfo()
 *
 *        Input:  stream
 *                pix
 *                text (an identifying string that is printed)
 *        Return: 0 if OK, 1 on error.
 */
INT32
pixPrintStreamInfo(FILE    *fp,
                   PIX     *pix,
                   char    *text)
{
INT32         ncolors;
PIXCMAP      *cmap;  

    PROCNAME("pixPrintStreamInfo");

    if (!fp)
        return ERROR_INT("fp not defined", procName, 1);
    if (!pix)
        return ERROR_INT("pix not defined", procName, 1);
    if (!text)
        return ERROR_INT("text not defined", procName, 1);

    fprintf(fp, "  Pix Info for %s:\n", text);
    fprintf(fp, "    width = %d, height = %d, depth = %d\n",
               pixGetWidth(pix), pixGetHeight(pix), pixGetDepth(pix));
    fprintf(fp, "    wpl = %d, data = %x, refcount = %d\n",
               pixGetWpl(pix), pixGetData(pix), pixGetRefcount(pix));
    cmap = pixGetColormap(pix);
    if (cmap) {
        if (cmap->type == CM_BMP)
	    ncolors = cmap->size / sizeof(RGBQUAD);
	else     /* cmap->type == CM_GIF */
	    ncolors = cmap->size / sizeof(GIFColorTableEntry);
	fprintf(fp, "    colormap: size = %d bytes; ncolors = %d\n",
	        cmap->size, ncolors);
    }
    else
        fprintf(fp, "    no colormap\n");

    return 0;
}

