/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *  pixarith.c
 *
 *
 *      One-image grayscale arithmetic operations (8 bpp)
 *           PIX        *pixAddConstantGray()
 *
 *      Two-image grayscale arithmetic operations (8 bpp)
 *           PIX        *pixAddGray()
 *           PIX        *pixSubtractGray()
 *
 *      Image accumulator arithmetic operations (onto 16 bpp)
 *           PIX        *pixInit8To16()
 *           INT32       pixAccumulate8To16() 
 *           INT32       pixMultConstant16()
 *           PIX        *pixFinal16To8()
 *           PIX        *pixDivide8Into16() 
 *
 *      Adjustment for maximum dynamic range in 8 bpp image:
 *           PIX        *pixMaxDynamicRange()
 *
 *
 *      The image accumulator operations are used when you expect
 *      overflow from 8 bits on intermediate results.  For example,
 *      you might want a tophat contrast operator which is
 *         3*I - opening(I,S) - closing(I,S)
 *      To use these operations, first use the init to generate
 *      a 16 bpp image, use the accumulate to add or subtract 8 bpp
 *      images from that, or the multiply constant to multiply
 *      by a small constant (much less than 256 -- we don't want
 *      overflow from the 16 bit images!), and when you're finished
 *      use final to bring the result back to 8 bpp, clipped
 *      if necessary.  There is also a divide function, which
 *      can be used to divide one image by another, scaling the
 *      result for maximum dynamic range, and giving back the
 *      8 bpp result.
 *      
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "allheaders.h"



/*-------------------------------------------------------------*
 *          One-image grayscale arithmetic operations          *
 *-------------------------------------------------------------*/
/*!
 *  pixAddConstantGray()
 *
 *          Input:  pixs (8 bpp)
 *                  val  (amount to add to each pixel; this can be < 0)
 *          Return: pixd (result), or null on error
 *
 *          Note: This is only defined for 8 bpp.  It can be used to
 *                add or subtract a constant image, and the the result
 *                for each pixel is clipped to the interval [0, 255]
 *                if necessary.
 */
INT32
pixAddConstantGray(PIX     *pixs,
                   INT32    val)
{
INT32     w, h, wpl;
UINT32   *data;

    PROCNAME("pixAddConstantGray");

    if (!pixs)
	return ERROR_INT("pixs not defined", procName, 1);
    if (pixGetDepth(pixs) != 8)
	return ERROR_INT("pixs not 8 bpp", procName, 1);
    if (ABS(val) > 255)
	return ERROR_INT("|val| must be < 256", procName, 1);

    data = pixGetData(pixs);
    wpl = pixGetWpl(pixs);
    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    addConstantGrayLow(data, w, h, wpl, val);

    return 0;
}

	    

/*-------------------------------------------------------------*
 *             Two-image grayscale arithmetic ops              *
 *-------------------------------------------------------------*/
/*!
 *  pixAddGray()
 *
 *    Input:  pixd (<optional>)
 *            pixs1 (can be == to pixd)
 *            pixs2 (can be == to pixd)
 *    Return:  pixd always
 *    Action:  Arithmetic addition of two 8 bpp images.
 *             The result is pixd = Min(pixs1 + pixs2, 255)
 *             There are 3 cases.  The result can go to a new dest,
 *             in-place to pixs1, or to an existing input dest:
 *              o  pixd == null   (src1 + src2) --> new pixd
 *              o  pixd == pixs1  (src1 + src2) --> src1  (in-place)
 *              o  pixd != pixs1 && pixd != pixs2)
 *                                (src1 + src2) --> input pixd
 */
PIX *
pixAddGray(PIX     *pixd,
           PIX     *pixs1,
           PIX     *pixs2)
{
INT32     w, h, wpls, wpld;
UINT32   *datas, *datad;

    PROCNAME("pixAddGray");

    if (!pixs1)
	return (PIX *)ERROR_PTR("pixs1 not defined", procName, pixd);
    if (!pixs2)
	return (PIX *)ERROR_PTR("pixs2 not defined", procName, pixd);
    if (pixs1 == pixs2)
	return (PIX *)ERROR_PTR("pixs1 and pixs2 must differ", procName, pixd);
    if (!pixSizesEqual(pixs1, pixs2))
	return (PIX *)ERROR_PTR("pixs1 and pixs2 not equal size",
	                         procName, pixd);
    if (pixd) {
	if (!pixSizesEqual(pixs1, pixd))
	    return (PIX *)ERROR_PTR("pixs1 and pixd not equal size",
	                             procName, pixd);
    }
    if (pixGetDepth(pixs1) != 8)
	return (PIX *)ERROR_PTR("pixs1 not 8 bpp", procName, pixd);

    w = pixGetWidth(pixs1);
    h = pixGetHeight(pixs1);

    if (pixs1 != pixd) {
	if ((pixd = pixCopy(pixd, pixs1)) == NULL)
	    return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    }

    datas = pixGetData(pixs2);
    datad = pixGetData(pixd);
    wpls = pixGetWpl(pixs2);
    wpld = pixGetWpl(pixd);

    addGrayLow(datad, w, h, wpld, datas, wpls);

    return pixd;
}


/*!
 *  pixSubtractGray()
 *
 *    Input:  pixd (<optional>)
 *            pixs1 (can be == to pixd)
 *            pixs2 (can be == to pixd)
 *    Return:  pixd always
 *    Action:  Arithmetic subtraction of two 8 bpp images.
 *             Source pixs2 is always subtracted from source pixs1.
 *             The result is pixd = Max(pixs1 - pixs2, 0)
 *             There are 4 cases.  The result can go to a new dest,
 *             in-place to either pixs1 or pixs2, or to an
 *             existing input dest:
 *              o  pixd == null   (src1 - src2) --> new pixd
 *              o  pixd == pixs1  (src1 - src2) --> src1  (in-place)
 *              o  pixd == pixs2  (src1 - src2) --> src2  (in-place)
 *              o  pixd != pixs1 && pixd != pixs2)
 *                                (src1 - src2) --> input pixd
 *             The third case is the one that requires swap = 1.
 *             For the other three, we can arrange for the data in pixs1
 *             to be put in pixd before calling the low-level code.
 */
PIX *
pixSubtractGray(PIX     *pixd,
                PIX     *pixs1,
                PIX     *pixs2)
{
INT32     w, h, wpls, wpld, swap;
UINT32   *datas, *datad;

    PROCNAME("pixSubtractGray");

    if (!pixs1)
	return (PIX *)ERROR_PTR("pixs1 not defined", procName, pixd);
    if (!pixs2)
	return (PIX *)ERROR_PTR("pixs2 not defined", procName, pixd);
    if (pixs1 == pixs2)
	return (PIX *)ERROR_PTR("pixs1 and pixs2 must differ", procName, pixd);
    if (!pixSizesEqual(pixs1, pixs2))
	return (PIX *)ERROR_PTR("pixs1 and pixs2 not equal size",
	                         procName, pixd);
    if (pixd) {
	if (!pixSizesEqual(pixs1, pixd))
	    return (PIX *)ERROR_PTR("pixs1 and pixd not equal size",
	                             procName, pixd);
    }
    if (pixGetDepth(pixs1) != 8)
	return (PIX *)ERROR_PTR("pixs1 not 8 bpp", procName, pixd);

    w = pixGetWidth(pixs1);
    h = pixGetHeight(pixs1);

    if (!pixd || (pixd != pixs1 && pixd != pixs2)) {  /* not in-place */
	if ((pixd = pixCopy(pixd, pixs1)) == NULL)
	    return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    }

    if (pixd != pixs2) {   /* straightforward:  pixd - pixs2 ==> pixd  */
	datas = pixGetData(pixs2);
	datad = pixGetData(pixd);
	wpls = pixGetWpl(pixs2);
	wpld = pixGetWpl(pixd);
	swap = 0;
    }
    else {  /* swap around:  pixs1 - pixd ==> pixd */
	datas = pixGetData(pixs1);
	datad = pixGetData(pixd);
	wpls = pixGetWpl(pixs1);
	wpld = pixGetWpl(pixd);
	swap = 1;
    }
    subtractGrayLow(datad, w, h, wpld, datas, wpls, swap);

    return pixd;
}



/*-------------------------------------------------------------*
 *     Image accumulator arithmetic operations (onto 16 bpp)   *
 *-------------------------------------------------------------*/
/*!
 *  pixInitTo16()
 *
 *          Input:  pixs (8 bpp)
 *          Return: pixd (16 bpp), or null on error
 *
 *          Action: This initializes the 16 bpp image to have the value
 *                  0x8000 and then adds the value of pixs, pixel-wise.
 */
PIX *
pixInit8To16(PIX    *pixs)
{
INT32     w, h, wpls, wpld;
UINT32   *datas, *datad;
PIX      *pixd;

    PROCNAME("pixInit8To16");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 8)
	return (PIX *)ERROR_PTR("pixs not 8 bpp", procName, NULL);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    if ((pixd = pixCreate(w, h, 16)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    datas = pixGetData(pixs);
    datad = pixGetData(pixd);
    wpls = pixGetWpl(pixs);
    wpld = pixGetWpl(pixd);

    init8To16Low(datad, w, h, wpld, datas, wpls);
    return pixd;
}


/*!
 *  pixAccumulate8To16() 
 *
 *          Input:  pixd (16 bpp)
 *                  pixs (8 bpp)
 *                  op  (ADD or SUBTRACT)
 *          Return: 0 if OK; 1 on error
 *
 *          Action: This adds or subtracts each pixd value from pixs
 */
INT32
pixAccumulate8To16(PIX   *pixd,
                   PIX   *pixs,
		   INT32  op)
{
INT32     w, h, wpls, wpld;
UINT32   *datas, *datad;

    PROCNAME("pixAccumulate8To16");

    if (!pixd)
	return ERROR_INT("pixd not defined", procName, 1);
    if (!pixs)
	return ERROR_INT("pixs not defined", procName, 1);
    if (pixGetDepth(pixd) != 16)
	return ERROR_INT("pixd not 16 bpp", procName, 1);
    if (pixGetDepth(pixs) != 8)
	return ERROR_INT("pixs not 8 bpp", procName, 1);
    if (op != ADD && op != SUBTRACT)
	return ERROR_INT("op must be in {ADD, SUBTRACT}", procName, 1);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    if (pixGetWidth(pixd) != w)
	return ERROR_INT("pix widths not equal", procName, 1);
    if (pixGetHeight(pixd) != h)
	return ERROR_INT("pix heights not equal", procName, 1);

    datas = pixGetData(pixs);
    datad = pixGetData(pixd);
    wpls = pixGetWpl(pixs);
    wpld = pixGetWpl(pixd);

    accumulate8To16Low(datad, w, h, wpld, datas, wpls, op);
    return 0;
}


/*!
 *  pixMultConstant16() 
 *
 *          Input:  pixs (16 bpp)
 *                  factor (multiplicative float in range [-255., 255.])
 *          Return: 0 if OK; 1 on error
 *
 *          Action: This multiplies each pixel, relative to INIT_16_VALUE,
 *                  by the integer.
 */
INT32
pixMultConstant16(PIX     *pixs,
		  FLOAT32  factor)
{
INT32     w, h, wpl;
UINT32   *data;

    PROCNAME("pixMultConstant16");

    if (!pixs)
	return ERROR_INT("pixs not defined", procName, 1);
    if (pixGetDepth(pixs) != 16)
	return ERROR_INT("pixs not 16 bpp", procName, 1);
    if (ABS(factor) > 255.)
	return ERROR_INT("|factor| must be <= 255.", procName, 1);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    data = pixGetData(pixs);
    wpl = pixGetWpl(pixs);

    multConstant16Low(data, w, h, wpl, factor);
    return 0;
}


/*!
 *  pixFinal16To8()
 *
 *          Input:  pixs (16 bpp)
 *          Return: pixd (8 bpp), or null on error
 *
 *          Action: This subtracts the value 0x8000 from the src
 *                  16 bpp image and puts the result, clipped to
 *                  [0, 255], in the destination 8 bpp pix.
 */
PIX *
pixFinal16To8(PIX   *pixs)
{
INT32     w, h, wpls, wpld;
UINT32   *datas, *datad;
PIX      *pixd;

    PROCNAME("pixFinal16To8");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 16)
	return (PIX *)ERROR_PTR("pixs not 16 bpp", procName, NULL);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    if ((pixd = pixCreate(w, h, 8)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    datas = pixGetData(pixs);
    datad = pixGetData(pixd);
    wpls = pixGetWpl(pixs);
    wpld = pixGetWpl(pixd);

    final16To8Low(datad, w, h, wpld, datas, wpls);
    return pixd;
}


/*!
 *  pixDivide8Into16() 
 *
 *          Input:  pixs8 (8 bpp)
 *                  pixs16 (16 bpp)
 *          Return: pixd (8 bpp), or null on error
 *
 *          Action: This divides each pixel in pixs16 by the one
 *                  in pixs8, then converts the 16 bpp result
 *                  into an 8 bpp image using maximum dynamic range.
 *
 *          Note:   This uses only "positive" input values in pixs16.
 *                  We first subtract INIT_16_VALUE from each pixel.
 *                  The result should be > 0.  If it is not, we force
 *                  it to be = 0.
 */
PIX *
pixDivide8Into16(PIX   *pixs8,
                 PIX   *pixs16)
{
INT32     w, h, wpls8, wpls16;
UINT32   *datas8, *datas16;
PIX      *pixd;

    PROCNAME("pixDivide8Into16");

    if (!pixs8)
	return (PIX *)ERROR_PTR("pixs8 not defined", procName, NULL);
    if (!pixs16)
	return (PIX *)ERROR_PTR("pixs16 not defined", procName, NULL);
    if (pixGetDepth(pixs8) != 8)
	return (PIX *)ERROR_PTR("pixs8 not 8 bpp", procName, NULL);
    if (pixGetDepth(pixs16) != 16)
	return (PIX *)ERROR_PTR("pixs16 not 16 bpp", procName, NULL);

    w = pixGetWidth(pixs8);
    h = pixGetHeight(pixs8);
    if (pixGetWidth(pixs16) != w)
	return (PIX *)ERROR_PTR("pix widths not equal", procName, NULL);
    if (pixGetHeight(pixs16) != h)
	return (PIX *)ERROR_PTR("pix heights not equal", procName, NULL);

    datas8 = pixGetData(pixs8);
    datas16 = pixGetData(pixs16);
    wpls8 = pixGetWpl(pixs8);
    wpls16 = pixGetWpl(pixs16);

	/* removes bias; divides */
    divide8Into16Low(datas16, w, h, wpls16, datas8, wpls8);

	/* get max dynm range, converting to 8 bpp */
    if ((pixd = pixMaxDynamicRange(pixs16, LINEAR_SCALE)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    return pixd;
}



/*-----------------------------------------------------------------------*
 *          Adjustment for maximum dynamic range in 8 bpp image          *
 *-----------------------------------------------------------------------*/
/*!
 *  pixMaxDynamicRange()
 *
 *   Input:  pixs  (8 or 16 bpp source)
 *           type  (LINEAR_SCALE or LOG_SCALE)
 *   Return: pixd, or null on error
 *   Action: scales pixel values up or down linearly to fit maximally
 *           within the 8 bpp pixd
 *
 */
PIX *
pixMaxDynamicRange(PIX    *pixs,
                   INT32   type)
{
INT32     w, h, d, wpls, wpld;
UINT32   *datas, *datad;
PIX      *pixd;

    PROCNAME("pixMaxDynamicRange");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    d = pixGetDepth(pixs);
    if (d != 8 && d != 16)
        return (PIX *)ERROR_PTR("pixs must be 8 or 16 bpp", procName, NULL);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    if ((pixd = pixCreate(w, h, 8)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    datas = pixGetData(pixs);
    datad = pixGetData(pixd);
    wpls = pixGetWpl(pixs);
    wpld = pixGetWpl(pixd);

    maxDynamicRangeLow(datad, w, h, wpld, datas, d, wpls, type);

    return pixd;
}


