/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/



/*
 *  rop.c
 *
 *      General rasterop
 *           INT32       pixRasterop()
 *
 *      In-place full band rasterop
 *           INT32       pixRasteropVip()
 *           INT32       pixRasteropHip()
 *
 *      In-place shift
 *           INT32       pixRasteropIP()
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "allheaders.h"




/*--------------------------------------------------------------------*
 *               Intermediate-level rasterop interface                *
 *--------------------------------------------------------------------*/
/*!
 *  pixRasterop()
 *
 *       Input: 9 args (standard args for rasterop)
 *              These are
 *                   pixd   (dest pix)
 *                   dx     (x val of UL corner of dest rectangle)
 *                   dy     (y val of UL corner of dest rectangle)
 *                   dw     (width of dest rectangle)
 *                   dh     (height of dest rectangle)
 *                   op     (op code)
 *                   pixs   (src pix)
 *                   sx     (x val of UL corner of src rectangle)
 *                   sy     (y val of UL corner of src rectangle)
 *
 *       Return: 0 if OK; 1 on error.
 *
 *       Action: if the operation involves only dest, calls
 *               rasteropUniLow().  Otherwise, checks depth of the
 *               src and dest, and if they match, calls rasteropLow().
 *
 *       Background:
 *       -----------
 *
 *       There are 18 operations, described by the op codes in
 *       pix.h.
 *
 *       One, PIX_DST, is a no-op.
 *
 *       Three, PIX_CLR, PIX_SET, and PIX_NOT(PIX_DST) operate
 *       only on the dest.  These are handled by the low-level
 *       rasteropUniLow().
 *
 *       The other 14 involve the both the src and the dest, 
 *       and depend on the bit values of either just the src
 *       or the bit values of both src and dest.  They are
 *       handled by rasteropLow():
 *
 *               PIX_SRC                             s
 *               PIX_NOT(PIX_SRC)                   ~s
 *               PIX_SRC | PIX_DST                   s | d
 *               PIX_SRC & PIX_DST                   s & d
 *               PIX_SRC ^ PIX_DST                   s ^ d
 *               PIX_NOT(PIX_SRC) | PIX_DST         ~s | d
 *               PIX_NOT(PIX_SRC) & PIX_DST         ~s & d
 *               PIX_NOT(PIX_SRC) ^ PIX_DST         ~s ^ d
 *               PIX_SRC | PIX_NOT(PIX_DST)          s | ~d
 *               PIX_SRC & PIX_NOT(PIX_DST)          s & ~d
 *               PIX_SRC ^ PIX_NOT(PIX_DST)          s ^ ~d
 *               PIX_NOT(PIX_SRC | PIX_DST)         ~(s | d)
 *               PIX_NOT(PIX_SRC & PIX_DST)         ~(s & d)
 *               PIX_NOT(PIX_SRC ^ PIX_DST)         ~(s ^ d)
 *
 *       Each of these is implemented with one of three low-level
 *       functions, depending on the alignment of the left edge
 *       of the src and dest rectangles:
 *           * a fastest implementation if both left edges are
 *             (32-bit) word aligned
 *           * a very slightly slower implementation if both left
 *             edges have the same relative (32-bit) word alignment
 *           * the general routine that is invoked when
 *             both left edges have different word alignment
 *
 *       Of the 14 binary rasterops above, only 12 are unique
 *       logical combinations (out of a possible 16) of src
 *       and dst bits:
 *
 *             (sd)         (11)   (10)   (01)   (00)
 *        -----------------------------------------------
 *              s            1      1      0      0
 *             ~s            0      1      0      1
 *            s | d          1      1      1      0
 *            s & d          1      0      0      0
 *            s ^ d          0      1      1      0
 *           ~s | d          1      0      1      1
 *           ~s & d          0      0      1      0
 *           ~s ^ d          1      0      0      1
 *            s | ~d         1      1      0      1
 *            s & ~d         0      1      0      0
 *            s ^ ~d         1      0      0      1
 *           ~(s | d)        0      0      0      1
 *           ~(s & d)        0      1      1      1
 *           ~(s ^ d)        1      0      0      1
 *
 *      Note that the following three operations are equivalent:
 *           ~(s ^ d)
 *           ~s ^ d
 *           s ^ ~d
 *      and in the implementation, we call them out with the first form;
 *      namely, ~(s ^ d).
 *
 *      Of the 16 possible binary combinations of src and dest bits,
 *      the remaining 4 unique ones are independent of the src bit.
 *      They depend on either just the dest bit or on neither
 *      the src nor dest bits:
 *              d            1      0      1      0    (indep. of s)
 *             ~d            0      1      0      1    (indep. of s)
 *             CLR           0      0      0      0    (indep. of both s & d)
 *             SET           1      1      1      1    (indep. of both s & d)
 *
 *      As mentioned above, three of these are implemented by
 *      rasteropUniLow(), and one is a no-op.
 *
 *      How can these operation codes be represented by bits
 *      in such a way that when the basic operations are performed
 *      on the bits the results are unique for unique
 *      operations, and mimic the logic table given above?
 *
 *      The answer is to choose a particular order of the pairings:
 *             (sd)         (11)   (10)   (01)   (00)
 *      (which happens to be the same as in the above table)
 *      and to translate the result into 4-bit representations
 *      of s and d.  For example, the Sun rasterop choice
 *      (omitting the extra bit for clipping) is
 *
 *           PIX_SRC      0xc
 *           PIX_DST      0xa
 *
 *      This corresponds to our pairing order given above:
 *             (sd)         (11)   (10)   (01)   (00)
 *      where for s = 1 we get the bit pattern
 *           PIX_SRC:        1      1      0      0     (0xc)
 *      and for d = 1 we get the pattern
 *           PIX_DST:         1      0      1      0    (0xa)
 *
 *      OK, that's the pairing order that Sun chose.  How
 *      many different ways can we assign bit patterns to PIX_SRC
 *      and PIX_DST to get the boolean ops to work out?
 *      Any of the 4 pairs can be put in the first position,
 *      any of the remaining 3 pairs can go in the second;
 *      and one of the remaining 2 pairs can go the the third.
 *      There is a total of 4*3*2 = 24 ways these pairs can
 *      be permuted.
 */
INT32
pixRasterop(PIX    *pixd,
            INT32   dx,
            INT32   dy,
            INT32   dw,
            INT32   dh,
            INT32   op,
            PIX    *pixs,
            INT32   sx,
            INT32   sy)
{
INT32   dd;

    PROCNAME("pixRasterop");

    if (!pixd)
	return ERROR_INT("pixd not defined", procName, 1);

    if (op == PIX_DST)   /* no-op */
	return 0;

	/* check if operation is only on dest */
    dd = pixGetDepth(pixd);
    if (op == PIX_CLR || op == PIX_SET || op == PIX_NOT(PIX_DST)) {
        rasteropUniLow(pixGetData(pixd),
	               pixGetWidth(pixd), pixGetHeight(pixd), dd,
	 	       pixGetWpl(pixd),
		       dx, dy, dw, dh,
		       op);
	return 0;
    }

    if (!pixs)
	return ERROR_INT("pixs not defined", procName, 1);

	/* check depth of src and depth; these must agree */
    if (dd != pixGetDepth(pixs))
	return ERROR_INT("depths of pixs and pixd differ", procName, 1);

    rasteropLow(pixGetData(pixd),
		pixGetWidth(pixd), pixGetHeight(pixd), dd,
		pixGetWpl(pixd),
		dx, dy, dw, dh,
		op,
		pixGetData(pixs),
		pixGetWidth(pixs), pixGetHeight(pixs),
		pixGetWpl(pixs),
		sx, sy);

    return 0;
}


/*--------------------------------------------------------------------*
 *                     In-place full band rasterop                    *
 *--------------------------------------------------------------------*/
/*!
 *  pixRasteropVip()
 *
 *         Input:   pixd (in-place)
 *                  x    (left edge of vertical band)
 *                  w    (width of vertical band)
 *                  vshift (vertical shift of band; vshift > 0 is down)
 *
 *         Return: 0 if OK; 1 on error
 */
INT32
pixRasteropVip(PIX    *pixd,
               INT32   x,
               INT32   w,
               INT32   vshift)
{

    PROCNAME("pixRasteropVip");

    if (!pixd)
	return ERROR_INT("pixd not defined", procName, 1);

    rasteropVipLow(pixGetData(pixd),
                   pixGetWidth(pixd), pixGetHeight(pixd),
		   pixGetDepth(pixd), pixGetWpl(pixd),
		   x, w, vshift);

    return 0;
}


/*!
 *  pixRasteropHip()
 *
 *         Input:   pixd (in-place)
 *                  y    (top of horizontal band)
 *                  h    (height of horizontal band)
 *                  hshift (horizontal shift of band; hshift > 0 is to right)
 *
 *         Return: 0 if OK; 1 on error
 *
 *         Note:  we need to set the pad bits to 0 for left shifts
 */
INT32
pixRasteropHip(PIX    *pixd,
               INT32   y,
               INT32   h,
               INT32   hshift)
{
INT32       height;

    PROCNAME("pixRasteropHip");

    if (!pixd)
	return ERROR_INT("pixd not defined", procName, 1);

    if (pixGetDepth(pixd) == 1)
	pixSetPadBitsBand(pixd, y, h, 0);

    rasteropHipLow(pixGetData(pixd), pixGetHeight(pixd),
		   pixGetDepth(pixd), pixGetWpl(pixd),
		   y, h, hshift);

    return 0;
}


/*--------------------------------------------------------------------*
 *                            In-place shift                          *
 *--------------------------------------------------------------------*/
/*!
 *  pixRasteropIP()
 *
 *         Input:   pixd (in-place)
 *                  hshift (horizontal shift; hshift > 0 is to right)
 *                  vshift (vertical shift; vshift > 0 is down)
 *
 *         Return: 0 if OK; 1 on error
 */
INT32
pixRasteropIP(PIX    *pixd,
              INT32   hshift,
              INT32   vshift)
{
INT32   w, h;

    PROCNAME("pixRasteropIP");

    if (!pixd)
	return ERROR_INT("pixd not defined", procName, 1);

    w = pixGetWidth(pixd);
    h = pixGetHeight(pixd);
    pixRasteropHip(pixd, 0, h, hshift);
    pixRasteropVip(pixd, 0, w, vshift);

    return 0;
}


