/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *  rotateorth.c
 *
 *      180-degree rotation
 *            PIX     *pixRotate180()
 *
 *       90-degree rotation (both directions)
 *            PIX     *pixRotate90()
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "allheaders.h"



/*!
 *  pixRotate180()
 *
 *     Input:  pixd  (<optional>; can be equal to pixs)
 *             pixs
 *     Return: pixd
 */
PIX *
pixRotate180(PIX    *pixd,
             PIX    *pixs)
{
UINT8    *tab;
INT32     w, h, d, wplt, wpld;
UINT32   *datat, *datad;
PIX      *pixt;

    PROCNAME("pixRotate180");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, pixd);

    d = pixGetDepth(pixs);
    if (d != 1 && d != 2 && d != 4 && d != 8 && d != 16 && d != 32)
	return (PIX *)ERROR_PTR("pixs not in {1,2,4,8,,16,32} bpp",
	       procName, pixd);

    if (!pixd) {
	if ((pixd = pixCreateTemplate(pixs)) == NULL)
	    return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
	pixt = pixClone(pixs);
    }
    else {
	if (!pixSizesEqual(pixs, pixd))
	    return (PIX *)ERROR_PTR("pix sizes unequal", procName, pixd);

	if (pixs == pixd) {
	    if ((pixt = pixCopy(NULL, pixs)) == NULL)
		return (PIX *)ERROR_PTR("pixt not made", procName, pixd);
	}
	else {
	    pixt = pixClone(pixs);
	}

	    /* we're not fully overwriting pixd */
	pixClearAll(pixd);
    }

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    datat = pixGetData(pixt);
    wplt = pixGetWpl(pixt);
    datad = pixGetData(pixd);
    wpld = pixGetWpl(pixd);

    if ((tab = makeReverseByteTab()) == NULL)
	return (PIX *)ERROR_PTR("tab not made", procName, pixd);

    rotate180low(datad, w, h, d, wpld, datat, wplt, tab);

    FREE((void *)tab);

    pixDestroy(&pixt);

    return pixd;
}
    

/*!
 *  pixRotate90()
 *
 *     Input:  pixs
 *             direction (1 = clockwise,  -1 = counter-clockwise)
 *     Return: pixd, or null on error
 */
PIX *
pixRotate90(PIX    *pixs,
            INT32   direction)
{
INT32     wd, hd, d, wpls, wpld;
UINT32   *datas, *datad;
PIX      *pixd;

    PROCNAME("pixRotate90");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, pixd);

    d = pixGetDepth(pixs);
    if (d != 1 && d != 2 && d != 4 && d != 8 && d != 16 && d != 32)
	return (PIX *)ERROR_PTR("pixs not in {1,2,4,8,,16,32} bpp",
	       procName, pixd);

    hd = pixGetWidth(pixs);
    wd = pixGetHeight(pixs);
    if ((pixd = pixCreate(wd, hd, d)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    pixCopyColormap(pixd, pixs);

    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);
    datad = pixGetData(pixd);
    wpld = pixGetWpl(pixd);

    rotate90low(datad, wd, hd, d, wpld, datas, wpls, direction);

    return pixd;
}


