/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *  rotateorthlow.c
 *
 *      180-degree rotation
 *            void      rotate180low()
 *            UINT8    *makeReverseByteTab()
 *
 *      90-degree rotation (cw)
 *            void      rotate90low()
 */

#include <stdio.h>
#include <stdlib.h>

#include "allheaders.h"



/*------------------------------------------------------------------*
 *                          180 degree rotation                     *
 *------------------------------------------------------------------*/
/*!
 *  rotate180low()
 *
 *     Action: The pixel access routines allow a trivial
 *             implementation.  However, for d < 8,
 *             it is more efficient to right-justify each line
 *             to the 32-bit boundary and then extract bytes and
 *             do pixel reversing.   This becomes most important
 *             for 1 bpp images, so for them, we right shift the
 *             source data (if necessary) to right-justify on the
 *             32 bit boundary, then read the bytes off each
 *             raster line in reverse order and reverse
 *             the bit (pixel) order in each byte.  At the end,
 *             we left shift (if necessary) the source image
 *             back so that it is again left-justified.
 *
 *       Note: The dest must be cleared in advance because not
 *             all source pixels are written to the destination.
 */
void
rotate180low(UINT32   *datad,
             INT32     w,
             INT32     h,
             INT32     d,
             INT32     wpld,
             UINT32   *datas, 
             INT32     wpls,
	     UINT8    *tab)
{
INT32       extra, shift, databpl, bpl, i, j;
UINT32      val;
UINT32     *lines, *lined;

    PROCNAME("rotate180low");

    switch (d)
    {
	case 32: 
	    for (i = 0; i < h; i++) {
		lines = datas + (h - 1 - i) * wpls;
		lined = datad + i * wpld;
		for (j = 0; j < w; j++)
		    lined[j] = lines[w - 1 - j];
	    }
	    break;
	case 16:
	    for (i = 0; i < h; i++) {
		lines = datas + (h - 1 - i) * wpls;
		lined = datad + i * wpld;
		for (j = 0; j < w; j++) {
		    if (val = GET_DATA_TWO_BYTES(lines, w - 1 - j))
			SET_DATA_TWO_BYTES(lined, j, val);
		}
	    }
	    break;
	case 8:
	    for (i = 0; i < h; i++) {
		lines = datas + (h - 1 - i) * wpls;
		lined = datad + i * wpld;
		for (j = 0; j < w; j++) {
		    if (val = GET_DATA_BYTE(lines, w - 1 - j))
			SET_DATA_BYTE(lined, j, val);
		}
	    }
	    break;
	case 4:
	    for (i = 0; i < h; i++) {
		lines = datas + (h - 1 - i) * wpls;
		lined = datad + i * wpld;
		for (j = 0; j < w; j++) {
		    if (val = GET_DATA_QBIT(lines, w - 1 - j))
			SET_DATA_QBIT(lined, j, val);
		}
	    }
	    break;
	case 2:
	    for (i = 0; i < h; i++) {
		lines = datas + (h - 1 - i) * wpls;
		lined = datad + i * wpld;
		for (j = 0; j < w; j++) {
		    if (val = GET_DATA_DIBIT(lines, w - 1 - j))
			SET_DATA_DIBIT(lined, j, val);
		}
	    }
	    break;
	case 1:
	    extra = (w * d) & 31;
	    if (extra)
		shift = 32 - extra;
	    else
		shift = 0;
	    if (shift)
		rasteropHipLow(datas, h, d, wpls, 0, h, shift);

	    databpl = (w + 7) / 8;
	    bpl = 4 * wpls;
	    for (i = 0; i < h; i++) {
		lines = datas + (h - 1 - i) * wpls;
		lined = datad + i * wpld;
		for (j = 0; j < databpl; j++) {
		    if (val = GET_DATA_BYTE(lines, bpl - 1 - j))
			SET_DATA_BYTE(lined, j, tab[val]);
		}
	    }
	    
	    if (shift)
		rasteropHipLow(datas, h, d, wpls, 0, h, -shift);

	    break;
	default:
	    ERROR_VOID("depth not permitted for rot 180", procName);
	    return;
    }

    return;
}


/*!
 *  makeReverseByteTab()
 */
UINT8 *
makeReverseByteTab(void)
{
INT32    i;
UINT8   *tab;

    PROCNAME("makeReverseByteTab");

    if ((tab = (UINT8 *)CALLOC(256, sizeof(UINT8))) == NULL)
	return (UINT8 *)ERROR_PTR("calloc fail for tab", procName, NULL);

    for (i = 0; i < 256; i++)
	tab[i] = ((0x80 & i) >> 7) |
	         ((0x40 & i) >> 5) |
	         ((0x20 & i) >> 3) |
	         ((0x10 & i) >> 1) |
	         ((0x08 & i) << 1) |
	         ((0x04 & i) << 3) |
	         ((0x02 & i) << 5) |
	         ((0x01 & i) << 7);

    return tab;
}



/*------------------------------------------------------------------*
 *                           90 degree rotation                     *
 *------------------------------------------------------------------*/
/*!
 *  rotate90low()
 *
 *       direction:  1 for cw rotation
 *                  -1 for ccw rotation
 *
 *       Note: The dest must be cleared in advance because not
 *             all source pixels are written to the destination.
 */
void
rotate90low(UINT32   *datad,
            INT32     wd,
            INT32     hd,
            INT32     d,
            INT32     wpld,
            UINT32   *datas, 
            INT32     wpls,
	    INT32     direction)
{
INT32       extra, shift, i, j, k, m, iend, nswords;
UINT32      val, word;
UINT32     *lines, *lined;

    PROCNAME("rotate90low");

    if (direction == 1) {  /* clockwise */
	switch (d)
	{
	    case 32: 
		for (i = 0; i < hd; i++) {
		    lined = datad + i * wpld;
		    lines = datas + (wd - 1) * wpls;
		    for (j = 0; j < wd; j++) {
			lined[j] = lines[i];
			lines -= wpls;
		    }
		}
		break;
	    case 16:
		for (i = 0; i < hd; i++) {
		    lined = datad + i * wpld;
		    lines = datas + (wd - 1) * wpls;
		    for (j = 0; j < wd; j++) {
			if (val = GET_DATA_TWO_BYTES(lines, i))
			    SET_DATA_TWO_BYTES(lined, j, val);
			lines -= wpls;
		    }
		}
		break;
	    case 8:
		for (i = 0; i < hd; i++) {
		    lined = datad + i * wpld;
		    lines = datas + (wd - 1) * wpls;
		    for (j = 0; j < wd; j++) {
			if (val = GET_DATA_BYTE(lines, i))
			    SET_DATA_BYTE(lined, j, val);
			lines -= wpls;
		    }
		}
		break;
	    case 4:
		for (i = 0; i < hd; i++) {
		    lined = datad + i * wpld;
		    lines = datas + (wd - 1) * wpls;
		    for (j = 0; j < wd; j++) {
			if (val = GET_DATA_QBIT(lines, i))
			    SET_DATA_QBIT(lined, j, val);
			lines -= wpls;
		    }
		}
		break;
	    case 2:
		for (i = 0; i < hd; i++) {
		    lined = datad + i * wpld;
		    lines = datas + (wd - 1) * wpls;
		    for (j = 0; j < wd; j++) {
			if (val = GET_DATA_DIBIT(lines, i))
			    SET_DATA_DIBIT(lined, j, val);
			lines -= wpls;
		    }
		}
		break;
	    case 1:
		nswords = hd / 32;
		for (j = 0; j < wd; j++) {
		    lined = datad;
		    lines = datas + (wd - 1 - j) * wpls;
		    for (k = 0; k < nswords; k++) {
		        word = lines[k];
			if (!word) {
			    lined += 32 * wpld;
			    continue;
			}
			else {
			    iend = 32 * (k + 1);
			    for (m = 0, i = 32 * k; i < iend; i++, m++) {
				if ((word << m) & 0x80000000)
				    SET_DATA_BIT(lined, j);
				lined += wpld;
			    }
			}
		    }
		    for (i = 32 * nswords; i < hd; i++) {
			if (GET_DATA_BIT(lines, i))
			    SET_DATA_BIT(lined, j);
			lined += wpld;
		    }
		}
		break;
	    default:
		ERROR_VOID("illegal depth", procName);
	}
    }
    else  {     /* direction counter-clockwise */
	switch (d)
	{
	    case 32: 
		for (i = 0; i < hd; i++) {
		    lined = datad + i * wpld;
		    lines = datas;
		    for (j = 0; j < wd; j++) {
			lined[j] = lines[hd - 1 - i];
			lines += wpls;
		    }
		}
		break;
	    case 16:
		for (i = 0; i < hd; i++) {
		    lined = datad + i * wpld;
		    lines = datas;
		    for (j = 0; j < wd; j++) {
			if (val = GET_DATA_TWO_BYTES(lines, hd - 1 - i))
			    SET_DATA_TWO_BYTES(lined, j, val);
			lines += wpls;
		    }
		}
		break;
	    case 8:
		for (i = 0; i < hd; i++) {
		    lined = datad + i * wpld;
		    lines = datas;
		    for (j = 0; j < wd; j++) {
			if (val = GET_DATA_BYTE(lines, hd - 1 - i))
			    SET_DATA_BYTE(lined, j, val);
			lines += wpls;
		    }
		}
		break;
	    case 4:
		for (i = 0; i < hd; i++) {
		    lined = datad + i * wpld;
		    lines = datas;
		    for (j = 0; j < wd; j++) {
			if (val = GET_DATA_QBIT(lines, hd - 1 - i))
			    SET_DATA_QBIT(lined, j, val);
			lines += wpls;
		    }
		}
		break;
	    case 2:
		for (i = 0; i < hd; i++) {
		    lined = datad + i * wpld;
		    lines = datas;
		    for (j = 0; j < wd; j++) {
			if (val = GET_DATA_DIBIT(lines, hd - 1 - i))
			    SET_DATA_DIBIT(lined, j, val);
			lines += wpls;
		    }
		}
		break;
	    case 1:
		nswords = hd / 32;
		for (j = 0; j < wd; j++) {
		    lined = datad + (hd - 1) * wpld;
		    lines = datas + (wd - 1 - j) * wpls;
		    for (k = 0; k < nswords; k++) {
		        word = lines[k];
			if (!word) {
			    lined -= 32 * wpld;
			    continue;
			}
			else {
			    iend = 32 * (k + 1);
			    for (m = 0, i = 32 * k; i < iend; i++, m++) {
				if ((word << m) & 0x80000000)
				    SET_DATA_BIT(lined, wd - 1 - j);
				lined -= wpld;
			    }
			}
		    }
		    for (i = 32 * nswords; i < hd; i++) {
			if (GET_DATA_BIT(lines, i))
			    SET_DATA_BIT(lined, wd - 1 - j);
			lined -= wpld;
		    }
		}
		break;
	    default:
		ERROR_VOID("illegal depth", procName);
	}
    }

    return;
}


