/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *  rotateshear.c
 *
 *      Shear rotation about arbitrary point using 2 and 3 shears
 *
 *              PIX        *pixRotateShear()
 *              PIX        *pixRotate2Shear()
 *              PIX        *pixRotate3Shear()
 *
 *      Shear rotation in-place about arbitrary point using 3 shears
 *              INT32       pixRotateShearIP()
 *
 *      Shear rotation around the image center
 *              PIX        *pixRotateShearCenter()    (2 or 3 shears)
 *              INT32       pixRotateShearCenterIP()  (3 shears)
 *
 *      Pointwise Euclidean Rotation
 *              PIX        *pixRotateEuclidean()   [deprecated: too slow]
 *
 *
 *      Rotation is measured in radians; clockwise rotations
 *      are positive.
 *
 *      Rotation by shear works on images of any depth, 
 *      including 8 bpp color paletted images and 24 bpp
 *      rgb images.  It works by translating each src pixel
 *      value to the appropriate pixel in the rotated dest.
 *      For 8 bpp grayscale images, it is about 10-15x faster
 *      than rotation by area-mapping.
 *
 *      This speed and flexibility comes at the following cost,
 *      relative to area-mapped rotation:
 *
 *        -  Jaggies are created on edges of straight lines
 *
 *        -  For large angles, where you must use 3 shears,
 *           there is some extra clipping from the shears.
 *
 *      For small angles, typically less than 0.05 radians,
 *      rotation can be done with 2 orthogonal shears.
 *      Two such continuous shears (as opposed to the discrete
 *      shears on a pixel lattice that we have here) give
 *      a rotated image that has a distortion in the lengths
 *      of the two rotated and still-perpendicular axes.  The
 *      length/width ratio changes by a fraction 
 *
 *           0.5 * (angle)**2
 *
 *      For an angle of 0.05 radians, this is about 1 part in
 *      a thousand.  This distortion is absent when you use
 *      3 continuous shears with the correct angles (see below).
 *
 *      Of course, the image is on a discrete pixel lattice.
 *      Rotation by shear gives an approximation to a continuous
 *      rotation, leaving pixel jaggies at sharp boundaries.
 *      For very small rotations, rotating from a corner gives
 *      better sensitivity than rotating from the image center.
 *      Here's why.  Define the shear "center" to be the line such
 *      that the image is sheared in opposite directions on
 *      each side of and parallel to the line.  For small
 *      rotations there is a "dead space" on each side of the
 *      shear center of width equal to half the shear angle,
 *      in radians.  Thus, when the image is sheared about the center,
 *      the dead space width equals the shear angle, but when
 *      the image is sheared from a corner, the dead space
 *      width is only half the shear angle.
 *
 *      All horizontal and vertical shears are implemented by
 *      rasterop.  The in-place rotation uses special in-place
 *      shears that copy rows sideways or columns vertically
 *      without buffering, and that erase old pixels that are
 *      no longer covered by sheared pixels.  (Stated differently,
 *      it is as if the rows were padded on the outside with OFF
 *      pixels that move into the image when the row is sheared,
 *      and similarly for columns.)
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "allheaders.h"

#define  VERY_SMALL_ANGLE       0.001    /* radians; approx. 0.05 degrees */
#define  MAX_2_SHEAR_ANGLE      0.05     /* radians; approx. 3 degrees */



/*------------------------------------------------------------------*
 *                Rotations about an arbitrary point                *
 *------------------------------------------------------------------*/
/*!
 *  pixRotateShear()
 *
 *         Input: pixs
 *                x, y
 *                angle (radians)
 *         Return:  pixd, or null on error.
 *
 *         Note: this rotates a binary image about the given point,
 *               using either 2 or 3 shears.
 *               A positive angle gives a clockwise rotation.
 *
 */
PIX *
pixRotateShear(PIX      *pixs,
	       INT32     x,
	       INT32     y,
	       FLOAT32   angle)
{

    PROCNAME("pixRotateShear");

    if (!pixs)
        return (PIX *)(PIX *)ERROR_PTR("pixs not defined", procName, NULL);

    if (ABS(angle) < VERY_SMALL_ANGLE)
	return pixClone(pixs);

    if (ABS(angle) <= MAX_2_SHEAR_ANGLE)
	return pixRotate2Shear(pixs, x, y, angle);
    else
	return pixRotate3Shear(pixs, x, y, angle);

}


/*!
 *  pixRotate2Shear()
 *
 *         Input: pixs
 *                x, y
 *                angle (radians)
 *         Return:  pixd, or null on error.
 *
 *         Note: this rotates the image about the given point,
 *               using the 2-shear method.  It should only
 *               be used for angles smaller than MAX_2_SHEAR_ANGLE.
 *               A positive angle gives a clockwise rotation.
 *               2-shear rotation by a specified angle is equivalent
 *               to the sequential transformations
 *                    x' = x + tan(angle) * y      for x-shear
 *                    y' = y + tan(angle) * x      for y-shear
 *               Note that the tan(angle) is performed within
 *               the shear operation.
 */
PIX *
pixRotate2Shear(PIX     *pixs,
		INT32    x,
		INT32    y,
	        FLOAT32  angle)
{
PIX    	     *pixt, *pixd;

    PROCNAME("pixRotate2Shear");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    
    if ((pixt = pixHShear(NULL, pixs, y, angle)) == NULL)
        return (PIX *)ERROR_PTR("pixt not made", procName, NULL);
    if ((pixd = pixVShear(NULL, pixt, x, angle)) == NULL)
        return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    pixDestroy(&pixt);

    return pixd;
}


/*!
 *  pixRotate3Shear()
 *
 *         Input: pixs
 *                x, y
 *                angle (radians)
 *         Return:  pixd, or null on error.
 *
 *         Note: this rotates the image about the image center,
 *               using the 3-shear method.  It should be used
 *               for angles larger than MAX_2_SHEAR_ANGLE.
 *               A positive angle gives a clockwise rotation.
 *               3-shear rotation by a specified angle is equivalent
 *               to the sequential transformations
 *                  y' = y + tan(angle/2) * x      for first y-shear
 *                  x' = x + sin(angle) * y        for x-shear
 *                  y' = y + tan(angle/2) * x      for second y-shear
 *               Note that tan(angle) is performed in the shear
 *               operations.
 */
PIX *
pixRotate3Shear(PIX     *pixs,
		INT32    x,
		INT32    y,
	        FLOAT32  angle)
{
FLOAT32       hangle;
PIX    	     *pixt, *pixd;

    PROCNAME("pixRotate3Shear");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    
    hangle = atan(sin(angle));
    if ((pixd = pixVShear(NULL, pixs, x, angle / 2.)) == NULL)
        return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    if ((pixt = pixHShear(NULL, pixd, y, hangle)) == NULL)
        return (PIX *)ERROR_PTR("pixt not made", procName, NULL);
    pixVShear(pixd, pixt, x, angle / 2.);
    pixDestroy(&pixt);

    return pixd;
}


/*------------------------------------------------------------------*
 *             Rotations in-place about an arbitrary point          *
 *------------------------------------------------------------------*/
/*!
 *  pixRotateShearIP()
 *
 *         Input: pixs
 *                x, y
 *                angle (radians)
 *         Return:  0 if OK; 1 on error.
 *
 *         Note: this does an in-place rotation of the image
 *               about the image center, using the 3-shear method.
 *               (It is useful because the regular (not InPlace)
 *               version requires 3 images in memory.  For 24-bit
 *               color, represented by 32 bit pixels, three page 
 *               images at 300 ppi take about 100 MB.)
 *               A positive angle gives a clockwise rotation.
 *               3-shear rotation by a specified angle is equivalent
 *               to the sequential transformations
 *                  y' = y + tan(angle/2) * x      for first y-shear
 *                  x' = x + sin(angle) * y        for x-shear
 *                  y' = y + tan(angle/2) * x      for second y-shear
 *               Note that tan(angle) is performed in the shear
 *               operations.
 */
INT32
pixRotateShearIP(PIX     *pixs,
		 INT32    x,
		 INT32    y,
	         FLOAT32  angle)
{
FLOAT32       hangle;

    PROCNAME("pixRotateShearIP");

    if (!pixs)
        return ERROR_INT("pixs not defined", procName, 1);
    
    hangle = atan(sin(angle));
    pixHShearIP(pixs, y, angle / 2.);
    pixVShearIP(pixs, x, hangle);
    pixHShearIP(pixs, y, angle / 2.);

    return 0;
}


/*------------------------------------------------------------------*
 *                    Rotations about the image center              *
 *------------------------------------------------------------------*/
/*!
 *   pixRotateShearCenter()
 *
 *          Input:  pixs
 *                  angle (radians)
 *          Return: pixd, or null on error
 */
PIX *
pixRotateShearCenter(PIX     *pixs,
	             FLOAT32  angle)
{

    PROCNAME("pixRotateShearCenter");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    
    return pixRotateShear(pixs, pixGetWidth(pixs) / 2,
                          pixGetHeight(pixs) / 2, angle);
}


/*!
 *   pixRotateShearCenterIP()
 *
 *          Input:  pixs
 *                  angle (radians)
 *          Return: pixd, or null on error
 */
INT32
pixRotateShearCenterIP(PIX     *pixs,
	               FLOAT32  angle)
{

    PROCNAME("pixRotateShearCenterIP");

    if (!pixs)
        return ERROR_INT("pixs not defined", procName, 1);
    
    return pixRotateShearIP(pixs, pixGetWidth(pixs) / 2,
                            pixGetHeight(pixs) / 2, angle);
}


/*------------------------------------------------------------------*
 *                   Pointwise Euclidean Rotation                   *
 *------------------------------------------------------------------*/
/*
 *   We include this function to allow you to compare the
 *   fast 2-shear and 3-shear results with this very slow but
 *   slightly more accurate function.  The difference in
 *   accuracy when compared to 3-shear and (for angles
 *   less than 2 degrees, when compared to 2-shear) is
 *   less than 1 pixel at any point, so in nearly all cases it
 *   does NOT make sense to use this function.  For a 2 degree
 *   rotation, this function is about 50x slower than 2-shear
 *   rotation and about 20x slower than a 3-shear rotation.
 *
 *   There has been some work on what is called a "quasishear
 *   rotation" ("The Quasi-Shear Rotation, Eric Andres,
 *   DGCI 1996, pp. 307-314).  I believe they use a 3-shear
 *   approximation to the continuous rotation, exactly as
 *   we do here.  The approximation is due to being on
 *   a square pixel lattice.  They also use integers to specify
 *   the rotation angle and center offset, but that makes
 *   little sense on a machine where you have a few GFLOPS
 *   and only a few hundred floating point operations to do (!) 
 *   They also allow subpixel specification of the center of
 *   rotation, which I haven't bothered with, and claim that
 *   better results are possible if each of the 4 quadrants is
 *   handled separately.
 * 
 *   But the bottom line is that for binary images, the quality
 *   of the simple 3-shear rotation is about as good as you can do,
 *   visually, without dithering the result.  The effect of dither
 *   is  to break up the horizontal and vertical shear lines.
 *   It's a bit tricky to dither with block shears -- you have to
 *   dither the pixels on the block boundaries -- and I leave
 *   it as an exercise for you!  (Send me your implementation and
 *   I'll include it in leptonica.)  So, here's the simple Euclidean
 *   rotation function.  It takes about 125 machine cycles/pixel,
 *   independent of the rotation angle.  Ugh!
 */

/*!
 *    pixRotateEuclidean()
 *
 *      Input:  pix (1 bpp; can easily be generalized, but why bother?)
 *              xcen, ycen  (center of rotation)
 *              angle (in radians)
 *      Return: rotated pix, or null on error
 */
PIX *
pixRotateEuclidean(PIX     *pixs,
                   INT32    xcen,
		   INT32    ycen,
	           FLOAT32  angle)
{
INT32         i, j, xdif, ydif, xp, yp;
INT32         w, h, wm1, hm1, wpls, wpld;
UINT32       *datas, *datad, *lines, *lined;
FLOAT32       sina, cosa;
PIX    	     *pixd;

    PROCNAME("pixRotateEuclidean");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 1)
        return (PIX *)ERROR_PTR("pixs must be 8 bpp", procName, NULL);

    if (ABS(angle) < VERY_SMALL_ANGLE)
	return pixClone(pixs);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);
    pixd = pixCreateTemplate(pixs);
    datad = pixGetData(pixd);
    wpld = pixGetWpl(pixd);

    wm1 = w - 1;
    hm1 = h - 1;
    xcen = w / 2;
    ycen = h / 2;
    sina = sin(angle);
    cosa = cos(angle);

    for (i = 0; i < h; i++) {
	ydif = ycen - i;
	lined = datad + i * wpld;
	for (j = 0; j < w; j++) {
	    xdif = xcen - j;
	    xp = xcen + (INT32)(-xdif * cosa - ydif * sina + 0.5);
	    yp = ycen + (INT32)(-ydif * cosa + xdif * sina + 0.5);

		/* if off the edge, just write white */
	    if (xp < 0 || yp < 0 || xp > wm1 || yp > hm1) {
		CLEAR_DATA_BIT(lined, j);
		continue;
	    }

	    lines = datas + yp * wpls;
	    if (GET_DATA_BIT(lines, xp))
		SET_DATA_BIT(lined, j);
	}
    }

    return pixd;
}


