/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *  scale.c
 *
 *         Top-level scaling
 *               PIX        *pixScale()
 *
 *         24-bit Color (linearly interpolated) scaling
 *               PIX        *pixScaleColorLI()
 *               PIX        *pixScaleColor2xLI()
 *               PIX        *pixScaleColor4xLI()
 *
 *         Grayscale (linearly interpolated) scaling
 *               PIX        *pixScaleGrayLI()
 *               PIX        *pixScaleGray2xLI()
 *               PIX        *pixScaleGray4xLI()
 *
 *         General scaling by closest pixel sampling
 *               PIX        *pixScaleBySampling()
 *
 *         Binary scaling by closest pixel sampling
 *               PIX        *pixScaleBinary()
 *
 *         Scale-to-gray (3x)
 *               PIX        *pixScaleToGray3()
 *
 *         Scale-to-gray (4x)
 *               PIX        *pixScaleToGray4()
 *
 *         Scale 2x followed by binarization
 *               PIX        *pixScaleGray2xLIThresh()
 *               PIX        *pixScaleGray2xLIDither()
 *
 *         Scale 4x followed by binarization
 *               PIX        *pixScaleGray4xLIThresh()
 *               PIX        *pixScaleGray4xLIDither()
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "allheaders.h"



/*------------------------------------------------------------------*
 *                    Top level scaling dispatcher                  *
 *------------------------------------------------------------------*/
/*
 *  pixScale()
 *
 *         Input: pixs
 *                scalex, scaley
 *         Return:  pixd, or null on error
 * 
 *     This function scales 24 bpp full color; 2, 4 or 8 bpp palette color;
 *     8 bpp gray, and binary images.
 *
 *     When the input has palette color, the colormap is removed and
 *     a 24 bpp full color pix is made, which is then scaled with
 *     pixScaleColorLI().
 *
 *     Color and gray images are scaled using linear interpolation;
 *     binary image are scaled by sampling the closest pixel, without
 *     any low-pass filtering (averaging of neighboring pixels).
 *
 *     Note: see notes below regarding use of filtering when
 *           subsampling
 */
PIX *
pixScale(PIX      *pixs,
	 FLOAT32   scalex,
	 FLOAT32   scaley)
{
INT32         d;
PIX          *pixt, *pixd;

    PROCNAME("pixScale");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);

    d = pixGetDepth(pixs);
    
    if (d == 32) {
        if (scalex == 2.0 && scaley == 2.0)
	    return pixScaleColor2xLI(pixs);
        else if (scalex == 4.0 && scaley == 4.0)
	    return pixScaleColor4xLI(pixs);
	else
	    return pixScaleColorLI(pixs, scalex, scaley);
    }
    else if (d == 1)
	return pixScaleBinary(pixs, scalex, scaley);
    else if (pixGetColormap(pixs)) {   /* 2, 4 or 8 bpp */
	pixt = pixRemoveColormap(pixs, CONVERT_TO_FULL_COLOR);
	pixd = pixScaleColorLI(pixt, scalex, scaley);
/*	pixt = pixRemoveColormap(pixs, CONVERT_TO_GRAYSCALE);
	pixd = pixScaleGrayLI(pixt, scalex, scaley);  */
	pixDestroy(&pixt);
	return pixd;
    }
    else if (d == 8) {
        if (scalex == 2.0 && scaley == 2.0)
	    return pixScaleGray2xLI(pixs);
        else
	    return pixScaleGrayLI(pixs, scalex, scaley);
    }
    else
        return (PIX *)ERROR_PTR("pixs must be {1,8,32} bpp", procName, NULL);

}


/*------------------------------------------------------------------*
 *            24-bit color scaling by linear interpolation          *
 *------------------------------------------------------------------*/
/*
 *  pixScaleColorLI()
 *
 *         Input: pixs  (32 bpp, representing 24 bpp 3-color)
 *                scalex, scaley
 *         Return:  pixd, or null on error
 */
PIX *
pixScaleColorLI(PIX     *pixs,
	       FLOAT32   scalex,
	       FLOAT32   scaley)
{
PIX          *pixRed, *pixGreen, *pixBlue;
PIX          *pixRedScale, *pixGreenScale, *pixBlueScale;
PIX    	     *pixd;

    PROCNAME("pixScaleColorLI");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 32)
        return (PIX *)ERROR_PTR("pixs must be 32 bpp", procName, NULL);

    pixRed = pixGetRGBComponent(pixs, COLOR_RED);
    pixRedScale = pixScaleGrayLI(pixRed, scalex, scaley);
    pixDestroy(&pixRed);
    pixGreen = pixGetRGBComponent(pixs, COLOR_GREEN);
    pixGreenScale = pixScaleGrayLI(pixGreen, scalex, scaley);
    pixDestroy(&pixGreen);
    pixBlue = pixGetRGBComponent(pixs, COLOR_BLUE);
    pixBlueScale = pixScaleGrayLI(pixBlue, scalex, scaley);
    pixDestroy(&pixBlue);

    if ((pixd = pixCreateRGBImage(pixRedScale, pixGreenScale, pixBlueScale))
            == NULL)
        return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    pixDestroy(&pixRedScale);
    pixDestroy(&pixGreenScale);
    pixDestroy(&pixBlueScale);

    return pixd;
}


/*
 *  pixScaleColor2xLI()
 *
 *      Input: pixs  (32 bpp, representing 24 bpp 3-color)
 *      Return:  pixd, or null on error
 *
 *      Special fast version for scaling up by 2x
 */
PIX *
pixScaleColor2xLI(PIX   *pixs)
{
PIX          *pixRed, *pixGreen, *pixBlue;
PIX          *pixRedScale, *pixGreenScale, *pixBlueScale;
PIX    	     *pixd;

    PROCNAME("pixScaleColor2xLI");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 32)
        return (PIX *)ERROR_PTR("pixs must be 32 bpp", procName, NULL);

    pixRed = pixGetRGBComponent(pixs, COLOR_RED);
    pixRedScale = pixScaleGray2xLI(pixRed);
    pixDestroy(&pixRed);
    pixGreen = pixGetRGBComponent(pixs, COLOR_GREEN);
    pixGreenScale = pixScaleGray2xLI(pixGreen);
    pixDestroy(&pixGreen);
    pixBlue = pixGetRGBComponent(pixs, COLOR_BLUE);
    pixBlueScale = pixScaleGray2xLI(pixBlue);
    pixDestroy(&pixBlue);

    if ((pixd = pixCreateRGBImage(pixRedScale, pixGreenScale, pixBlueScale))
            == NULL)
        return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    pixDestroy(&pixRedScale);
    pixDestroy(&pixGreenScale);
    pixDestroy(&pixBlueScale);

    return pixd;
}


/*
 *  pixScaleColor4xLI()
 *
 *      Input: pixs  (32 bpp, representing 24 bpp 3-color)
 *      Return:  pixd, or null on error
 *
 *      Special fast version for scaling up by 4x
 */
PIX *
pixScaleColor4xLI(PIX   *pixs)
{
PIX          *pixRed, *pixGreen, *pixBlue;
PIX          *pixRedScale, *pixGreenScale, *pixBlueScale;
PIX    	     *pixd;

    PROCNAME("pixScaleColor4xLI");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 32)
        return (PIX *)ERROR_PTR("pixs must be 32 bpp", procName, NULL);

    pixRed = pixGetRGBComponent(pixs, COLOR_RED);
    pixRedScale = pixScaleGray4xLI(pixRed);
    pixDestroy(&pixRed);
    pixGreen = pixGetRGBComponent(pixs, COLOR_GREEN);
    pixGreenScale = pixScaleGray4xLI(pixGreen);
    pixDestroy(&pixGreen);
    pixBlue = pixGetRGBComponent(pixs, COLOR_BLUE);
    pixBlueScale = pixScaleGray4xLI(pixBlue);
    pixDestroy(&pixBlue);

    if ((pixd = pixCreateRGBImage(pixRedScale, pixGreenScale, pixBlueScale))
            == NULL)
        return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    pixDestroy(&pixRedScale);
    pixDestroy(&pixGreenScale);
    pixDestroy(&pixBlueScale);

    return pixd;
}


/*------------------------------------------------------------------*
 *              General linear interpolated scaling                 *
 *------------------------------------------------------------------*/
/*
 *  pixScaleGrayLI()
 *
 *         Input: pixs (8 bpp grayscale)
 *                scalex 
 *                scaley 
 *         Return:  pixd, or null on error
 *
 *         This function is appropriate for upscaling
 *         (magnification: scale factors > 1), and for a
 *         small amount of downscaling (reduction: scale
 *         factors > 0.5).   For scale factors less than 0.5,
 *         the best result is obtained by area mapping,
 *         but this is very expensive.  So for such large
 *         reductions, it is more appropriate to do low pass
 *         filtering followed by subsampling, a combination
 *         which is effectively a cheap form of area mapping.
 *
 *         Some details follow.
 *
 *         For each pixel in the dest, this does a linear
 *         interpolation of 4 neighboring pixels in the src.
 *         Specifically, consider the UL corner of src and
 *         dest pixels.  The UL corner of the dest falls within
 *         a src pixel, whose four corners are the UL corners
 *         of 4 adjacent src pixels.  The value of the dest
 *         is taken by linear interpolation using the values of
 *         the four src pixels and the distance of the UL corner
 *         of the dest from each corner.
 *
 *         If the image is expanded so that the dest pixel is
 *         smaller than the src pixel, such interpolation
 *         is a reasonable approach.  This interpolation is
 *         also good for a small image reduction factor that
 *         is not more than a 2x reduction.
 *
 *         Note that the linear interpolation algorithm for scaling
 *         is identical in form to the area-mapping algorithm
 *         for grayscale rotation.  The latter corresponds to a
 *         translation of each pixel without scaling.
 * 
 *         This function is NOT optimal if the scaling involves
 *         a large reduction.    If the image is significantly
 *         reduced, so that the dest pixel is much larger than
 *         the src pixels, this interpolation, which is over src
 *         pixels only near the UL corner of the dest pixel,
 *         is not going to give a good area-mapping average.
 *         Because area mapping for image scaling is considerably
 *         more computationally intensive than linear interpolation,
 *         we choose not to use it.   For large image reduction,
 *         linear interpolation over adjacent src pixels
 *         degenerates asymptotically to subsampling.  But
 *         subsampling without a low-pass pre-filter causes
 *         aliasing by the nyquist theorem.  To avoid aliasing,
 *         a low-pass filter (e.g., an averaging filter) of
 *         size roughly equal to the dest pixel (i.e., the
 *         reduction factor) should be applied to the src before
 *         subsampling.
 *
 *         As an alternative to low-pass filtering and subsampling
 *         for large reduction factors, linear interpolation can
 *         also be done between the (widely separated) src pixels in
 *         which the corners of the dest pixel lie.  This also is
 *         not optimal, as it samples src pixels only near the
 *         corners of the dest pixel, and it is not implemented.
 *
 */
PIX *
pixScaleGrayLI(PIX      *pixs,
	       FLOAT32   scalex,
	       FLOAT32   scaley)
{
INT32         d, ws, hs, wpls, wd, hd, wpld;
UINT32       *datas, *datad;
PIX    	     *pixd;

    PROCNAME("pixScaleGrayLI");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);

    if ((d = pixGetDepth(pixs)) != 8)
        return (PIX *)ERROR_PTR("pixs must be 8 bpp", procName, NULL);
    
    ws = pixGetWidth(pixs);
    hs = pixGetHeight(pixs);
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);
    wd = (INT32)(scalex * (FLOAT32)ws + 0.5);
    hd = (INT32)(scaley * (FLOAT32)hs + 0.5);
    if ((pixd = pixCreate(wd, hd, 8)) == NULL)
        return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    datad = pixGetData(pixd);
    wpld = pixGetWpl(pixd);

    scaleGrayLILow(datad, wd, hd, wpld, datas, ws, hs, wpls);

    return pixd;
}


/*
 *    pixScaleGray2xLI()
 * 
 *        Input: pixs (8 bpp grayscale)
 *        Return:  pixd, or null on error
 *
 *        This is a special case of linear interpolated scaling
 *        for 2x upscaling.
 */
PIX *
pixScaleGray2xLI(PIX   *pixs)
{
INT32         d, ws, hs, wpls, wpld;
UINT32       *datas, *datad;
PIX    	     *pixd;

    PROCNAME("pixScaleGray2xLI");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);

    if ((d = pixGetDepth(pixs)) != 8)
        return (PIX *)ERROR_PTR("pixs must be 8 bpp", procName, NULL);
    
    ws = pixGetWidth(pixs);
    hs = pixGetHeight(pixs);
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);
    if ((pixd = pixCreate(2 * ws, 2 * hs, 8)) == NULL)
        return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    datad = pixGetData(pixd);
    wpld = pixGetWpl(pixd);

    scaleGray2xLILow(datad, wpld, datas, ws, hs, wpls);

    return pixd;
}


/*
 *    pixScaleGray4xLI()
 * 
 *        Input: pixs (8 bpp grayscale)
 *        Return:  pixd, or null on error
 *
 *        This is a special case of linear interpolated scaling
 *        for 4x upscaling.
 */
PIX *
pixScaleGray4xLI(PIX   *pixs)
{
INT32         d, ws, hs, wpls, wpld;
UINT32       *datas, *datad;
PIX    	     *pixd;

    PROCNAME("pixScaleGray4xLI");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);

    if ((d = pixGetDepth(pixs)) != 8)
        return (PIX *)ERROR_PTR("pixs must be 8 bpp", procName, NULL);
    
    ws = pixGetWidth(pixs);
    hs = pixGetHeight(pixs);
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);
    if ((pixd = pixCreate(4 * ws, 4 * hs, 8)) == NULL)
        return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    datad = pixGetData(pixd);
    wpld = pixGetWpl(pixd);

    scaleGray4xLILow(datad, wpld, datas, ws, hs, wpls);

    return pixd;
}



/*------------------------------------------------------------------*
 *              General scaling by closest pixel sampling           *
 *------------------------------------------------------------------*/
/*
 *  pixScaleBySampling()
 *
 *         Input: pixs (1, 2, 4, 8, 16, 32 bpp)
 *                scalex, scaley
 *         Return:  pixd, or null on error
 * 
 *         Note: this function samples from the source without
 *               filtering, so that aliasing will result for
 *               subsampling (scalex and scaley < 1.0).
 */
PIX *
pixScaleBySampling(PIX      *pixs,
	           FLOAT32   scalex,
	           FLOAT32   scaley)
{
INT32         ws, hs, d, wpls, wd, hd, wpld;
UINT32       *datas, *datad;
PIX    	     *pixd;

    PROCNAME("pixScaleBySampling");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);

    if ((d = pixGetDepth(pixs)) == 1)
        return pixScaleBinary(pixs, scalex, scaley);

    ws = pixGetWidth(pixs);
    hs = pixGetHeight(pixs);
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);
    wd = (INT32)(scalex * (FLOAT32)ws + 0.5);
    hd = (INT32)(scaley * (FLOAT32)hs + 0.5);
    if ((pixd = pixCreate(wd, hd, d)) == NULL)
        return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    datad = pixGetData(pixd);
    wpld = pixGetWpl(pixd);

    if (pixGetColormap(pixs))
	pixCopyColormap(pixd, pixs);

    scaleBySamplingLow(datad, wd, hd, wpld, datas, ws, hs, d, wpls);

    return pixd;
}



/*------------------------------------------------------------------*
 *               Binary scaling by closest pixel sampling           *
 *------------------------------------------------------------------*/
/*
 *  pixScaleBinary()
 *
 *         Input: pixs (1 bpp)
 *                scalex, scaley
 *         Return:  pixd, or null on error
 * 
 *         Note: this function samples from the source without
 *               filtering, so that aliasing will result for
 *               subsampling (scalex and scaley < 1.0).
 */
PIX *
pixScaleBinary(PIX      *pixs,
	       FLOAT32   scalex,
	       FLOAT32   scaley)
{
INT32         ws, hs, wpls, wd, hd, wpld;
UINT32       *datas, *datad;
PIX    	     *pixd;

    PROCNAME("pixScaleBinary");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 1)
        return (PIX *)ERROR_PTR("pixs must be 1 bpp", procName, NULL);

    ws = pixGetWidth(pixs);
    hs = pixGetHeight(pixs);
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);
    wd = (INT32)(scalex * (FLOAT32)ws + 0.5);
    hd = (INT32)(scaley * (FLOAT32)hs + 0.5);
    if ((pixd = pixCreate(wd, hd, 1)) == NULL)
        return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    datad = pixGetData(pixd);
    wpld = pixGetWpl(pixd);

    scaleBinaryLow(datad, wd, hd, wpld, datas, ws, hs, wpls);

    return pixd;
}



/*------------------------------------------------------------------*
 *                         Scale-to-gray 3x                         *
 *------------------------------------------------------------------*/
/*
 *  pixScaleToGray3()
 *       Input:  1 bpp pix
 *       Return: 8 bpp pix, scaled down by 3x in each direction,
 *               or NULL on error.
 *       Time: about 0.1 sec on 300 ppi page image (866 MHz P3)
 *       Speed: about 1 src pix per 10 cycles
 */
PIX *
pixScaleToGray3(PIX   *pixs)
{
UINT8        *valtab;
INT32         ws, hs, wd, hd;
INT32         wpld, wpls;
UINT32       *sumtab;
UINT32       *datas, *datad;
PIX          *pixd;

    PROCNAME("pixScaleToGray3");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 1)
        return (PIX *)ERROR_PTR("pixs must be 1 bpp", procName, NULL);

    ws = pixGetWidth(pixs);
    hs = pixGetHeight(pixs);
    wd = (ws / 3) & 0xfffffff8;    /* truncate to factor of 8 */
    hd = hs / 3;
    if (wd == 0 || hd == 0)
        return (PIX *)ERROR_PTR("pixs too small", procName, NULL);

    if ((pixd = pixCreate(wd, hd, 8)) == NULL)
        return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    datas = pixGetData(pixs);
    datad = pixGetData(pixd);
    wpls = pixGetWpl(pixs);
    wpld = pixGetWpl(pixd);

    if ((sumtab = makeSumTabSG3()) == NULL)
	return (PIX *)ERROR_PTR("sumtab not made", procName, NULL);

    if ((valtab = makeValTabSG3()) == NULL)
	return (PIX *)ERROR_PTR("valtab not made", procName, NULL);

    scaleToGray3Low(datad, wd, hd, wpld, datas, wpls, sumtab, valtab);

    FREE((void *)sumtab);
    FREE((void *)valtab);

    return pixd;
}


/*------------------------------------------------------------------*
 *                         Scale-to-gray 4x                         *
 *------------------------------------------------------------------*/
/*
 *  pixScaleToGray4()
 *       Input:  1 bpp pix
 *       Return: 8 bpp pix, scaled down by 4x in each direction,
 *               or NULL on error.
 */
PIX *
pixScaleToGray4(PIX   *pixs)
{
UINT8        *valtab;
INT32         ws, hs, wd, hd;
INT32         wpld, wpls;
UINT32       *sumtab;
UINT32       *datas, *datad;
PIX          *pixd;

    PROCNAME("pixScaleToGray4");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 1)
        return (PIX *)ERROR_PTR("pixs must be 1 bpp", procName, NULL);

    ws = pixGetWidth(pixs);
    hs = pixGetHeight(pixs);
    wd = (ws / 4) & 0xfffffffe;    /* truncate to factor of 2 */
    hd = hs / 4;
    if (wd == 0 || hd == 0)
        return (PIX *)ERROR_PTR("pixs too small", procName, NULL);

    if ((pixd = pixCreate(wd, hd, 8)) == NULL)
        return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    datas = pixGetData(pixs);
    datad = pixGetData(pixd);
    wpls = pixGetWpl(pixs);
    wpld = pixGetWpl(pixd);

    if ((sumtab = makeSumTabSG4()) == NULL)
	return (PIX *)ERROR_PTR("sumtab not made", procName, NULL);

    if ((valtab = makeValTabSG4()) == NULL)
	return (PIX *)ERROR_PTR("valtab not made", procName, NULL);

    scaleToGray4Low(datad, wd, hd, wpld, datas, wpls, sumtab, valtab);

    FREE((void *)sumtab);
    FREE((void *)valtab);

    return pixd;
}


/*------------------------------------------------------------------*
 *                Scale 2x followed by binarization                 *
 *------------------------------------------------------------------*/
/*
 *  pixScaleGray2xLIThresh()
 *
 *      Input:   pixs (8 bpp)
 *               thresh  (between 0 and 255)
 *      Return:  pixd (1 bpp), or null on error
 *
 *      This does 2x upscale on pixs, using linear interpolation,
 *      followed by thresholding to binary.
 *      Buffers are used to avoid making a large grayscale image.
 */
PIX *
pixScaleGray2xLIThresh(PIX     *pixs,
                       INT32    thresh)
{
INT32    i, ws, hs, hsm, wd, hd, wpls, wplb, wpld;
UINT32  *datas, *datad, *lines, *lined, *lineb;
PIX     *pixd;

    PROCNAME("pixScaleGray2xLIThresh");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 8)
        return (PIX *)ERROR_PTR("pixs must be 8 bpp", procName, NULL);
    if (thresh < 0 || thresh > 255)
        return (PIX *)ERROR_PTR("thresh must be in [0, ... 255]",
	    procName, NULL);

    ws = pixGetWidth(pixs);
    hs = pixGetHeight(pixs);
    wd = 2 * ws;
    hd = 2 * hs;
    hsm = hs - 1;
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);

        /* make line buffer for 2 lines of virtual intermediate image */
    wplb = (wd + 3) / 4;
    if ((lineb = (UINT32 *)CALLOC(2 * wplb, sizeof(UINT32))) == NULL)
        return (PIX *)ERROR_PTR("lineb not made", procName, NULL);

	/* make dest binary image */
    if ((pixd = pixCreate(wd, hd, 1)) == NULL)
        return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    wpld = pixGetWpl(pixd);
    datad = pixGetData(pixd);

        /* do all but last src line */
    for (i = 0; i < hsm; i++) {
	lines = datas + i * wpls;
	lined = datad + 2 * i * wpld;  /* do 2 dest lines at a time */
	scaleGray2xLILineLow(lineb, wplb, lines, ws, wpls, 0);
        thresholdToBinaryLineLow(lined, wd, lineb, 8, thresh);
        thresholdToBinaryLineLow(lined + wpld, wd, lineb + wplb, 8, thresh);
    }

        /* do last src line */
    lines = datas + hsm * wpls;
    lined = datad + 2 * hsm * wpld;
    scaleGray2xLILineLow(lineb, wplb, lines, ws, wpls, 1);
    thresholdToBinaryLineLow(lined, wd, lineb, 8, thresh);
    thresholdToBinaryLineLow(lined + wpld, wd, lineb + wplb, 8, thresh);

    FREE((void *)lineb);

    return pixd;
}


/*
 *  pixScaleGray2xLIDither()
 *
 *      Input:   pixs (8 bpp)
 *      Return:  pixd (1 bpp), or null on error
 *
 *      This does 2x upscale on pixs, using linear interpolation,
 *      followed by Floyd-Steinberg dithering to binary.
 *
 *      Buffers are used to avoid making a large grayscale image.
 *          * We use 2 line buffers for the src, required for the 2x
 *            LI upscale.
 *          * We use 3 line buffers for the intermediate image.
 *            Two are filled with each 2xLI row operation; the third
 *            is needed because the upscale and dithering ops are
 *            out of sync.
 */
PIX *
pixScaleGray2xLIDither(PIX  *pixs)
{
INT32    i, ws, hs, hsm, wd, hd, wpls, wplb, wpld;
UINT32  *datas, *datad;
UINT32  *lined;
UINT32  *lineb;   /* 2 intermediate buffer lines */
UINT32  *linebp;  /* 1 intermediate buffer line */
UINT32  *bufs;    /* 2 source buffer lines */
PIX     *pixd;

    PROCNAME("pixScaleGray2xLIDither");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 8)
        return (PIX *)ERROR_PTR("pixs must be 8 bpp", procName, NULL);

    ws = pixGetWidth(pixs);
    hs = pixGetHeight(pixs);
    wd = 2 * ws;
    hd = 2 * hs;
    hsm = hs - 1;
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);

        /* make line buffers for 2 lines of src image */
    if ((bufs = (UINT32 *)CALLOC(2 * wpls, sizeof(UINT32))) == NULL)
        return (PIX *)ERROR_PTR("bufs not made", procName, NULL);

        /* make line buffer for 2 lines of virtual intermediate image */
    wplb = (wd + 3) / 4;
    if ((lineb = (UINT32 *)CALLOC(2 * wplb, sizeof(UINT32))) == NULL)
        return (PIX *)ERROR_PTR("lineb not made", procName, NULL);

        /* make line buffer for 1 line of virtual intermediate image */
    if ((linebp = (UINT32 *)CALLOC(wplb, sizeof(UINT32))) == NULL)
        return (PIX *)ERROR_PTR("linebp not made", procName, NULL);

	/* make dest binary image */
    if ((pixd = pixCreate(wd, hd, 1)) == NULL)
        return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    wpld = pixGetWpl(pixd);
    datad = pixGetData(pixd);

        /* start with the first src and the first dest line */
    memcpy(bufs, datas, 4 * wpls);   /* first src line */
    memcpy(bufs + wpls, datas + wpls, 4 * wpls);  /* 2nd src line */
    scaleGray2xLILineLow(lineb, wplb, bufs, ws, wpls, 0);  /* 2 i lines */
    lined = datad;
    ditherFSLineLow(lined, wd, lineb, lineb + wplb, 0);  /* first d line */

        /* do all but last src line */
    for (i = 1; i < hsm; i++) {
	memcpy(bufs, datas + i * wpls, 4 * wpls);  /* i-th src line */
	memcpy(bufs + wpls, datas + (i + 1) * wpls, 4 * wpls);
	memcpy(linebp, lineb + wplb, 4 * wplb);
	scaleGray2xLILineLow(lineb, wplb, bufs, ws, wpls, 0);  /* 2 i lines */
	lined = datad + 2 * i * wpld;
        ditherFSLineLow(lined - wpld, wd, linebp, lineb, 0);  /* odd d line */
        ditherFSLineLow(lined, wd, lineb, lineb + wplb, 0);  /* even d line */
    }

       /* do the last src line and the last 3 dest lines */
    memcpy(bufs, datas + hsm * wpls, 4 * wpls);  /* hsm-th src line */
    memcpy(linebp, lineb + wplb, 4 * wplb);   /* 1 i line */
    scaleGray2xLILineLow(lineb, wplb, bufs, ws, wpls, 1);  /* 2 i lines */
    ditherFSLineLow(lined + wpld, wd, linebp, lineb, 0);  /* odd d line */
    ditherFSLineLow(lined + 2 * wpld, wd, lineb, lineb + wplb, 0);  /* even d line */
    ditherFSLineLow(lined + 3 * wpld, wd, lineb + wplb, NULL, 1); /* last d line */

    FREE((void *)bufs);
    FREE((void *)lineb);
    FREE((void *)linebp);

    return pixd;
}


/*------------------------------------------------------------------*
 *                Scale 4x followed by binarization                 *
 *------------------------------------------------------------------*/
/*
 *  pixScaleGray4xLIThresh()
 *
 *      Input:   pixs (8 bpp)
 *               thresh  (between 0 and 255)
 *      Return:  pixd (1 bpp), or null on error
 *
 *      This does 4x upscale on pixs, using linear interpolation,
 *      followed by thresholding to binary.
 *      Buffers are used to avoid making a large grayscale image.
 *
 *      If a full 4x expanded grayscale image can be kept in memory,
 *      this function is only about 10% faster than separately doing
 *      a linear interpolation to a large grayscale image, followed
 *      by thresholding to binary.
 */
PIX *
pixScaleGray4xLIThresh(PIX     *pixs,
                       INT32    thresh)
{
INT32    i, j, ws, hs, hsm, wd, hd, wpls, wplb, wpld;
UINT32  *datas, *datad, *lines, *lined, *lineb;
PIX     *pixd;

    PROCNAME("pixScaleGray4xLIThresh");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 8)
        return (PIX *)ERROR_PTR("pixs must be 8 bpp", procName, NULL);
    if (thresh < 0 || thresh > 255)
        return (PIX *)ERROR_PTR("thresh must be in [0, ... 255]",
	    procName, NULL);

    ws = pixGetWidth(pixs);
    hs = pixGetHeight(pixs);
    wd = 4 * ws;
    hd = 4 * hs;
    hsm = hs - 1;
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);

        /* make line buffer for 4 lines of virtual intermediate image */
    wplb = (wd + 3) / 4;
    if ((lineb = (UINT32 *)CALLOC(4 * wplb, sizeof(UINT32))) == NULL)
        return (PIX *)ERROR_PTR("lineb not made", procName, NULL);

	/* make dest binary image */
    if ((pixd = pixCreate(wd, hd, 1)) == NULL)
        return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    wpld = pixGetWpl(pixd);
    datad = pixGetData(pixd);

        /* do all but last src line */
    for (i = 0; i < hsm; i++) {
	lines = datas + i * wpls;
	lined = datad + 4 * i * wpld;  /* do 4 dest lines at a time */
	scaleGray4xLILineLow(lineb, wplb, lines, ws, wpls, 0);
	for (j = 0; j < 4; j++) {
	    thresholdToBinaryLineLow(lined + j * wpld, wd,
	                             lineb + j * wplb, 8, thresh);
	}
    }

        /* do last src line */
    lines = datas + hsm * wpls;
    lined = datad + 4 * hsm * wpld;
    scaleGray4xLILineLow(lineb, wplb, lines, ws, wpls, 1);
    for (j = 0; j < 4; j++) {
	thresholdToBinaryLineLow(lined + j * wpld, wd,
				 lineb + j * wplb, 8, thresh);
    }

    FREE((void *)lineb);

    return pixd;
}


/*
 *  pixScaleGray4xLIDither()
 *
 *      Input:   pixs (8 bpp)
 *      Return:  pixd (1 bpp), or null on error
 *
 *      This does 4x upscale on pixs, using linear interpolation,
 *      followed by Floyd-Steinberg dithering to binary.
 *
 *      Buffers are used to avoid making a large grayscale image.
 *          * We use 2 line buffers for the src, required for the
 *            4xLI upscale.
 *          * We use 5 line buffers for the intermediate image.
 *            Four are filled with each 4xLI row operation; the fifth
 *            is needed because the upscale and dithering ops are
 *            out of sync.
 *
 *      If a full 4x expanded grayscale image can be kept in memory,
 *      this function is only about 5% faster than separately doing
 *      a linear interpolation to a large grayscale image, followed
 *      by error-diffusion dithering to binary.
 */
PIX *
pixScaleGray4xLIDither(PIX  *pixs)
{
INT32    i, j, ws, hs, hsm, wd, hd, wpls, wplb, wpld;
UINT32  *datas, *datad;
UINT32  *lined;
UINT32  *lineb;   /* 4 intermediate buffer lines */
UINT32  *linebp;  /* 1 intermediate buffer line */
UINT32  *bufs;    /* 2 source buffer lines */
PIX     *pixd;

    PROCNAME("pixScaleGray4xLIDither");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 8)
        return (PIX *)ERROR_PTR("pixs must be 8 bpp", procName, NULL);

    ws = pixGetWidth(pixs);
    hs = pixGetHeight(pixs);
    wd = 4 * ws;
    hd = 4 * hs;
    hsm = hs - 1;
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);

        /* make line buffers for 2 lines of src image */
    if ((bufs = (UINT32 *)CALLOC(2 * wpls, sizeof(UINT32))) == NULL)
        return (PIX *)ERROR_PTR("bufs not made", procName, NULL);

        /* make line buffer for 4 lines of virtual intermediate image */
    wplb = (wd + 3) / 4;
    if ((lineb = (UINT32 *)CALLOC(4 * wplb, sizeof(UINT32))) == NULL)
        return (PIX *)ERROR_PTR("lineb not made", procName, NULL);

        /* make line buffer for 1 line of virtual intermediate image */
    if ((linebp = (UINT32 *)CALLOC(wplb, sizeof(UINT32))) == NULL)
        return (PIX *)ERROR_PTR("linebp not made", procName, NULL);

	/* make dest binary image */
    if ((pixd = pixCreate(wd, hd, 1)) == NULL)
        return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    wpld = pixGetWpl(pixd);
    datad = pixGetData(pixd);

        /* start with the first src and the first 3 dest lines */
    memcpy(bufs, datas, 4 * wpls);   /* first src line */
    memcpy(bufs + wpls, datas + wpls, 4 * wpls);  /* 2nd src line */
    scaleGray4xLILineLow(lineb, wplb, bufs, ws, wpls, 0);  /* 4 b lines */
    lined = datad;
    for (j = 0; j < 3; j++) {  /* first 3 d lines of Q */
	ditherFSLineLow(lined + j * wpld, wd, lineb + j * wplb,
	                lineb + (j + 1) * wplb, 0);
    }

        /* do all but last src line */
    for (i = 1; i < hsm; i++) {
	memcpy(bufs, datas + i * wpls, 4 * wpls);  /* i-th src line */
	memcpy(bufs + wpls, datas + (i + 1) * wpls, 4 * wpls);
	memcpy(linebp, lineb + 3 * wplb, 4 * wplb);
	scaleGray4xLILineLow(lineb, wplb, bufs, ws, wpls, 0);  /* 4 b lines */
	lined = datad + 4 * i * wpld;
	ditherFSLineLow(lined - wpld, wd, linebp, lineb, 0); /* 4th d line of Q */
	for (j = 0; j < 3; j++) {  /* next 3 d lines of Quad */
            ditherFSLineLow(lined + j * wpld, wd, lineb + j * wplb,
	                    lineb + (j + 1) * wplb, 0);
        }			     
    }

       /* do the last src line and the last 5 dest lines */
    memcpy(bufs, datas + hsm * wpls, 4 * wpls);  /* hsm-th src line */
    memcpy(linebp, lineb + 3 * wplb, 4 * wplb);   /* 1 b line */
    scaleGray4xLILineLow(lineb, wplb, bufs, ws, wpls, 1);  /* 4 b lines */
    lined = datad + 4 * hsm * wpld;
    ditherFSLineLow(lined - wpld, wd, linebp, lineb, 0);  /* 4th d line of Q */
    for (j = 0; j < 3; j++) {  /* next 3 d lines of Quad */
	ditherFSLineLow(lined + j * wpld, wd, lineb + j * wplb,
			lineb + (j + 1) * wplb, 0);
    }
       /* and finally, the last d line */
    ditherFSLineLow(lined + 3 * wpld, wd, lineb + 3 * wplb, NULL, 1);

    FREE((void *)bufs);
    FREE((void *)lineb);
    FREE((void *)linebp);

    return pixd;
}


