/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/



/*
 *  sel1.c
 *
 *      Basic ops on Sels and Selas
 *
 *         Create/destroy/copy:
 *            SELA        *selaCreate()
 *            void         selaDestroy()
 *            SEL         *selCreate()
 *            void         selDestroy()
 *            SEL         *selCopy()
 *            SEL         *selCreateBrick()
 *
 *         Helper proc:
 *            INT32     **create2dIntArray()
 *
 *         Extension of sela:
 *            SELA       *selaAddSel()
 *            INT32       selaExtendArray()
 *
 *         Accessors:
 *            INT32       selaGetCount()
 *            SEL        *selaGetSel()
 *            char       *selGetName()
 *            INT32       selaFindSelByName()
 *            INT32       selSetElement()
 *
 *         Max extent vals for erosion and hmt:
 *            INT32       selFindMaxTranslations()
 *
 *         Write/read & visualization:
 *            INT32       selaWrite()
 *            INT32       selaWriteStream()
 *            INT32       selWriteStream()
 *            INT32       selaRead()
 *            INT32       selaReadStream()
 *            INT32       selReadStream()
 *
 *
 *     Usage note:
 *        Consistency, being the hobgoblin of small minds,
 *        is adhered to here in the dimensioning and accessing of sels.
 *        Everything is done in standard matrix (row, column) order.
 *        We have 2 functions that make sels:
 *             selCreate(), with input (h, w, [name])
 *             selCreateBrick(), with input (h, w, cy, cx, val)
 *        When we set specific elements in a sel, we likewise use
 *        (row, col) ordering:
 *             selSetElement(), with input (row, col, type)
 */



#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "allheaders.h"


#define  INITIAL_PTR_ARRAYSIZE   50   /* n'import quoi */
#define  MANY_SELS               1000
#define  BUFSIZE                 256



/*------------------------------------------------------------------------*
 *                      Create / Destroy / Copy                           *
 *------------------------------------------------------------------------*/
/*!
 *  selaCreate()
 *
 *       Input: n (initial number of sel ptrs; use 0 for default)
 *       Return: sela, or null on error
 */
SELA *
selaCreate(INT32  n)
{
SELA    *sela;

    PROCNAME("selaCreate");

    if (n <= 0)
	n = INITIAL_PTR_ARRAYSIZE;
    if (n > MANY_SELS)
	WARNING_INT("%d sels", procName, n);

    if ((sela = (SELA *)CALLOC(1, sizeof(SELA))) == NULL)
	return (SELA *)ERROR_PTR("sela not made", procName, NULL);

    sela->nalloc = n;
    sela->n = 0;

	/* make array of se ptrs */
    if ((sela->sel = (SEL **)CALLOC(n, sizeof(SEL *))) == NULL)
	return (SELA *)ERROR_PTR("sel ptrs not made", procName, NULL);

    return sela;
}


/*!
 *  selaDestroy()
 *
 *     Input:  &sela (<to be nulled>)
 *     Return: void
 */
void
selaDestroy(SELA  **psela)
{
SELA       *sela;
INT32       i;

    if (!psela) return;
    if ((sela = *psela) == NULL)
	return;

    for (i = 0; i < sela->n; i++)
	selDestroy(&sela->sel[i]);
    FREE((void *)sela->sel);

    FREE((void *)sela);
    *psela = NULL;
    return;
}


/*!
 *  selCreate()
 *
 *     Input:  height, width
 *             name (<optional> sel name)
 *     Return: sel, or null on error
 *
 *     Usage note: selCreate() initializes all values to 0.
 *                 After this call, (cy,cx) and nonzero data
 *                 values must be assigned.  If a text name
 *                 is not assigned here, it will be needed
 *                 later when the sel is put into a sela.
 */
SEL *
selCreate(INT32  height,
	  INT32  width,
	  char  *name)
{
SEL   *sel;

    PROCNAME("selCreate");

    if ((sel = (SEL *)CALLOC(1, sizeof(SEL))) == NULL)
	return (SEL *)ERROR_PTR("sel not made", procName, NULL);
    if (name)
	sel->name = stringNew(name);
    sel->sy = height;
    sel->sx = width;
    if ((sel->data = create2dIntArray(height, width)) == NULL)
	return (SEL *)ERROR_PTR("data not allocated", procName, NULL);

    return sel;
}


/*!
 *  selDestroy()
 *
 *     Input: &sel (<to be nulled>)
 *     Return: void
 */
void
selDestroy(SEL  **psel)
{
INT32     i;
SEL      *sel;

    if (!psel) return;
    if ((sel = *psel) == NULL)
	return;

    for (i = 0; i < sel->sy; i++)
	FREE((void *)sel->data[i]);
    FREE((void *)sel->data);

    if (sel->name)
	FREE((void *)sel->name);

    FREE((void *)sel);

    *psel = NULL;
    return;
}


/*!
 *  selCopy() 
 *
 *      Input:  sel
 *      Returns: a copy of the sel, or null on error
 *              
 */
SEL *
selCopy(SEL  *sel)
{
INT32     i, j;
SEL      *csel;

    PROCNAME("selCopy");

    if (!sel)
	return (SEL *)ERROR_PTR("sel not defined", procName, NULL);

    if ((csel = (SEL *)CALLOC(1, sizeof(SEL))) == NULL)
	return (SEL *)ERROR_PTR("csel not made", procName, NULL);

    csel->sy = sel->sy;
    csel->sx = sel->sx;
    csel->cy = sel->cy;
    csel->cx = sel->cx;

    if ((csel->data = create2dIntArray(sel->sy, sel->sx)) == NULL)
	return (SEL *)ERROR_PTR("sel data not made", procName, NULL);

    for (i = 0; i < sel->sy; i++)
	for (j = 0; j < sel->sx; j++)
	    csel->data[i][j] = sel->data[i][j];

    csel->name = stringNew(sel->name);

    return csel;
}



/*!
 *  selCreateBrick()
 *
 *     Input:  height, width
 *             cy, cx  (center, relative to UL corner at 0,0)
 *             val  (HIT, MISS, or DONT_CARE)
 *     Return: sel, or null on error
 *     Action: a "brick" is a rectangular array of either hits,
 *             misses, or don't-cares.
 *
 */
SEL *
selCreateBrick(INT32  h,
	       INT32  w,
	       INT32  cy,
	       INT32  cx,
	       INT32  val)
{
INT32    i, j;
SEL     *sel;

    PROCNAME("selCreateBrick");

    if (h <= 0 || w <= 0)
	return (SEL *)ERROR_PTR("h and w must both be > 0", procName, NULL);
    if (val != HIT && val != MISS && val != DONT_CARE)
	return (SEL *)ERROR_PTR("val must be in {0,1,2}", procName, NULL);

    if ((sel = selCreate(h, w, NULL)) == NULL)
	return (SEL *)ERROR_PTR("sel not made", procName, NULL);
    sel->cy = cy;
    sel->cx = cx;
    for (i = 0; i < h; i++)
	for (j = 0; j < w; j++)
	    sel->data[i][j] = val;

    return sel;
}


/*!
 *  create2dIntArray()
 *         Input: sy (rows == height)
 *                sx (columns == width)
 *         Return: doubly indexed array (i.e., an array of
 *                 sy row pointers, each of which points to
 *                 an array of sx ints)
 *         Note: the array[sy][sx] is indexed in standard
 *               "matrix notation" with the row index first.
 */
INT32 **
create2dIntArray(INT32  sy,
                 INT32  sx)
{
INT32    i;
INT32  **array;

    PROCNAME("create2dIntArray");

    if ((array = (INT32 **)CALLOC(sy, sizeof(INT32 *))) == NULL)
	return (INT32 **)ERROR_PTR("ptr array not made", procName, NULL);

    for (i = 0; i < sy; i++) {
	if ((array[i] = (INT32 *)CALLOC(sx, sizeof(INT32))) == NULL)
	    return (INT32 **)ERROR_PTR("array not made", procName, NULL);
    }

    return array;
}



/*------------------------------------------------------------------------*
 *                           Extension of sela                            *
 *------------------------------------------------------------------------*/
/*!
 *  selaAddSel()
 *
 *           Input:  sela
 *                   sel to be added
 *                   selname (ignored if already defined in sel;
 *                            req'd in sel when added to a sela)
 *                   copyflag (for sel: 0 inserts, 1 copies)
 *
 *           Returns: 0 if OK; 1 on error
 *
 *           Action:  Adds sel to arrays, making a copy if flagged.
 *                    Copies the name to the sel if necessary.
 *                    Increments the sel count.
 */
INT32
selaAddSel(SELA     *sela,
	   SEL      *sel,
	   char     *selname,
	   INT32     copyflag)
{
INT32     n;
SEL      *csel;

    PROCNAME("selaAddSel");

    if (!sela)
	return ERROR_INT("sela not defined", procName, 1);
    if (!sel)
	return ERROR_INT("sel not defined", procName, 1);
    if (!sel->name && !selname)
	return ERROR_INT("added sel must have name", procName, 1);

    if (copyflag == TRUE) {
	if ((csel = selCopy(sel)) == NULL)
	    return ERROR_INT("csel not made", procName, 1);
    }
    else   /* copyflag is false; insert directly */
	csel = sel;

    if (csel->name == NULL)
	csel->name = stringNew(selname);

    n = selaGetCount(sela);
    if (n >= sela->nalloc)
	selaExtendArray(sela);
    sela->sel[n] = csel;
    sela->n++;

    return 0;
}
    

/*!
 *     selaExtendArray()
 *
 *         Input:  sela
 *         Return: 0 if OK; 1 on error
 *
 *         Action: doubles the ptr array; copies the old ptr addresses
 *                 into the new array; updates ptr array size
 *
 */
INT32
selaExtendArray(SELA  *sela)
{

    PROCNAME("selaExtendArray");

    if (!sela)
        return ERROR_INT("sela not defined", procName, 1);
    
    if ((sela->sel = (SEL **)reallocNew((void **)&sela->sel,
				4 * sela->nalloc,
				8 * sela->nalloc)) == NULL)
	    return ERROR_INT("new ptr array not returned", procName, 1);

    sela->nalloc = 2 * sela->nalloc;
    return 0;
}



/*----------------------------------------------------------------------*
 *                               Accessors                              *
 *----------------------------------------------------------------------*/
/*!
 *  selaGetCount()
 *
 *       Input: sela
 *       Return: count, or 0 on error
 */
INT32
selaGetCount(SELA  *sela)
{

    PROCNAME("selaGetCount");

    if (!sela)
	return ERROR_INT("sela not defined", procName, 0);

    return sela->n;
}


/*!
 *  selaGetSel()
 *
 *       Input:  sela
 *               index of sel to be retrieved (not copied)
 */
SEL *
selaGetSel(SELA    *sela,
	   INT32    i)
{

    PROCNAME("selaGetSel");

    if (!sela)
	return (SEL *)ERROR_PTR("sela not defined", procName, NULL);

    if (i < 0 || i >= sela->n)
	return (SEL *)ERROR_PTR("invalid index", procName, NULL);

    return sela->sel[i];
}


/*!
 *   selGetName()
 *
 *        Input: sel
 *        Return:  sel name (not copied), or null if no name or on error
 */
char *
selGetName(SEL  *sel)
{

    PROCNAME("selGetName");

    if (!sel)
	return (char *)ERROR_PTR("sel not defined", procName, NULL);

    return sel->name;
}


/*!
 *   selaFindSelByName()
 *
 *       Input:  sela
 *               sel name
 *               &index (<return>)
 *               &sel  (<return> sel (not a copy))
 *       Return: 0 if OK; 1 on error
 */
INT32
selaFindSelByName(SELA    *sela, 
	          char    *name,
		  INT32   *pindex,
		  SEL    **psel)
{
INT32      i, n;
char      *sname;
SEL       *sel;

    PROCNAME("selaFindSelByName");

    *pindex = -1;
    *psel = NULL;

    if (!sela)
	return ERROR_INT("sela not defined", procName, 1);

    n = selaGetCount(sela);
    for (i = 0; i < n; i++)
    {
	if ((sel = selaGetSel(sela, i)) == NULL) {
	    WARNING("missing sel", procName);
	    continue;
	}
	    
	sname = selGetName(sel);
	if (sname && (!strcmp(name, sname))) {
	    *pindex = i;
	    *psel = sel;
	    return 0;
	}
    }
    
    return 1;
}


/*!
 *   selSetElement()
 *
 *        Input:  sel
 *                row
 *                col
 *                type  (HIT, MISS, DONT_CARE)
 *        Return: 0 if OK; 1 on error
 */
INT32
selSetElement(SEL    *sel,
              INT32   row,
	      INT32   col,
	      INT32   type)
{

    PROCNAME("selSetElement");

    if (!sel)
	return ERROR_INT("sel not defined", procName, 1);
    if (type != HIT && type != MISS && type != DONT_CARE)
	return ERROR_INT("type not in {HIT,MISS,DONT_CARE}", procName, 1);
    if (row < 0 || row >= sel->sy)
	return ERROR_INT("sel row out of bounds", procName, 1);
    if (col < 0 || col >= sel->sx)
	return ERROR_INT("sel col out of bounds", procName, 1);

    sel->data[row][col] = type;
    return 0;
}



/*----------------------------------------------------------------------*
 *                Max translations for erosion and hmt                  *
 *----------------------------------------------------------------------*/
/*!
 *   selFindMaxTranslations()
 *
 *       Input:  sel
 *               &xp, &yp, &xn, &yn  (<return> max shifts)
 *       Return: 0 if OK; 1 on error
 *
 *       Note: these are the maximum shifts for the erosion operation.
 *             For example, when j < cx, the shift of the image
 *             is +x to the cx.  This is a positive xp shift.
 */
INT32
selFindMaxTranslations(SEL    *sel,
                       INT32  *pxp,
                       INT32  *pyp,
                       INT32  *pxn,
                       INT32  *pyn)
{
INT32      i, j, maxxp, maxyp, maxxn, maxyn;

    PROCNAME("selaFindMaxTranslations");

    *pxp = *pyp = *pxn = *pyn = 0;

    if (!sel)
	return ERROR_INT("sel not defined", procName, 1);

    maxxp = maxyp = maxxn = maxyn = 0;
    for (i = 0; i < sel->sy; i++) {
	for (j = 0; j < sel->sx; j++) {
	    if (sel->data[i][j] == 1) {
		maxxp = MAX(maxxp, sel->cx - j);
		maxyp = MAX(maxyp, sel->cy - i);
		maxxn = MAX(maxxn, j - sel->cx);
		maxyn = MAX(maxyn, i - sel->cy);
	    }
	}
    }

    *pxp = maxxp;
    *pyp = maxyp;
    *pxn = maxxn;
    *pyn = maxyn;

    return 0;
}



/*----------------------------------------------------------------------*
 *                    I/O and visualizing Sela & Sel                    *
 *----------------------------------------------------------------------*/
/*!
 *  selaWrite()
 *
 *     Input: filename
 *            sela
 *     Return: 0 if OK, 1 on error
 */
INT32
selaWrite(char   *fname,
	  SELA   *sela)
{
FILE     *fp;

    PROCNAME("selaWriteFile");

    if (!fname)
	return ERROR_INT("fname not defined", procName, 1);
    if (!sela)
	return ERROR_INT("sela not defined", procName, 1);

    if ((fp = fopen(fname, "wb")) == NULL)
	return ERROR_INT("stream not opened", procName, 1);
    selaWriteStream(fp, sela);
    fclose(fp);

    return 0;
}


/*!
 *  selaWriteStream()
 *
 *     Input:  stream
 *             sela
 *     Return: 0 if OK, 1 on error
 */
INT32
selaWriteStream(FILE    *fp,
	        SELA    *sela)
{
INT32     i, n;
SEL      *sel;

    PROCNAME("selaWriteStream");

    if (!fp)
	return ERROR_INT("stream not defined", procName, 1);
    if (!sela)
	return ERROR_INT("sela not defined", procName, 1);

    n = selaGetCount(sela);
    fprintf(fp, "sel array: number of sels = %d\n\n", n);
    for (i = 0; i < n; i++)
    {
	if ((sel = selaGetSel(sela, i)) == NULL)
	    continue;
	selWriteStream(fp, sel);
    }
    return 0;
}


/*!
 *  selWriteStream()
 *
 *     Input:  stream
 *             sel
 *     Return: 0 if OK, 1 on error
 */
INT32
selWriteStream(FILE  *fp,
	       SEL   *sel)
{
INT32      i, j;

    PROCNAME("selWriteStream");

    if (!fp)
	return ERROR_INT("stream not defined", procName, 1);
    if (!sel)
	return ERROR_INT("sel not defined", procName, 1);

    fprintf(fp, "  ------  %s  ------\n", selGetName(sel));
    fprintf(fp, "  sy = %d, sx = %d, cy = %d, cx = %d\n",
	    sel->sy, sel->sx, sel->cy, sel->cx);
    for (i = 0; i < sel->sy; i++) {
	fprintf(fp, "    ");
	for (j = 0; j < sel->sx; j++)
	    fprintf(fp, "%d", sel->data[i][j]);
	fprintf(fp, "\n");
    }
    fprintf(fp, "\n");

    return 0;
}


/*!
 *  selaRead()
 *
 *      Input: filename
 *      Return: sela, or null on error
 */
SELA  *
selaRead(char   *fname)
{
FILE     *fp;
SELA     *sela;

    PROCNAME("selaRead");

    if (!fname)
	return (SELA *)ERROR_PTR("fname not defined", procName, NULL);

    if ((fp = fopen(fname, "rb")) == NULL)
	return (SELA *)ERROR_PTR("stream not opened", procName, NULL);
    if ((sela = selaReadStream(fp)) == NULL)
	return (SELA *)ERROR_PTR("sela not returned", procName, NULL);
    fclose(fp);

    return sela;
}


/*!
 *  selaReadStream()
 *
 *     Input: stream
 *     Return: sela, or null on error
 */
SELA  *
selaReadStream(FILE    *fp)
{
INT32     i, n;
SEL      *sel;
SELA     *sela;

    PROCNAME("selaReadStream");

    if (!fp)
	return (SELA *)ERROR_PTR("stream not defined", procName, NULL);

    if (fscanf(fp, "sel array: number of sels = %d\n\n", &n) != 1)
	return (SELA *)ERROR_PTR("not a sela", procName, NULL);

    if ((sela = selaCreate(n)) == NULL)
	return (SELA *)ERROR_PTR("sela not made", procName, NULL);
    sela->nalloc = n;

    for (i = 0; i < n; i++)
    {
	if ((sel = selReadStream(fp)) == NULL)
	    return (SELA *)ERROR_PTR("sel not made", procName, NULL);
	selaAddSel(sela, sel, NULL, 0);
    }

    return sela;
}


/*!
 *  selReadStream()
 *
 *      Input: stream
 *      Return: sel, or null on error
 */
SEL  *
selReadStream(FILE  *fp)
{
char     *selname;
char      linebuf[BUFSIZE];
INT32     sy, sx, cy, cx, i, j;
SEL      *sel;

    PROCNAME("selReadStream");

    if (!fp)
	return (SEL *)ERROR_PTR("stream not defined", procName, NULL);

    fgets(linebuf, BUFSIZE, fp);
    selname = stringNew(linebuf);
    sscanf(linebuf, "  ------  %s  ------", selname);

    if (fscanf(fp, "  sy = %d, sx = %d, cy = %d, cx = %d\n",
	    &sy, &sx, &cy, &cx) != 4)
	return (SEL *)ERROR_PTR("dimensions not read", procName, NULL);

    if ((sel = selCreate(sy, sx, selname)) == NULL)
	return (SEL *)ERROR_PTR("sel not made", procName, NULL);
    sel->cy = cy;
    sel->cx = cx;

    for (i = 0; i < sy; i++) {
	fscanf(fp, "    ");
	for (j = 0; j < sx; j++)
	    fscanf(fp, "%1d", &sel->data[i][j]);
	fscanf(fp, "\n");
    }
    fscanf(fp, "\n");

    return sel;
}

