/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *  utils.c
 *
 *           error and warning procs
 *               INT32      errorInt()
 *               FLOAT32    errorFloat()
 *               void      *errorPtr()
 *               void       errorVoid()
 *               void       warning()
 *               void       warningInt()
 *
 *           safe string procs
 *               char      *stringNew()
 *               INT32      stringReplace()
 *               char      *stringJoin()
 *
 *           safe realloc
 *               void      *reallocNew()
 *
 *           file to memory
 *               UINT8     *fileToArray()
 *               UINT8     *streamToArray()
 *               INT32      nbytesInFile()
 *
 *           memory to file
 *               INT32      arrayToFile()
 *
 *           data conversion machine-end <--> little-end
 *               UINT16     convertLittleEnd16()
 *               UINT32     convertLittleEnd32()
 *
 *           file opening and name operations
 *               FILE      *fopenReadStream()
 *               INT32      splitPathname()
 *               char      *genPathname()
 *
 *           timing
 *               void       startTimer()
 *               FLOAT32    stopTimer()
 *
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "allheaders.h"



/*----------------------------------------------------------------------------*
 *                      Error and Warning Procs                            
 *----------------------------------------------------------------------------*/
/*
 *  These all print messages to stderr
 */

/*!
 *  errorInt()
 *       Input:  msg  (error message)
 *               procname
 *               ival (return this, typically 1)
 */
INT32
errorInt(char	*msg, 
	 char	*procname, 
	 INT32	 ival)
{

    fprintf(stderr, "Error in %s: %s\n", procname, msg);
    return ival;
}


/*!
 *  errorFloat()
 *        Input:  msg  (error message)
 *                procname
 *                fval  (return this)
 */
FLOAT32
errorFloat(char     *msg, 
	   char     *procname, 
	   FLOAT32   fval)
{

    fprintf(stderr, "Error in %s: %s\n", procname, msg);
    return fval;
}


/*!
 *  errorPtr()
 *        Input:  msg  (error message)
 *                procname
 *                pval  (return this; typically null)
 */
void *
errorPtr(char	*msg,
 	 char	*procname, 
	 void	*pval)
{

    fprintf(stderr, "Error in %s: %s\n", procname, msg);
    return (void*)pval;
}


/*!
 *  errorVoid()
 *        Input:  msg  (error message)
 *                procname
 */
void
errorVoid(char	*msg, 
	  char	*procname)
{
    fprintf(stderr, "Error in %s: %s\n", procname, msg);
    return;
}


/*!
 *  warning()
 *        Input:  msg  (warning message)
 *                procname
 *        Return: void
 */
void
warning(char   *msg,
        char   *procname)
{
    fprintf(stderr, "Warning in %s: %s\n", procname, msg);
    return;
}


/*!
 *  warningInt()
 *        Input:  msg  (warning message)
 *                procname
 *                ival (embedded in warning message via %d)
 */
void
warningInt(char   *msg,
	   char   *procname, 
	   INT32   ival)
{
INT32    bufsize;
char    *charbuf;

    if (!msg || !procname) {
	ERROR_VOID("msg or procname not defined in warningInt()", procname);
	return;
    }

    bufsize = strlen(msg) + strlen(procname) + 128;
    if ((charbuf = (char *)CALLOC(bufsize, sizeof(char))) == NULL) {
	ERROR_VOID("charbuf not made in warningInt()", procname);
	return;
    }

    sprintf(charbuf, "Warning in %s: %s\n", procname, msg);
    fprintf(stderr, charbuf, ival);

    FREE((void *)charbuf);
    return;
}



/*--------------------------------------------------------------------*
 *                       Safe string operations                       *
 *--------------------------------------------------------------------*/
/*!
 *  stringNew()
 *
 *        Input:  src string
 *        Return: dest copy of src string, or null on error
 */
char *
stringNew(char  *src) 
{
char       *dest;

    PROCNAME("stringNew");

    if (!src)
	return (char *)ERROR_PTR("src not defined", procName, NULL);
    
    if ((dest = (char *)CALLOC(strlen(src) + 2, sizeof(char))) == NULL)
	return (char *)ERROR_PTR("dest not made", procName, NULL);
    strcpy(dest, src);

    return dest;
}
    

/*!
 *  stringReplace()
 *
 *        Input:  &dest string (<return> copy)
 *                src string
 *        Return: 0 if OK; 1 on error
 *        Action: (1) frees any existing dest string
 *                (2) puts a copy of src string in the dest
 *        Nonstandard usage: does reasonable thing if either
 *                           or both strings are null
 */
INT32
stringReplace(char  **pdest,
	      char   *src)
{
char       *scopy;

    PROCNAME("stringReplace");

    if (!pdest)
        return ERROR_INT("pdest not defined", procName, 1);

    if (*pdest)
	FREE((void *)(*pdest));
    
    if (src) {
	if ((scopy = (char *)CALLOC(strlen(src) + 2, sizeof(char))) == NULL)
	    return ERROR_INT("scopy not made", procName, 1);
	strcpy(scopy, src);
	*pdest = scopy;
    }
    else
	*pdest = NULL;

    return 0;
}
    

/*!
 *  stringJoin()
 *
 *     Safe version of strcat.
 *
 *     Note: It is not an error if either or both of the strings
 *     are empty, or if either or both the pointers are null.
 *
 *     Input:  src1 string (<optional>)
 *             src2 string (<optional>)
 *     Return: concatenated string, or null on error
 */
char *
stringJoin(char  *src1, 
	   char  *src2)
{
char      *dest;
INT32      srclen1, srclen2, destlen;

    PROCNAME("stringJoin");

    srclen1 = srclen2 = 0;
    if (src1)
	srclen1 = strlen(src1);
    if (src2)
	srclen2 = strlen(src2);
    destlen = srclen1 + srclen2 + 3;

    if ((dest = (char *)CALLOC(destlen, sizeof(char))) == NULL)
        return (char *)ERROR_PTR("calloc fail for dest", procName, NULL);

    if (src1)
	strcpy(dest, src1);
    if (src2)
	strcat(dest, src2);
    return dest;
}



/*--------------------------------------------------------------------*
 *                             Safe realloc                           *
 *--------------------------------------------------------------------*/
/*!
 *  reallocNew()
 *          Input:  &indata (<used and returned>)
 *                  size of input data to be copied (bytes)
 *                  size of data to be reallocated (bytes)
 *          Return: ptr to new data, or null on error
 *
 *          Action:  !N.B. (3) and (4)!
 *                (1) allocates memory, initialized to 0.
 *                (2) copies as much of the input data as possible
 *                    to the new block, truncating the copy if
 *                    necessary.
 *                (3) frees the input data
 *                (4) zeroes the input data ptr
 *
 *        Notes on nonstandard use:
 *              (1) if newsize <=0, just frees input data and nulls ptr
 *              (2) if input ptr is null, just callocs new memory
 */
void *
reallocNew(void  **pindata,
	   INT32   oldsize,
	   INT32   newsize)
{
INT32       minsize;
void       *indata; 
void       *newdata;

    PROCNAME("reallocNew");

    if (!pindata)
	return ERROR_PTR("input data not defined", procName, NULL);

    indata = *pindata;

    if (newsize <= 0) {   /* nonstandard usage */
	if (indata) {
	    FREE((void *)indata);
	    *pindata = NULL;
	}
	return NULL;
    }

    if (!indata)   /* nonstandard usage */
    {
	if ((newdata = (void *)CALLOC(1, newsize)) == NULL)
	    return ERROR_PTR("newdata not made", procName, NULL);
	return newdata;
    }

	/* standard usage */
    if ((newdata = (void *)CALLOC(1, newsize)) == NULL)
	return ERROR_PTR("newdata not made", procName, NULL);

    minsize = MIN(oldsize, newsize);
    memcpy((char *)newdata, (char *)indata, minsize);

    FREE((void *)indata);
    *pindata = NULL;

    return newdata;
}
    


/*--------------------------------------------------------------------*
 *                       Reading Bytes from File                      *
 *--------------------------------------------------------------------*/
/*!
 *  fileToArray()
 *
 *       Input: filename
 *              &nbytes (<return> number of bytes read)
 *       Return: array, or null on error
 *
 */
UINT8 *
fileToArray(char   *fname, 
 	    INT32  *pnbytes)
{
UINT8    *data;
FILE     *fp;

    PROCNAME("fileToArray");

    if (!fname)
        return (UINT8 *)ERROR_PTR("fname not defined", procName, NULL);
    if (!pnbytes)
        return (UINT8 *)ERROR_PTR("pnbytes not defined", procName, NULL);
    *pnbytes = 0;

    if ((fp = fopen(fname, "r")) == NULL)
        return (UINT8 *)ERROR_PTR("file stream not opened", procName, NULL);

    data = streamToArray(fp, pnbytes);
    fclose(fp);

    return data;
}


/*!
 *  streamToArray()
 *
 *       Input:  stream
 *               &nbytes (<return> number of bytes read)
 *       Return: null-terminated array, or null on error
 *               (reading 0 bytes is not an error)
 *
 *     N.B. -- Side effect:
 *           This always re-positions stream ptr to the
 *           beginning of the file.
 */
UINT8 *
streamToArray(FILE   *fp, 
	      INT32  *pnbytes)
{
UINT8    *data;

    PROCNAME("streamToArray");

    if (!fp)
        return (UINT8 *)ERROR_PTR("stream not defined", procName, NULL);
    if (!pnbytes)
        return (UINT8 *)ERROR_PTR("ptr to nbytes not defined", procName, NULL);

    *pnbytes = nbytesInFile(fp);

    if ((data = (UINT8 *)CALLOC(1, *pnbytes + 1)) == NULL)
        return (UINT8 *)ERROR_PTR("CALLOC fail for data", procName, NULL);
    fread(data, *pnbytes, 1, fp);

    return data;
}


/*!
 *  nbytesInFile()
 *
 *     Input: file stream
 *     Return: nbytes in file; 0 on error
 */
INT32
nbytesInFile(FILE  *fp)
{
INT32    nbytes, pos;

    PROCNAME("nbytesInFile");

    if (!fp)
	return ERROR_INT("stream not open", procName, 0);

    pos = ftell(fp);          /* initial position */
    fseek(fp, 0, SEEK_END);   /* EOF */
    nbytes = ftell(fp);
    fseek(fp, 0, pos);        /* back to initial position */

    return nbytes;
}



/*--------------------------------------------------------------------*
 *                          Writing Bytes to File                     *
 *--------------------------------------------------------------------*/
/*!
 *  arrayToFile()
 *
 *       Input:  filename (output)
 *               operation  ("w" for write; "a" for append)
 *               data  (binary data to be written)
 *               nbytes  (size of data array)
 *       Return: 0 if OK; 1 on error
 */
INT32
arrayToFile(char  *filename,
	    char  *operation,
	    void  *data,
	    INT32  nbytes)
{
FILE     *fp;

    PROCNAME("arrayToFile");

    if (!filename)
	return ERROR_INT("filename not defined", procName, 1);
    if (!operation)
	return ERROR_INT("operation not defined", procName, 1);
    if (!data)
	return ERROR_INT("data not defined", procName, 1);
    if (nbytes <= 0)
	return ERROR_INT("nbytes must be > 0", procName, 1);

    if (!strcmp(operation, "w") && !strcmp(operation, "a"))
	return ERROR_INT("operation not one of {'w','a'}", procName, 1);

    if ((fp = fopen(filename, operation)) == NULL)
	return ERROR_INT("stream not opened", procName, 1);
    fwrite(data, 1, nbytes, fp);
    fclose(fp);

    return 0;
}


/*--------------------------------------------------------------------*
 *         Data conversions machine-endian to little endian           *
 *                                                                    * 
 *      These two procedures swap bytes if on a big-endian machine.   *
 *      They are used for two things:                                 *
 *          (1) conversion to machine endian from little-endian       *
 *              data that was read from a file                        *
 *          (2) conversion to little-endian from machine-endian       *
 *              data before writing little-endian data to file        *
 *      We use them in bmpio.c, which reads and writes data           *
 *      in little-endian format from and to files.                    *
 *                                                                    * 
 *--------------------------------------------------------------------*/

/*--------------------------------------------------------------------*
 *                        convertLittleEnd16()                        *
 *--------------------------------------------------------------------*/
#ifdef _BIG_ENDIAN_

UINT16
convertLittleEnd16(UINT16  shortin)
{
UINT16      shortout;

    shortout = (shortin << 8) | (shortin >> 8);
    return shortout;
}

#else     /* _LITTLE_ENDIAN_ */

UINT16
convertLittleEnd16(UINT16  shortin)
{
    return  shortin;
}

#endif  /* _BIG_ENDIAN_ */


/*--------------------------------------------------------------------*
 *                        convertLittleEnd32()                        *
 *--------------------------------------------------------------------*/
#ifdef _BIG_ENDIAN_

UINT32
convertLittleEnd32(UINT32   wordin)
{
UINT32   wordout;

    wordout = (wordin << 24) | ((wordin << 8) & 0x00ff0000) |
              ((wordin >> 8) & 0x0000ff00) | (wordin >> 24);
    return wordout;
}

#else  /*  _LITTLE_ENDIAN_ */

UINT32
convertLittleEnd32(UINT32   wordin)
{
    return wordin;
}

#endif  /* _BIG_ENDIAN_ */



/*--------------------------------------------------------------------*
 *                         Opening read stream                        *
 *--------------------------------------------------------------------*/
/*!
 *  fopenReadStream()
 *
 *    Input:  filename 
 *    Return: stream or null on error
 */
FILE *
fopenReadStream(char   *filename)
{
char     *tail;
FILE     *fp;

    PROCNAME("fopenReadStream");

    if (!filename)
	return (FILE *)ERROR_PTR("filename not defined", procName, NULL);

	/* try input filename */
    if (fp = fopen(filename, "rb"))
	return fp;

	/* else, strip directory and try locally */
    splitPathname(filename, NULL, &tail);
    if (fp = fopen(tail, "rb")) {
	FREE((void *)tail);
	return fp;
    }
    FREE((void *)tail);

    return (FILE *)ERROR_PTR("file not found", procName, NULL);
}


/*!
 *  splitPathname()
 *
 *      Input:   pathname  (full path; can be a directory)
 *              &dir  (<optional return> root directory name of
 *                     input path, without trailing '/')
 *              &tail (<optional return> path tail, which is either
 *                     the file name within the root directory or
 *                     the last sub-directory in the path)
 *      Return:  0 if OK, 1 on error
 *       
 *      Note: (1) if you only want the tail, input null for
 *                the root directory ptr.
 *            (2) if you only want the root directory name,
 *                input null for the tail ptr.
 */
INT32
splitPathname(char  *pathname,
              char **pdir,
              char **ptail)
{
char    *cpathname, *lastslash;
INT32    len;

    PROCNAME("splitPathname");

    if (!pathname)
        return ERROR_INT("pathname not defined", procName, 1);
    if (!pdir && !ptail)
        return ERROR_INT("null input for both strings", procName, 1);

        /* Make a copy and remove any trailing '/'.
	 * That way, this will also work on directory names
	 * that are written with a trailing '/'   */
    cpathname = stringNew(pathname);
    len = strlen(cpathname);
    if (cpathname[len - 1] == '/')
        cpathname[len - 1] = '\0';
     
    if (lastslash = strrchr(cpathname, '/')) {
	if (ptail)
	    *ptail = stringNew(lastslash + 1);
	if (pdir) {
	    *lastslash = '\0';
	    *pdir = cpathname;
	}
	else
	    FREE((void *)cpathname);
    }
    else {  /* no directory */
	if (pdir)
	    *pdir = stringNew("");
	if (ptail)
	    *ptail = cpathname;
	else
	    FREE((void *)cpathname);
    }

    return 0;
}


/*! 
 *  genPathname()
 *
 *      Input:   dir  (directory name, with or without trailing '/')
 *               fname  (file name within the directory)
 *      Return:  full pathname, or null on error
 */
char *
genPathname(char  *dir,
            char  *fname)
{
INT32     dirlen, namelen;
char     *charbuf;
    
    PROCNAME("genPathname");

    if (!dir)
	return (char *)ERROR_PTR("dir not defined", procName, NULL);
    if (!fname)
	return (char *)ERROR_PTR("fname not defined", procName, NULL);

    dirlen = strlen(dir);
    namelen = strlen(fname);
    if ((charbuf = (char *)CALLOC(dirlen + namelen + 10, sizeof(char)))
            == NULL)
	return (char *)ERROR_PTR("charbuf not made", procName, NULL);

    if (dir[dirlen - 1] != '/')
	sprintf(charbuf, "%s/", dir);
    else
	strcpy(charbuf, dir);
    strcat(charbuf, fname);
    return charbuf;
}



/*----------------------------------------------------------------------------*
 *                           Timing Procs                            
 *----------------------------------------------------------------------------*/
#include <sys/time.h>
#include <sys/resource.h>

static struct rusage rusage_before;
static struct rusage rusage_after;

/*
 *  Example of use:
 *
 *     startTimer();
 *     ....
 *     fprintf(stderr, "Elapsed time = %7.3f sec\n", stopTimer());
 */

void
startTimer(void)
{
    getrusage(RUSAGE_SELF, &rusage_before);
}


FLOAT32
stopTimer(void)
{
INT32    tsec, tusec;

    getrusage(RUSAGE_SELF, &rusage_after);

    tsec = rusage_after.ru_utime.tv_sec - rusage_before.ru_utime.tv_sec;
    tusec = rusage_after.ru_utime.tv_usec - rusage_before.ru_utime.tv_usec;

    return (tsec + ((FLOAT32)tusec) / 1000000.0);
}


