/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 * writefile.c
 *
 *     Top-level procedures for writing images to file:
 *        INT32    pixWrite() 
 *        INT32    pixWriteJpeg() 
 *        INT32    pixWriteStream() 
 *
 *     X Display using xv
 *        INT32    pixDisplay()
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "allheaders.h"

    /* for xv display */
#define   BUF_SIZE                256
#define   MAX_DISPLAY_WIDTH      1000
#define   MAX_DISPLAY_HEIGHT      800

    /* PostScript output for printing */
#define   DEFAULT_SCALING         1.0



/*---------------------------------------------------------------------*
 *           Top-level procedures for writing images to file           *
 *---------------------------------------------------------------------*/
/*!
 *  pixWrite()
 *
 *      Input:   filename
 *               pix
 *               format  (defined in imageio.h)
 *      Return: 0 if OK; 1 on error
 */
INT32
pixWrite(char    *filename, 
	 PIX     *pix, 
         INT32    format)
{
FILE   *fp;

    PROCNAME("pixWrite");

    if (!pix)
        return ERROR_INT("pix not defined", procName, 1);
    if (!filename)
	return ERROR_INT("filename not defined", procName, 1);

    if ((fp = fopen(filename, "w+")) == NULL)
	return ERROR_INT("stream not opened", procName, 1);

    if (pixWriteStream(fp, pix, format)) {
	fclose(fp);
        return ERROR_INT("pix not written to stream", procName, 1);
    }

    fclose(fp);
    return 0;
}


/*!
 *  pixWriteJpeg()
 *
 *      Input:   filename
 *               pix
 *               quality (1 - 100; 75 is default)
 *      Return: 0 if OK; 1 on error
 *      Note: this is a special function for jfif-jpeg only
 */
INT32
pixWriteJpeg(char    *filename, 
	     PIX     *pix, 
             INT32    quality)
{
FILE   *fp;

    PROCNAME("pixWriteJpeg");

    if (!pix)
        return ERROR_INT("pix not defined", procName, 1);
    if (!filename)
	return ERROR_INT("filename not defined", procName, 1);

    if ((fp = fopen(filename, "w+")) == NULL)
	return ERROR_INT("stream not opened", procName, 1);

    if (pixWriteStreamJpeg(fp, pix, quality)) {
	fclose(fp);
        return ERROR_INT("pix not written to stream", procName, 1);
    }

    fclose(fp);
    return 0;
}


/*!
 *  pixWriteStream()
 *
 *       Input:  stream
 *               pix
 *               format
 *       Return: 0 if OK; 1 on error.
 */
INT32
pixWriteStream(FILE   *fp, 
	       PIX    *pix, 
	       INT32   format)
{

    PROCNAME("pixWriteStream");

    if (!fp)
        return ERROR_INT("stream not defined", procName, 1);
    if (!pix)
        return ERROR_INT("pix not defined", procName, 1);

    switch(format)
    {
    case IFF_BMP:
        pixWriteStreamBmp(fp, pix);
        break;

    case IFF_JFIF_JPEG:
	return pixWriteStreamJpeg(fp, pix, 75);  /* default quality */
	break;
    
    case IFF_PNG:
	return pixWriteStreamPng(fp, pix);
	break;
    
    case IFF_PS:
	return pixWriteStreamPS(fp, pix, NULL, 0, DEFAULT_SCALING);
	break;
    
    default:
        return ERROR_INT("unknown format", procName, 1);
        break;
    }

    return 0;
}


/*---------------------------------------------------------------------*
 *                            X Display using xv                       *
 *---------------------------------------------------------------------*/
/*!
 *  pixDisplay()
 *
 *      Input:   pix
 *               x, y  (location of xv frame)
 *      Return: 0 if OK; 1 on error
 *
 *      Note: This uses xv to display.  Because xv reduces images
 *            to fit the screen, we do this reduction in advance,
 *            before writing out to file.  This function both erases and
 *            writes files in the current directory with the name
 *            "junk_temp_display.*"
 *            Also, this function uses a static internal variable to
 *            number output files written by a single process.  Behavior
 *            with a shared library may be unpredictable.
 */
INT32
pixDisplay(PIX     *pixs,
           INT32    x,
	   INT32    y)
{
static INT32  index = 0;
INT32    w, h, d;
FLOAT32  ratw, rath, ratmin;
char    *tempname;
char     buffer[BUF_SIZE];
PIX     *pixt;

    PROCNAME("pixDisplay");

    if (!pixs)
        return ERROR_INT("pixs not defined", procName, 1);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    d = pixGetDepth(pixs);
    if (w <= MAX_DISPLAY_WIDTH && h <= MAX_DISPLAY_HEIGHT)
	pixt = pixClone(pixs);
    else {
	ratw = (FLOAT32)MAX_DISPLAY_WIDTH / (FLOAT32)w;
	rath = (FLOAT32)MAX_DISPLAY_HEIGHT / (FLOAT32)h;
	ratmin = MIN(ratw, rath);
	if (ratmin < 0.33 && d == 1)
	    pixt = pixScaleToGray4(pixs);
	else if (ratmin < 0.5 && d == 1)
	    pixt = pixScaleToGray3(pixs);
        else
	    pixt = pixScale(pixs, ratmin, ratmin);
	if (!pixt)
	    return ERROR_INT("pixt not made", procName, 1);
    }

    if (index == 0) {
	sprintf(buffer, "rm -f junk_temp_display.*");
	system(buffer);
    }

    index++;
    sprintf(buffer, "junk_temp_display.%d", index);
    tempname = stringNew(buffer);
    if (pixGetDepth(pixt) == 1)
	pixWrite(tempname, pixt, IFF_BMP);
    else
	pixWrite(tempname, pixt, IFF_JFIF_JPEG);

    sprintf(buffer, "/usr/X11R6/bin/xv -quit -geometry +%d+%d %s &",
            x, y, tempname);
    system(buffer);

    pixDestroy(&pixt);
    FREE((void *)tempname);

    return 0;

}


