/**@file anacutoff.h 
 *\brief defines a struct to hold cutoffs and constants for the program, and define their default values as well.
*/

#ifndef ANACUTOFF_H
#define ANACUTOFF_H

/*default values*/
/** \brief the default value for the minimum marker roughness for marker checking*/
#define DEFAULT_MARKER_ROUGHNESS_CUTOFF 0.5
/** \brief the default value for the minimum marker correlation for marker checking*/
#define DEFAULT_MARKER_CORRELATION_CUTOFF 0.975
/** \brief the default value for the maximum number of interpolated lanes allowed in a zone*/
#define DEFAULT_INTERPOLATED_DATA_CUTOFF 40
/** \brief the default value for the maximum number of interpolated markers allowed in a zone*/
#define DEFAULT_INTERPOLATED_MARKER_CUTOFF 15
/** \brief the default value for the maximum number of consecutive interpolated lanes allowed in a zone*/
#define DEFAULT_CON_INTERPOLATED_DATA_CUTOFF 5
/** \brief the default value for the maximum number of consecutive interpolated markers allowed in a zone*/
#define DEFAULT_CON_INTERPOLATED_MARKER_CUTOFF 3
/** \brief the default value for the number segments into which zones are grouped for marker roughness and correlation determination*/
#define DEFAULT_MARKERCORSEGMENTS 3
/** \brief the default value for the maximum number of outliers corrected*/
#define DEFAULT_NOUTLIER_CUTOFF 10

/** \brief the default factor defining the range of acceptable lane spacing*/
#define DEFAULT_LANESPACINGCHECK_FACTOR .2

/** \brief the default minimum correlation of the two border markers for them to be accepted*/
#define DEFAULT_BORDER_COR_HARDCUTOFF 0.90
/** \brief the default minimum correlation of the two border markers not to raise a flag*/
#define DEFAULT_BORDER_COR_SOFTCUTOFF 0.98
/** \brief the default minimum correlation of the two border markers with their neighbours*/
#define DEFAULT_BORDER_NEIGHBOUR_COR_CUTOFF 0.98
/** \brief the default minimum roughness of the two border markers*/
#define DEFAULT_BORDER_ROUGHNESS_CUTOFF 1.8

/** \brief the default factor used to determine the span of the asymmetry calculation*/
#define DEFAULT_ASYM_WINDOW_FACTOR 0.58
/** \brief the default n for the nIQR test used to determine if an asymmetry is an outlier*/
#define DEFAULT_ASYM_WHISKER_FACTOR 3.0
/** \brief the default minimum number of consecutive asymmetric lanes that requires re-matching*/
#define DEFAULT_CON_ASYM_SOFT_CUTOFF 3
/** \brief the default minimum number of consecutive asymmetric lanes that will be flagged*/
#define DEFAULT_CON_ASYM_HARD_CUTOFF 5
/** \brief the default minimum difference between the intensity of pixel pairs for the difference to be included in the score*/
#define DEFAULT_ASYM_THRESHOLD (2.0/255.0)
/** \brief the default maximum offset of the asymmetric datum can be from the identified center of the lane*/
#define DEFAULT_ASYM_MAX_OFFSET (2.0/19.0)

/** \brief the default threshold used for working out thresholded roughness for contamination detection*/
#define DEFAULT_SLOPE_THRESHOLD 1
/** \brief the default factor used to determine the offset for contamination detection comparison*/
#define DEFAULT_MATCH_MAXOFFSET_FACTOR 20000.0
/** \brief the default factor used to determine the minimum valid thresholded roughness for contamination detection*/
#define DEFAULT_MATCH_MIN_ROUGHNESS_FACTOR 50.0
/** \brief the default minimum threshold used to calculate the thresholded roughness for contamination detection*/
#define DEFAULT_MATCH_MIN_VALID_THRESHOLD 1
/** \brief the default cutoff value above which the two lanes are declared contaminations*/
#define DEFAULT_MATCH_CONTAM_CUTOFF 0.5
/** \brief the default factor used to determine the roughness below which a lane is declared non-recombinant*/
#define DEFAULT_NON_RECOM_THRESROUGH_FACTOR 80.0
/** \brief the default minimum maxdiff for a lane to be classified as non-recombinant*/
#define DEFAULT_NON_RECOM_MAXDIFF 10

/*for noise galaxy detection*/
/** \brief default value noise to foreground ratio used in noise galaxy detection*/
#define DEFAULT_NOISE_TO_FG_RATIO 0.35
/** \brief default minimum intensity of foreground for pixel to be considered in noise galaxy detection*/
#define DEFAULT_MIN_FG 10
/** \brief default number of consecutive pixels labelled as noise for the lane to be labelled as noisy (in multiples of lanespacing*/
#define DEFAULT_NOISE_SPAN_FACTOR 1.0

/** \brief struct to hold all "magic numbers" for easy passing*/
struct AnaCutoff {
  /*for border determination*/
  /** \brief the minimum correlation of the two border markers for them to be accepted*/
  float border_cor_hardcutoff;
/** \brief the minimum correlation of the two border markers not to raise a flag*/
  float border_cor_softcutoff;
/** \brief the minimum correlation of the two border markers with their neighbours*/
  float border_neighbour_cor_cutoff;
/** \brief the minimum roughness of the two border markers*/
  float border_roughness_cutoff;

 /*for result checks*/
/** \brief the value for the minimum marker roughness for marker checking*/
  float marker_roughness_cutoff;
/** \brief the value for the minimum marker correlation for marker checking*/
  float marker_correlation_cutoff;
/** \brief the value for the maximum number of interpolated lanes allowed in a zone*/
  int interpolated_data_cutoff;
/** \brief the value for the maximum number of interpolated markers allowed in a zone*/
  int interpolated_marker_cutoff;
/** \brief the value for the maximum number of consecutive interpolated lanes allowed in a zone*/
  int consec_data_cutoff;
/** \brief the value for the maximum number of consecutive interpolated markers allowed in a zone*/
  int consec_marker_cutoff;
/** \brief the value for the number segments into which zones are grouped for marker roughness and correlation determination*/
  int markercorsegments;
/** \brief the value for the maximum number of outliers corrected*/
  int noutlier_cutoff;
/** \brief the factor defining the range of acceptable lane spacing*/
  float lanespacingcheck_factor;

  /*for asymmetry*/
/** \brief the factor used to determine the size of the window used for the asymmetry calculation*/
  float asym_window_factor;
/** \brief the n for the nIQR test used to determine if an asymmetry is an outlier*/
  float asym_whisker_factor;
/** \brief the minimum number of consecutive asymmetric lanes that requires re-matching*/
  int consec_asym_soft_cutoff;
/** \brief the minimum number of consecutive asymmetric lanes that will be flagged*/
  int consec_asym_hard_cutoff;
/** \brief the minimum difference between the intensity of pixel pairs for the difference to be included in the score*/
  float asym_threshold;
/** \brief the maximum offset of the asymmetric datum can be from the identified center of the lane*/
  float asym_max_offset;

  /*for no growth detection*/
/** \brief the threshold used for working out thresholded roughness for contamination detection*/
  int slope_threshold;
  /*for contam matching*/
/** \brief the factor used to determine the offset for contamination detection comparison*/
  float match_maxoffset_factor;
/** \brief the factor used to determine the minimum valid thresholded roughness for contamination detection*/
  float match_min_roughness_factor;
/** \brief the minimum threshold used to calculate the thresholded roughness for contamination detection*/
  int match_min_valid_threshold;
/** \brief the cutoff value above which the two lanes are declared contaminations*/
  float match_contam_cutoff;
/** \brief the factor used to determine the roughness below which a lane is declared non-recombinant
 *imagewidth/factor gives the min thresrough for lane to be considered for contamination
*/
  float non_recom_thresrough_factor; 
/** \brief the minimum maxdiff for a lane to be classified as non-recombinant*/
  int non_recom_maxdiff;

  /*for noise galaxy detection*/
/** \brief noise to foreground ratio used in noise galaxy detection*/
  float noise_to_FG_ratio;
/** \brief the minimum intensity of foreground for pixel to be considered in noise galaxy detection*/
  int min_FG;
/** \brief number of consecutive pixels labelled as noise for the lane to be labelled as noisy (in multiples of lanespacing*/
  float noise_span_factor;
};

#endif
