/**@file checknogrowths.c
 *\brief contains functions to extract the median row for each lane and 
 *outputs the thresholded roughness and maxdiff to a file 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "checknogrowths.h"
#include "asymmetry.h"

/**\brief go down an image and place the median value in each column into an array
 *@param laneintensity pointer to 2D array holding the straightened lane
 *@param imageWidth width of the 2D array
 *@param imageHeight height of the 2D array
 *@param median array of the median value in each column of the lane
*/
void extractRow(float **laneintensity, int imageWidth, int imageHeight, float* median)
{
  int col, row;
  int medianindex=(int) ((float)imageHeight/2.0);
  float *rowvalue=calloc(imageHeight, sizeof(float));  

  for(col = 0; col < imageWidth; col++)
    {
      for(row = 0; row <imageHeight; row++)
	{
	  rowvalue[row] =  laneintensity[col][row];
	}
      shellSort(rowvalue, imageHeight);
      median[col] = rowvalue[medianindex];
    }
  free(rowvalue);
}

/**\brief threshold valuein to be vector of 1, 0, -1 based if the running difference is larger than, equal or smaller than the threshold value.  
*
*elements before start and after stop are not changed
*also returns the absolute maximum of slope in the area of interest
*
*@param valuein array of values to be thresholded
*@param valueout array where results is written to
*@param size size of the array
*@param start first element to be tested
*@param stop last element to be tested
*@param threshold threshold used
*@returns the absolute maximum of the slope
*/
static float slopeSign(float* valuein, int* valueout, int size, int start, int stop, int threshold)
{
  int i;
  int slope;
  float maxdiff=0;
  
  for (i=start; i<stop; i++)
    {
      slope = valuein[i+1]- valuein[i];
      if (fabs(slope)>maxdiff) maxdiff=fabs(slope);
      if (slope > threshold) valueout[i]=1;
      else if (slope < -1*threshold) valueout[i]=-1;
      else valueout[i]=0;
    }
  return maxdiff;

}

/**\brief finds the similarity between two thresholded vectors with values of {-1,0,1}.  
 *
 *the number of columns where the vectors have the same non-zero value is counted
 *an offset is allowed and the best score is returned
 *
 *@param valuein1 array of first vector for comparison
 *@param valuein2 array of second vector for comparison
 *@param start first element to be compared
 *@param stop last element to be compared
 *@param maxoffset maximum offset allowed for comparison
 *@returns the number of elements that are similar
 */ 
static float Similarity(int* valuein1, int* valuein2, int start, int stop, int maxoffset)
{
  int  i, j;
  int  value1, value2;
  int	offset, bestoff;
  int diffsum, bestsum;
  bestoff = 0;
  bestsum = 0; /* 2*(stop-start); */
  for (offset = -maxoffset; offset <= maxoffset; offset++) {
    diffsum=0;
    for (i = start; i < stop; i++) {
      value1 =  valuein1[i];
      value2 = valuein2[i+offset];
      if((value1==1 && value2==1) || (value1==-1 && value2==-1))diffsum++;
    }
    if (diffsum > bestsum) {
      bestsum = diffsum;
      bestoff = offset;
    }
  }
  return (float) bestsum;
}

/**\brief compares the similarity between two 1D array using thresholded roughness
 *  
 *The threshold is varied until the absolute sum of the thresholded vector is just below 
 *a certain number to correct for intensity differences.
 *The two thresholded vector is tested for similarity and the score is normalized by the 
 *smaller of the two abs sum of the thresholded vectors.
 *Effectively, a score of 1 means that one of the signal is completely embedded in the other one.
 *The score is returned
 *
 *@param row1 the first row for the comparison
 *@param row2 the second row for comparison
 *@param start the leftmost column for comparison
 *@param stop the rightmost column for comparison
 *@param GelInfo holds various gel constants
 *@param AnaCutoff holds various analysis cutoff constants
 *@param Verbose turns on text output on screen
 *@returns similarity score [0,1], 1 being one signal is completely embedded in the other
 */
static float compareRows(float *row1, float *row2, int start, int stop, struct GelInfo *GelInfo, struct AnaCutoff *AnaCutoff, int Verbose)
{
   int *slope1, *slope2;
   slope1 =  calloc(GelInfo->ImageWidth-1, sizeof(int));
   slope2 =  calloc(GelInfo->ImageWidth-1, sizeof(int));
   
   int roughness1, roughness2, lastroughness;
   roughness1 = 0;
   roughness2 = 0;
   int index;
   int threshold;
   int maxoffset = (int)((float)GelInfo->ImageWidth / (float) AnaCutoff->match_maxoffset_factor + 0.5);
   float similar;


   slopeSign(row1, slope1, GelInfo->ImageWidth, start, stop, 0);
   slopeSign(row2, slope2, GelInfo->ImageWidth, start, stop, 0);

   for(index =0; index< GelInfo->ImageWidth-1; index++)
     {
       roughness1 +=abs(slope1[index]);
       roughness2 += abs(slope2[index]);
     }

  /*threshold is set so that thresholded roughness just before width / min_roughness_factor, 
  *to normalize for intensity difference*/
  threshold = 0;
  while(roughness1 > (float)(stop-start)/AnaCutoff->match_min_roughness_factor)
    {
      threshold++;
      slopeSign(row1, slope1, GelInfo->ImageWidth, start, stop, threshold);
      lastroughness = roughness1;
      roughness1 = 0;
      for(index =0; index<GelInfo->ImageWidth-1; index++)
	{
	  roughness1 +=abs(slope1[index]);
	}
    }
  if (threshold<AnaCutoff->match_min_valid_threshold) return 0;
  else
    {
      roughness1 = lastroughness;
      slopeSign(row1, slope1, GelInfo->ImageWidth, start, stop, threshold-1);
    }

  threshold = 0;
  while(roughness2 > (float)(stop-start)/AnaCutoff->match_min_roughness_factor)
    {
      threshold++;
      slopeSign(row2, slope2, GelInfo->ImageWidth, start, stop, threshold);
      lastroughness = roughness2;
      roughness2 = 0;
      for(index =0; index<GelInfo->ImageWidth-2; index++)
	{
	  roughness2 +=abs(slope2[index]);
	}
    }
  if (threshold<AnaCutoff->match_min_valid_threshold) return 0;
  else
    {
      roughness2 = lastroughness;
      slopeSign(row2, slope2, GelInfo->ImageWidth, start, stop, threshold-1);
    }
  
  /*compare the two signals*/
  similar = Similarity(slope1, slope2, start, stop, maxoffset);

  free(slope1);
  free(slope2);
  float minroughness = roughness1>roughness2?(float)roughness2:(float)roughness1;
  /*normalize by the smaller roughness [0,1]*/
  return (similar/minroughness);
}

/**\brief returns 1 if 2 given lanes are neighbouring wells on a 96 well plate
 *
 *hard coded for production gels
 *
 *@param lane1 lane ID of one lane
 *@param lane2 land ID of the other lane
 *@returns 1 if the two lanes are neighbours, 0 otherwise
*/
static int plateNeighbour(int lane1, int lane2)
{
  if(lane1%5==0 || lane2%5==0) return 0; /*markers not on plate*/
  if(abs(lane2-lane1)==5 && !((lane1<=64 &&lane1 >=56) && (lane2<=64 &&lane2 >=56))) return 1; /*EW neighbours*/
  if(abs(lane2-lane1)==1 && lane1%5!=0 && lane2%5!=0) return 1; /*NS neighbours*/
  if((lane2-lane1==57 && lane2%5==1)||(lane1-lane2==57 && lane1%5==1)) return 1; /*between row D and E*/
  return 0;
}
/*returns 1 if 2 given non-marker lanes are neighbouring lanes or off by 5 on a gel*/
static int gelNeighbour(int lane1, int lane2)
{
  if(lane1%5==0 || lane2%5==0) return 0; /*marker pairs not considered*/
  if(abs(lane2-lane1)==5) return 1; /*off by five*/
  if(abs(lane2-lane1)==1) return 1; /*off by one*/
  return 0;
}

/**\brief output the maxdiff and thresrough (threshold 1) of every lane
 *suspected contaminated lanes is marked by negative thresrough
 *
 *@param OUTFILE writable file stream for results output
 *@param pixGel pix representation of gel image used for analysis
 *@param PeakPosition holds y position of all nodes
 *@param NodePosition holds x position of nodes in each zone
 *@param GelInfo holds various gel constants
 *@param AnaCutoff holds various analysis cutoff constants
 *@param Verbose turns on text output on screen
 */
void checkNoGrows(FILE *OUTFILE, PIX *pixGel, float **PeakPosition, float *NodePosition, struct GelInfo *GelInfo, struct AnaCutoff *AnaCutoff, int Verbose)
{
  int lane,col,row;
  float **laneintensity;
  int   halfwidth, output_width;
  int *slope;

  float **medianrow;

  int nogrowthcount=0;
  int* roughness = calloc(GelInfo->NumberOfLanes, sizeof(int));
  float *maxdiff =calloc(GelInfo->NumberOfLanes, sizeof(float));
  float contam_sim_score = 0;
  float non_recom_thresrough=(float)GelInfo->ImageWidth/AnaCutoff->non_recom_thresrough_factor;
  int lane1;

  /*extract the median value of the section of image of each lane*/
  halfwidth = ((int)(GelInfo->LaneSpacing + 0.5)) / 2;  
  output_width = 2 * halfwidth + 1;

  laneintensity=(float**) malloc((GelInfo->ImageWidth) * sizeof(float*));
  medianrow =  calloc(GelInfo->NumberOfLanes, sizeof(float*));
  for(row = 0; row <GelInfo->NumberOfLanes; row++)
    {
        medianrow[row] =  calloc(GelInfo->ImageWidth, sizeof(float));
    }

  slope =  calloc(GelInfo->ImageWidth-1, sizeof(int));
  for (col = 0; col < GelInfo->ImageWidth; col++) 
   laneintensity[col] =(float*)  calloc(output_width, sizeof(float));

  for(lane=0; lane<GelInfo->NumberOfLanes;lane++)
    {
      /*getting lane straightened image of a lane*/
      followLane(pixGel, PeakPosition, NodePosition, GelInfo, lane, laneintensity);
      
      /*extract median row from lane image*/
      extractRow(laneintensity, GelInfo->ImageWidth, output_width, medianrow[lane]);

      /*get the slope of the median row*/
      maxdiff[lane]=slopeSign(medianrow[lane], slope, GelInfo->ImageWidth, GelInfo->StopPreData, (int) NodePosition[GelInfo->NumberOfZones], AnaCutoff->slope_threshold);

      /*get roughness for the lane*/
      for(col =GelInfo->StopPreData; col<NodePosition[GelInfo->NumberOfZones]; col++)
	{
	  roughness[lane] +=abs(slope[col]);    
	} 
    }

  /*checking for contamination*/
  for(lane=0; lane<GelInfo->NumberOfLanes;lane++)
    {
      for(lane1=lane+1; lane1<GelInfo->NumberOfLanes;lane1++)
	{
	  /*if the pair are neighbours on a plate*/
	  if(plateNeighbour(GelInfo->NumberOfLanes-1-lane, GelInfo->NumberOfLanes-1-lane1)||gelNeighbour(GelInfo->NumberOfLanes-1-lane, GelInfo->NumberOfLanes-1-lane1) )
	    {
	      /*compare the rows using thresholded difference*/
	      contam_sim_score = compareRows(medianrow[lane], medianrow[lane1], GelInfo->StopPreData, NodePosition[GelInfo->NumberOfZones],GelInfo, AnaCutoff, Verbose);
	      /*set roughness to be negative if contamination found, since either could have been an empty lane otherwise*/
	      if(contam_sim_score>AnaCutoff->match_contam_cutoff && (abs(roughness[lane])>non_recom_thresrough || maxdiff[lane]<AnaCutoff->non_recom_maxdiff) && (abs(roughness[lane1])>non_recom_thresrough||maxdiff[lane1]<AnaCutoff->non_recom_maxdiff))
		{
		  roughness[lane] = -1*abs(roughness[lane]);
		  roughness[lane1] = -1*abs(roughness[lane1]);
		  if(Verbose)fprintf(stderr, "found contam at lane %i and %i with score of %f\n", GelInfo->NumberOfLanes-1- lane, GelInfo->NumberOfLanes-1-lane1, contam_sim_score);	  
		}
	    }
	}
    }
 
  /*output to file*/
  for(lane=0; lane<GelInfo->NumberOfLanes;lane++)
    {
       fprintf(OUTFILE, "%i\t%i\t%.0f\n", GelInfo->NumberOfLanes-1-lane, roughness[lane], maxdiff[lane]); 
    } 
  

  for(row = 0; row <GelInfo->NumberOfLanes; row++)
    {
        free(medianrow[row]);
    }
  free(medianrow); 
  free(slope);
  for (col = 0; col < GelInfo->ImageWidth; col++) {
    free(laneintensity[col]);
  }
  free(laneintensity);

}


