/**@file node.c
 *\brief implements functions for extrapolating nodes to the edge, adjusting the boundary nodes, and writing lanes information to file*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>


#include "node.h"

/**\brief linear extrapolate out to get the peak positions at the edges (0th and last node), modifies PeakPosition and NodePosition accordingly
 *@param PeakPosition holds y position of all nodes
 *@param NodePosition holds x position of nodes in each zone
 *@param GelInfo holds various gel parameters
*/
void extrapolateNodes(float **PeakPosition, float *NodePosition,  struct GelInfo *GelInfo) {
			
int	nodeID, laneID;
float	slope, offset, deltaxlow, deltaxhigh, deltaxdata;

  deltaxlow = NodePosition[2] - NodePosition[1]; 
  deltaxhigh = NodePosition[GelInfo->NumberOfZones]-NodePosition[GelInfo->NumberOfZones-1];

  for (laneID = 0; laneID < GelInfo->NumberOfLanes; laneID++) {

    /*extrapolate for 0th node*/
    slope = (fabs(PeakPosition[2][laneID])- fabs(PeakPosition[1][laneID])) / deltaxlow;
    PeakPosition[0][laneID] =  -1* ( slope * (NodePosition[0]-NodePosition[1]) + fabs(PeakPosition[1][laneID]));
    
    /*extrapolate for last node*/
    slope = (fabs(PeakPosition[GelInfo->NumberOfZones][laneID])-fabs(PeakPosition[GelInfo->NumberOfZones-1][laneID]))/deltaxhigh;
    PeakPosition[GelInfo->NumberOfZones+1][laneID] = -1 *(slope * (NodePosition[GelInfo->NumberOfZones+1] - NodePosition[GelInfo->NumberOfZones]) + fabs(PeakPosition[GelInfo->NumberOfZones][laneID]));
  }
  
}

/**\brief move first node to stopPreData and extrapolate out to get peak positions at the edges, modifies PeakPosition and NodePosition accordingly
 *
 *the negative positions denoting interpolation is reset to positive
 *
 *@param PeakPosition holds y position of all nodes
 *@param NodePosition holds x position of nodes in each zone
 *@param GelInfo holds various gel parameters
*/
void adjustBoundaryNodes(float **PeakPosition, float *NodePosition,  struct GelInfo *GelInfo) {
			
int	nodeID, laneID;
float	slope, offset, deltaxlow, deltaxhigh, deltaxdata;
 int     oldnode1;
 float oldpeak1;


/*accept all interpolated peaks*/
  for (nodeID = 0; nodeID < GelInfo->NumberOfZones+2; nodeID++) {
    for (laneID = 0; laneID < GelInfo->NumberOfLanes; laneID++) {
      PeakPosition[nodeID][laneID] = fabs(PeakPosition[nodeID][laneID]);
    }
  }

  /*set node right before data to be at StopPreData
  * and extrapolate out to get the peak positions at the edges (0th and last node)*/

  deltaxlow = NodePosition[2] - NodePosition[1]; 
  deltaxhigh = NodePosition[GelInfo->NumberOfZones]-NodePosition[GelInfo->NumberOfZones-1];
  deltaxdata = NodePosition[GelInfo->FirstDataZone+1]-NodePosition[GelInfo->FirstDataZone];

  /*storing first node location in case it gets moved*/
  oldnode1 = NodePosition[1]; 

  NodePosition[GelInfo->FirstDataZone] = (float)GelInfo->StopPreData;

  for (laneID = 0; laneID < GelInfo->NumberOfLanes; laneID++) {

    /*storing first peak location in case it gets moved*/
    oldpeak1 = PeakPosition[1][laneID];

    /*move node before data*/
    slope = (PeakPosition[GelInfo->FirstDataZone+1][laneID]-PeakPosition[GelInfo->FirstDataZone][laneID]) / deltaxdata;
    PeakPosition[GelInfo->FirstDataZone][laneID] = slope * (NodePosition[GelInfo->FirstDataZone]-NodePosition[GelInfo->FirstDataZone+1]) + PeakPosition[GelInfo->FirstDataZone+1][laneID];
    
    /*extrapolate for 0th node*/
    slope = (PeakPosition[2][laneID]-oldpeak1) / deltaxlow;
    PeakPosition[0][laneID] =  slope * (NodePosition[0]-oldnode1) + oldpeak1;
    
    /*extrapolate for last node*/
    slope = (PeakPosition[GelInfo->NumberOfZones][laneID]-PeakPosition[GelInfo->NumberOfZones-1][laneID])/deltaxhigh;
    PeakPosition[GelInfo->NumberOfZones+1][laneID] = slope * (NodePosition[GelInfo->NumberOfZones+1] - NodePosition[GelInfo->NumberOfZones]) + PeakPosition[GelInfo->NumberOfZones][laneID];
  }
}

/**\brief write node and peak positions to lanes file
 *@param PeakPosition holds y position of all nodes
 *@param lanespath path to the destination lanes file
 *@param NodePosition holds x position of nodes in each zone
 *@param GelInfo holds various gel parameters
*/
void writeLanes(float **PeakPosition, char *lanespath, float *NodePosition, struct GelInfo *GelInfo) {

char	fullName[FILE_NAME_LEN];
FILE	*LANES;
int	nodeID, laneID;

  sprintf(fullName,"%s%s",GelInfo->ProjectDir,lanespath);
  if ((LANES=fopen(fullName,"w"))==NULL) {
    fprintf(stderr,"Cannot open lane file\n");
    exit(1);
  }
  fprintf(LANES,"LANES\n");
  for (nodeID = 0; nodeID < GelInfo->NumberOfZones+2; nodeID++) {
    fprintf(LANES, "z=%d x=%.1f", nodeID, NodePosition[nodeID]);
    for (laneID = GelInfo->NumberOfLanes-1; laneID >= 0; laneID--) {
      fprintf(LANES, " %.1f", PeakPosition[nodeID][laneID]);
    }
    fprintf(LANES, "\n");
  }
  fprintf(LANES, "StopPreData= %i\n", GelInfo->StopPreData);
  fclose(LANES);
}
