/**@file peaks.c
 *\brief contains function for finding guassian-like peaks in a vector of intensity
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "peaks.h"

/**\brief find all the peaks in an intensity profile.
 *
 *the algorithm looks for gaussian-like peaks by looking for local maxima in
 *a fit score.
 *the fit score is the difference of the integrated intensity within 1 fwhm and the integrated intensity between 1 and 2 fwhm away from the point, normalizded by the integrated intensity within 2 fwhm of the point.
 *
 *@param spec intensity trace in which peaks are found
 *@param length length of intensity trace
 *@param pcentroid array where centroid of located peaks are written
 *@param psize array where the area of the peaks are written
 *@param loch lower limit of the index for the search
 *@param hich upper limit of the index for the search
 *@param ifwhm integer full width at half maximum of expected peak
 *@param sigma minimum "sigma" (fit score) value a point must yield for the algorithm to look for a peak
 *@param maxpk maximum number of peak to be located
 *@returns the number of peaks found
 */
int specFindPeak(INT32 *spec, int length, float *pcentroid, float *psize,
		int loch, int hich, int ifwhm, float sigma, int maxpk) {

float	deno, root, peak1, peak2, peak3;
float	p1, p2, p3, s2, w4, pc, sum, sum2;
int	n, i, j, j1, j2, lock, jodd, jd, nd, jl, jr, np;
INT32   wings, ctr;
float 	found;
int	istart, istop;

  np = 0;
  n = 0;
  lock = 0;
  p1 = 0.f;
  p2 = 0.f;
  p3 = 0.f;
  peak1 = 0.f;
  peak2 = 0.f;
  peak3 = 0.f;
  sum = 0.f;
  sum2 = 0.f;
  if (ifwhm < 1) ifwhm = 1;
  j1 = -((ifwhm << 1) - 1) / 4; /*-(ifwhm * 2 - 1)/4*/
  j2 = j1 + ifwhm - 1;
  jodd = 1 - (ifwhm - (ifwhm / 2 << 1)); /*jodd 0 if ifwhm odd, 1 if ifwhm even*/

  istart = MAX(0, loch - ifwhm);
  istop = MIN(length - 1, hich + ifwhm);
  
  for (nd = istart; nd <= istop; ++nd) {
    /* sum counts center and wings*/
    ctr = 0;
    wings = 0;
    for (j = j1; j <= j2; ++j) {
      jd = nd + j;
      ctr += spec[jd - 1];
      if (j > 0) {
	jl = nd - j + j1;
	jr = jd + j2;
	wings = wings + spec[jl-1] + spec[jr-1];
      }
    }
    /* if ifwhm is odd, average data in split cells at ends to be part of the wings*/
    w4 = 0.f;
    if (jodd == 0) {
      jl = nd + j1 + j1;
      jr = nd + j2 + j2;
      w4 = (spec[jl-2] + spec[jl-1] + spec[jr-1] + spec[jr]) * .25f;
    }
    /* compute height of second derivative (neg) relative to noise*/
    s2 = sum; /*last sum = the area of center - wings*/
    sum = (float) (ctr - wings) - w4;
    root = sqrt((float) (ctr + wings + 1) + w4); /*sqrt of entire area*/
    p1 = p2; /*sum/deno of 2 points ago*/
    p2 = p3; /*sum/deno of last point*/
    deno = root;
    if (deno < 1e-6f) deno = 1e-6f;
    p3 = sum / deno; /*sum / sqrt of the entire area*/
    /*only look when center is large*/
    if (lock != 0) {
      if (p2 > peak2 && p3 < p2) /*if p2 is a local maximum*/{
	/* save three values at relative maximum */
	peak1 = p1;
	peak2 = p2;
	peak3 = p3;
	sum2 = s2;  /*ctr-wings of the peak*/
	np = nd - 1; /*position of the peak*/
      }
      if (p3 >= sigma) continue; /*keep looking until sum/deno is less than threshold again, then we have passed a peak*/

      /* estimate location and crude size of peak */
      deno = peak1 - peak2 + peak3 - peak2;
      if (fabs(deno) < 1e-6f) deno = 1e-6f;
      pc = (peak1 - peak3) * .5f / deno; /*fine adjustment of the peak location based on size of neighbours*/
      
      found = (float) np + pc + (float) jodd * .5f - 1.;
      if ((found >= loch)&&(found <= hich)) {
        if (++n > maxpk) n = maxpk;
        pcentroid[n-1] = found;
        psize[n-1] = sum2 + sum2; /*psize is ctr-wings * 2*/
      }
      /*reset, until a big enough center is seen*/
      lock = 0;
      peak2 = 0.f;
    }
    if (p3 >= sigma) lock = 1; /*if sum/deno of this point bigger than given cutoff, then start looking for peaks from this point*/
  }
  return n;
}
