/**@file spacing.c
 *\brief contains algorithm for finding the dominant lane spacing in the
 * gel as well as the dominant spacing of each zone
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "spacing.h"

/**\brief finds the maximum of FFT power spectrum within a set range in y-direction within each zone
 *@param ZoneProjection points to 1D y-projection in each of the zone
 *@param GelInfo holds various gel parameters
 *@param Spacing points to array to hold dominant spacing for each zone
 *@param minfreq minimum frequency of range of interest
 *@param maxfreq maximum frequency of range of interest
*/
static void fillSpacing(INT32 **ZoneProjection, struct GelInfo *GelInfo, float *Spacing, float minfreq, float maxfreq) {
fftw_complex *fft;
double *spectrum, *power;
fftw_plan plan;
int length, Nd2p1, specID, channel, chMax, StartSearch;
double Max, peakposition;
  
  length = GelInfo->ImageHeight;
  Nd2p1 = length/2 + 1;
  spectrum = fftw_malloc(sizeof(double) * length);
  fft = fftw_malloc(sizeof(fftw_complex) * Nd2p1);
  power = calloc(Nd2p1, sizeof(double));
  plan = fftw_plan_dft_r2c_1d(length, spectrum, fft, FFTW_ESTIMATE);

  StartSearch = GelInfo->NumberOfLanes / GelInfo->MarkerPeriod; /*initial guess, max number of marker lanes, avoids DC term*/

  for (specID = 0; specID < GelInfo->NumberOfZones; specID++) {
    for (channel = 0; channel < length; channel++) {
      spectrum[channel] = (double) (ZoneProjection[specID][channel]); /*YProjection of a zone*/
    }
    fftw_execute(plan);
    for (channel = 0; channel < Nd2p1; channel++) {
      power[channel] = (fft[channel][0]*fft[channel][0])+(fft[channel][1]*fft[channel][1]);
    }
    chMax = 1;
    Max = 0;
    for (channel = (int) minfreq; channel < (int) maxfreq; channel++) {
        if(power[channel] > Max) {
          chMax = channel;
	  Max = power[channel];
        }
    }
    /*weighted average of peak and neighbours to give better noise rejection, and higher resolution*/
    peakposition = (power[chMax-1]*(float)(chMax-1)) + (power[chMax]*(float)chMax) + (power[chMax+1]*(float)(chMax+1));
    peakposition /= power[chMax-1] + power[chMax] + power[chMax+1];
    Spacing[specID] = (float)length / peakposition; /*height/dominant frequency gives most likely spacing in pixels, this can be markerspacing or lane spacing*/
  }
  
  fftw_destroy_plan(plan);
  fftw_free(spectrum);
  fftw_free(fft);
  free(power);
}

/**\brief find lane spacing
 *
 *returns -1 if no valid spacing found
 *
 *@param FLAGOUT_FILE writable file handle to data log file
 *@param ZoneProjection points to 1D y-projection in each of the zone
 *@param GelInfo holds various gel parameters
 *@param Spacing points to array to hold dominant spacing for each zone
 *@param Verbose turn on text output to screen
 *@returns the generic lane spacing for the gel
 */
float findSpacing (FILE *FLAGOUT_FILE, INT32 **ZoneProjection, struct GelInfo *GelInfo, float *Spacing,
                  int Verbose) {

  int i, j, maxseen;
int *ISpacing, *seenNtimes, imin, nmin;
float minimum, GenericLaneSpacing;
float maxspacing, minspacing;
float maxfreq, minfreq;

   minfreq = (float)GelInfo->NumberOfLanes; /*if lanes span the entire image*/
   maxfreq = (float)GelInfo->NumberOfLanes / GelInfo->GelSpanFactor; /*if lanes span a minimum required portion of the image*/

   fillSpacing(ZoneProjection, GelInfo, Spacing, minfreq, maxfreq);
  if (Verbose) fprintf(stderr, "Probable lane spacing for each zone:\n");
  
  fprintf(FLAGOUT_FILE, "Probable lane spacing:");
  for (i = 0; i < GelInfo->NumberOfZones; i++) {
      if(Verbose) fprintf(stderr, "\tzone %d = %.2f\n", i, Spacing[i]);
      fprintf(FLAGOUT_FILE, "\t%.2f", Spacing[i]);
    }     
  fprintf(FLAGOUT_FILE, "\n");
  

  maxspacing = (float)GelInfo->ImageHeight/(float)GelInfo->NumberOfLanes;
   minspacing = (float)GelInfo->ImageHeight*GelInfo->GelSpanFactor/(float)GelInfo->NumberOfLanes; /*lanes should span set amount the gel*/


  ISpacing = (int *)calloc(GelInfo->NumberOfZones, sizeof(int));
  seenNtimes = (int *)calloc(GelInfo->NumberOfZones, sizeof(int));

  for (i = 0; i < GelInfo->NumberOfZones; i++) {
    ISpacing[i] = (int) (Spacing[i] + 0.5);
  }
  /*tracking how many times an ispacing is seen, the most frequent one closest to top of gel is used*/
  for (i = 0; i < GelInfo->NumberOfZones; i++) {
    for (j = 0; j < GelInfo->NumberOfZones; j++) {
      if (ISpacing[i] == ISpacing[j]) {
        seenNtimes[i]++;
      }
    }
    if (seenNtimes[i] < 2) ISpacing[i] = GelInfo->ImageHeight;
  } /*there must be at least one other zone with same ISpacing for the result to be used, spacing is set to image height (invalid) otherwise*/

  /*search for max, closest to top of gel*/
  imin = GelInfo->ImageHeight;
  maxseen = 0;
  for (i = 0; i < GelInfo->NumberOfZones; i++) {
    if (seenNtimes[i] > maxseen){
      imin = ISpacing[i];
      maxseen = seenNtimes[i];
    }
  }

  if (imin < GelInfo->ImageHeight) {
    nmin = 0;
    GenericLaneSpacing = 0.0;
    for (i = 0; i < GelInfo->NumberOfZones; i++) {
      if (ISpacing[i] == imin) {
        nmin++;
        GenericLaneSpacing += Spacing[i];
      }
    }
    GenericLaneSpacing /= nmin;
    
    if (Verbose) {
      fprintf(stderr,"Generic lane spacing = %.2f\n", GenericLaneSpacing);
    }
    fprintf(FLAGOUT_FILE,"Generic lane spacing:\t%.2f\n", GenericLaneSpacing);
  }
  else {
    GenericLaneSpacing = -1.0; /*invalid spacing*/
  }
  free(ISpacing);
  free(seenNtimes);

  /*refill spacing to identify marker dominant zones*/
  minfreq=GelInfo->NumberOfMarkers;
  maxfreq=GelInfo->ImageHeight/2 + 1;

  fillSpacing(ZoneProjection, GelInfo, Spacing, minfreq, maxfreq);
  fprintf(FLAGOUT_FILE, "Dominant spacing:");
  if (Verbose) fprintf(stderr, "Dominant spacing for each zone:\n");
  for (i = 0; i < GelInfo->NumberOfZones; i++) {
      if(Verbose) fprintf(stderr, "\tzone %d = %.2f\n", i, Spacing[i]);
      fprintf(FLAGOUT_FILE, "\t%.2f", Spacing[i]);
    }     

  fprintf(FLAGOUT_FILE, "\n");
  return GenericLaneSpacing;
}
